-- Displays the puzzle game (this is the world object for the BattleGameMenu)

use_safeglobals()

local Grid = require("Battlegrounds/Grid")
local Constants = require("Battlegrounds/BattleGroundConstants")
local FX = require("FXContainer")
local SpinHelpers = require("Utils/SpinHelpers")
local Spell = require("Spells/Spell")
local DraggingSupport = require("Utils/DraggingSupport")

class "BattleGroundView" (GameObject)

DraggingSupport:AddToClass(BattleGroundView)

BattleGroundView.AttributeDescriptions = AttributeDescriptionList()

-- This is the spacing between gems for various gem sizes 
--   (THESE WOULD ACTUALLY BE PLATFORM DEPENDENT)

local LARGE_GEM_SIZE 	= Constants.LARGE_GEM_SIZE
local LARGE_GEM_SCALE	= Constants.LARGE_GEM_SCALE
local MEDIUM_GEM_SIZE 	= Constants.MEDIUM_GEM_SIZE
local MEDIUM_GEM_SCALE	= Constants.MEDIUM_GEM_SCALE
local SMALL_GEM_SIZE 	= Constants.SMALL_GEM_SIZE
local SMALL_GEM_SCALE	= Constants.SMALL_GEM_SCALE

-- Defines the delay for a gem droping that is relative to its height on the board
--  (this delay is the delay that stops all gems in a row from falling at the same time)

local DROP_DELAY		= Constants.DROP_DELAY
local DROP_XDELAY		= Constants.DROP_XDELAY

local FALL_ACCELERATION	= Constants.FALL_ACCELERATION

local HEROIC_EFFORT_COUNT = Constants.HEROIC_EFFORT_COUNT

local CASCADE_SOUNDS	= {
	"pq2audio/board/fx_cascade1",
	"pq2audio/board/fx_cascade2",
	"pq2audio/board/fx_cascade3",
	"pq2audio/board/fx_cascade4",
	"pq2audio/board/fx_cascade5",
}

local SMALL_FONT_HEIGHT = 44

local TUTORIAL_STEALTH     = 1
local TUTORIAL_MATCH5      = 2
local TUTORIAL_MATCH4      = 3
local TUTORIAL_SPELLREADY  = 4
local TUTORIAL_ITEMREADY   = 5
local TUTORIAL_MATCHSKULL5 = 6
local TUTORIAL_MATCHSKULL  = 7
local TUTORIAL_MATCHWILD   = 8
local TUTORIAL_MATCHMANA   = 9
local TUTORIAL_MATCHACTION = 10
local TUTORIAL_BLOCKINGGEM = 11
local TUTORIAL_GAINEDMANA  = 12
local TUTORIAL_POSITIVEDEF = 13
local TUTORIAL_POSITIVEATT = 14
local TUTORIAL_ENEMYSTATUS = 15
local TUTORIAL_SELFSTATUS  = 16

function BattleGroundView:__init(clid)
	super("BGND")
	
	self.swap_request_id = 1
	self.isMoving = false
	self.reanimationCount = 0
	
	self:SetCursorInteract(false)
end

function BattleGroundView:RefreshView()
	Constants.SetBoardLayoutConstants()
	LARGE_GEM_SIZE		= Constants.LARGE_GEM_SIZE
	LARGE_GEM_SCALE		= Constants.LARGE_GEM_SCALE
	MEDIUM_GEM_SIZE 	= Constants.MEDIUM_GEM_SIZE
	MEDIUM_GEM_SCALE	= Constants.MEDIUM_GEM_SCALE
	SMALL_GEM_SIZE		= Constants.SMALL_GEM_SIZE
	SMALL_GEM_SCALE		= Constants.SMALL_GEM_SCALE
	
	if self.width and self.height then
		self:SetupView()
		for x=1, self.width do
			for y=1, self.height do
				if self.grid[x] and self.grid[x][y] then
					local name = self.grid[x][y].name
					
					self:DestroyGemAt(x, y)
					self:AddNewGemToBoard(x, y, name)
				end
			end
		end
	end
end

function BattleGroundView:PreDestroy()
	if leakdetect then
		leakdetect.PreDestroyHook(self)
	end
	scheduler:StopAllThreadsForObject(self)
	scheduler:StopAllThreadsForObject(Spell)		-- #3495, stops any scheduled particle effects from appearing after this world is destroyed
end

function BattleGroundView:GemCoordToPixel(x, y)
	return math.round((x - 1) * self.gem_step + (self.gem_step / 2)), math.round(y * self.gem_step - (self.gem_step / 2))
end

function BattleGroundView:GemPixelToCoord(px, py)
	return math.round((px - (self.gem_step / 2)) / self.gem_step + 1), math.round((py + (self.gem_step / 2)) / self.gem_step)
end

function BattleGroundView:SetClient(client) -- client can be nil, if this view accepts no inputs
	self.client = client
	self.player_id = client:GetAttribute("team")
	
	self.tutorials = {
	
		{ type = "Stealth",        check = true,  show = false, },
		{ type = "MatchFive",      check = true,  show = false, },
		{ type = "MatchFour",      check = true,  show = false, },
		{ type = "SpellReady",     check = true,  show = false, },
		{ type = "ItemReady",      check = true,  show = false, },
		{ type = "MatchSkull5",    check = true,  show = false, },
		{ type = "MatchSkull",     check = true,  show = false, },
		{ type = "MatchWild",      check = true,  show = false, },
		{ type = "MatchMana",      check = true,  show = false, },
		{ type = "MatchAction",    check = true,  show = false, },
		{ type = "BlockingGem",    check = false, show = false, },
		{ type = "GainedMana",     check = true,  show = false, },
		{ type = "NonzeroDefense", check = true,  show = false, },
		{ type = "NonzeroAttack",  check = true,  show = false, },
		{ type = "EnemyStatus",    check = true,  show = false, },
		{ type = "SelfStatus",     check = true,  show = false, },
	}
end

function BattleGroundView:AddLocalClient(client) -- adds a local multiplayer client
	if not self.localClients then
		self.localClients = { }
	end
	table.insert(self.localClients, client)
end

function BattleGroundView:SetCurrentMoveID(currentMoveID)
	self.currentMoveID = currentMoveID
end

function BattleGroundView:CreateGem(name)
	local gem = GameObjectManager:ConstructLocal("BGEM")
	
	gem:SetInfo(Gems[name], self.gem_scale, self.gem_size)
	gem:SetCursorInteract(true)

	return gem
end

function BattleGroundView:DestroyGem(gem)
	if gem then
		GameObjectManager:Destroy(gem)
	end
end

function BattleGroundView:DestroyGemAt(x, y)
	assert(x)
	assert(y)
	
	self:DestroyGem(self.grid[x][y])
	self.grid[x][y] = nil
end

function BattleGroundView:OnEventGameInfo(event)
	LOG("BattleGroundView:OnEventGameInfo(event)")

	self.width = event:GetAttribute("width")
	self.height = event:GetAttribute("height")
	self.grid = {}
	
	for x=1,self.width do
		self.grid[x] = {}
	end
	
	self.grid = Grid:Construct(self.width, self.height, self.grid)
	
	self.turnTimer = event:GetAttribute("turnTimer")
	self.turnLimit = event:GetAttribute("turnLimit")
	
	self:SetupView()
end

function BattleGroundView:SetupView()
	-- retrieve the arena size, its a PAD widget in the menu
	local menu = SCREENS.BattleGameMenu
	
	self.arena_width = menu:get_widget_w("pad_arena")
	self.arena_height = menu:get_widget_h("pad_arena")
	
	LOG("Grid is %dx%d with arena %dx%d", self.width, self.height, self.arena_width, self.arena_height)
	
	-- select size of gems
	local offset = GetScreenOffset()
	self.gem_step = self.arena_width / self.width
	
	if math.ceil(self.gem_step / SMALL_GEM_SCALE) < SMALL_GEM_SIZE then
		error(string.format("Board size of %dx%d is too large for arena size of %dx%d", self.width, self.height, self.arena_width, self.arena_height))
	elseif math.ceil(self.gem_step / MEDIUM_GEM_SCALE) < MEDIUM_GEM_SIZE then
		self.gem_size = SMALL_GEM_SIZE
		self.gem_scale = SMALL_GEM_SCALE
		if offset == 0 then	-- not widescreen
			self.temp_scale = 0.6
		else
			self.temp_scale = 0.7
		end
	elseif math.ceil(self.gem_step / LARGE_GEM_SCALE) < LARGE_GEM_SIZE then
		self.gem_size = MEDIUM_GEM_SIZE
		self.gem_scale = MEDIUM_GEM_SCALE
		if offset == 0 then	-- not widescreen
			self.temp_scale = 0.8
		else
			self.temp_scale = 0.9
		end
	else
		self.gem_size = LARGE_GEM_SIZE
		self.gem_scale = LARGE_GEM_SCALE
		if offset == 0 then	-- not widescreen
			self.temp_scale = 0.9
		else
			self.temp_scale = 1.0
		end
	end
	
	-- Construct our overlays here
	self.hint1 = GameObjectViewManager:Construct("GraphicAsset", "anim_hint")
	self.hint1:SetScale(self.temp_scale)
	self.hint1:SetSortingValue(Constants.HINT_ARROW_SORTING_VALUE)
	self.hint1:SetAlpha(0.0)	--Constants.HINT_ARROW_ALPHA_VALUE)
	
	self.hint2 = GameObjectViewManager:Construct("GraphicAsset", "anim_hint")
	self.hint2:SetScale(self.temp_scale)
	self.hint2:SetSortingValue(Constants.HINT_ARROW_SORTING_VALUE)
	self.hint2:SetAlpha(0.0)	--Constants.HINT_ARROW_ALPHA_VALUE)
	
	self.player1Overlay1 = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_g")
	self.player1Overlay1:SetScale(self.temp_scale)
	self.player1Overlay1:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self.OverlayFast = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_b")
	self.OverlayFast:SetScale(self.temp_scale)
	self.OverlayFast:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self.player1Overlay2 = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_g")
	self.player1Overlay2:SetScale(self.temp_scale)
	self.player1Overlay2:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self.player2Overlay1 = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_r")
	self.player2Overlay1:SetScale(self.temp_scale)
	self.player2Overlay1:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self.player2Overlay2 = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_r")
	self.player2Overlay2:SetScale(self.temp_scale)
	self.player2Overlay2:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self:OnDragInit(Constants.GEM_SELECTION_SORTING_VALUE, self.temp_scale)
end

function BattleGroundView:AddNewGemToBoard(x, y, name, invisible) -- noattach, does not add the gem as a child of the world
	assert(not self.grid[x][y])

	local newgem = self:CreateGem(name)
	
	newgem:SetPos(self:GemCoordToPixel(x, y))
	newgem.x = x
	newgem.y = y
	
	self.grid[x][y] = newgem
	
	if invisible then
		newgem:GetView():SetAlpha(0.0)
	end
	
	self:AddChild(newgem)
	
	return newgem
end

function BattleGroundView:OnEventBoardSnapshot(event)
	assert(event)
	assert(event:NumAttributes("x") == event:NumAttributes("y"))
	assert(event:NumAttributes("y") == event:NumAttributes("name"))
	
	local selectedGemX = 0
	local selectedGemY = 0
	if self.selectedGem then
		selectedGemX = self.selectedGem.x
		selectedGemY = self.selectedGem.y
	end
	
	for i=1, event:NumAttributes("x") do
		local x = event:GetAttributeAt("x", i)
		local y = event:GetAttributeAt("y", i)
		local gem = event:GetAttributeAt("name", i)
		local kind = event:GetAttribute("type")
		
		self:DestroyGemAt(x, y)
		
		if kind == "transform" or kind == "replace" then
			SoundSystem.Play("pq2audio/board/gemtransform")
		end
		
		if gem ~= "Empty" then
			self:AddNewGemToBoard(x, y, gem)
		end
	end
	
	if selectedGemX > 0 and selectedGemY > 0 then
		self.selectedGem = self.grid[selectedGemX][selectedGemY]
		self:SetSelectedGemOverlay(self.selectedGem, self.player_id, true)
		if SCREENS.BattleGameMenu.selected then
			self:SetSpinSpeed(true)
		end
	end
end

function BattleGroundView:SetSpinSpeed(fast)
	if self.selectedGem then
		if self.spinHandle and self.spinHandle[self.selectedGem] then
			SpinHelpers.EndSpin(self.spinHandle[self.selectedGem])
			if fast then
				--self.spinHandle[self.selectedGem] = SpinHelpers.StartOverlaySpin(self.selectedGem, "target", Constants.FAST_SPIN_RATE)
				self:ClearSelectedGemOverlay(self.selectedGem)
				self.selectedGem:AddOverlay("target", self.OverlayFast, 0, 0)
				if not self.spinHandle then
					self.spinHandle = {}
				end
				self.spinHandle[self.selectedGem] = SpinHelpers.StartOverlaySpin(self.selectedGem, "target", Constants.FAST_SPIN_RATE)
			else
				self:ClearSelectedGemOverlay(self.selectedGem)
				self.selectedGem:AddOverlay("target", self[string.format("player%dOverlay1", self.player_index)], 0, 0)
				if not self.spinHandle then
					self.spinHandle = {}
				end
				self.spinHandle[self.selectedGem] = SpinHelpers.StartOverlaySpin(self.selectedGem, "target", Constants.SPIN_RATE)
			end
		end
	end
end

function BattleGroundView:SetSelectedGemOverlay(gem, playerNum, isFirstSelection, noSpin) -- True if you are making the first selection before the swap
	if gem then
		SoundSystem.Play("pq2audio/board/select1")
		local overlay = self.player1Overlay1
		if playerNum == 1 then
			if not isFirstSelection then
				overlay = self.player1Overlay2
			end
		elseif playerNum == 2 then
			overlay = self.player2Overlay1
			if not isFirstSelection then
				overlay = self.player2Overlay2
			end
		end
		
		
		
		if self:IsLocalClient(self.activeClient)  or (not (self.client and self.client:GetAttribute("team") == playerNum)) then	-- HAX? - 2nd condition needed for AI overlay to show where the move was made. Why?
			self:ClearSelectedGemOverlay(gem)	-- For safetey's sake, clear the old one
			if self.targetIcon then
				gem:AddOverlay("spell", self.targetIcon, 0, 0)
			else
				gem:AddOverlay("target", overlay, 0, 0)
				
				if not noSpin then
					if not self.spinHandle then
						self.spinHandle = {}
					end
					self.spinHandle[gem] = SpinHelpers.StartOverlaySpin(gem, "target", Constants.SPIN_RATE)
				end
			end
		end
	end
end

function BattleGroundView:ClearSelectedGemOverlay(gem)
	if gem then
		SoundSystem.Play("pq2audio/board/select0")
		if self.spinHandle and self.spinHandle[gem] then
			SpinHelpers.EndSpin(self.spinHandle[gem])
		end
		if gem:GetOverlay("target").__ok then
			gem:RemoveOverlay("target")
		end
		if gem:GetOverlay("spell").__ok then
			gem:RemoveOverlay("spell")
		end
	end
	self:OnDragReset()
end

function BattleGroundView:IsAdjacent(gem1, gem2)
	local x1, y1 = gem1.x, gem1.y
	local x2, y2 = gem2.x, gem2.y
	
	if x1 == x2 then
		if math.abs(y1 - y2) == 1 then
			return true
		end
	elseif y1 == y2 then
		if math.abs(x1 - x2) == 1 then
			return true
		end
	end
	
	return false
end

function BattleGroundView:OnEventCursorEntered(event)
	local obj = event:GetAttribute("object")
	local x   = event:GetAttribute("x")
	local y   = event:GetAttribute("y")
	
	if obj and obj.IS_GEM then
		local foundTarget = true
		self.hoverGem = obj
		if self.targetOnly then
			foundTarget = false
			for _, gemName in ipairs(self.targetOnly) do
				if obj.name == gemName then
					foundTarget = true
				end
			end
			if foundTarget then
				local cursor = Interface.GetCurrentCursor()
				if cursor then
					cursor:SetGraphic("curs_select")
					cursor:SetActive(true)
				end
			else
				local cursor = Interface.GetCurrentCursor()
				if cursor then
					cursor:SetGraphic("curs_select_red")
					cursor:SetActive(true)
				end
			end
		end
		self:OnDragUpdate(obj)
	end
end

function BattleGroundView:OnEventCursorExited(event)
	local obj = event:GetAttribute("object")
	local x   = event:GetAttribute("x")
	local y   = event:GetAttribute("y")
	
	if self.hoverGem == obj then
		self.hoverGem = nil
	end
	
	if obj and obj.IS_GEM then
		local foundTarget = true
		if self.targetOnly then
			foundTarget = false
			for _, gemName in ipairs(self.targetOnly) do
				if obj.name == gemName then
					foundTarget = true
				end
			end
			if foundTarget then
				local cursor = Interface.GetCurrentCursor()
				if cursor then
					cursor:SetGraphic("curs_select_red")
					cursor:SetActive(true)
				end
			else
				local cursor = Interface.GetCurrentCursor()
				if cursor then
					cursor:SetGraphic("curs_select")
					cursor:SetActive(true)
				end
			end
		end
		self:OnDragUpdate(nil)
	end
end

function BattleGroundView:OnEventCursorAction(event)
	local x = event:GetAttribute("x")
	local y = event:GetAttribute("y")
	local obj = event:GetAttribute("object")
	local up = event:GetAttribute("up")
	
	if self.isMoving or self.pending_swap_request then	-- all gem clicks are ignored if there is a move pending
		return
	end
	
	local gx, gy = self:GemPixelToCoord(x, y)
	LOGF("BattleGroundView:OnEventCursorAction: x=%d, y=%d, obj=%s, up=%s (%d, %d)", x, y, tostring(obj), tostring(up), gx, gy)
	
	-- This view might be from a demo, in which case all input is ignored
	if (not self.client) and (not self.localClients) then
		return
	end
	
	-- It might not be this players turn
	if not self.allowMoves then
		LOG("Click ignored, not allowed to move")
		return
	end
	
	if IsOpen("OppTutorialMenu") then
		--[[
		if self.input_gems then
			if (obj ~= self.input_gems[1]) and (obj ~= self.input_gems[2]) then
				return
			end
		end
		self.input_gems = nil
		]]--
		SCREENS.OppTutorialMenu:Close()
	end
	
	-- could be targetting
	if self.targetCallback then
		if obj == self then -- click on world (check for a gem)
			SCREENS[self.activeClient]:DisableTargettingLockdown()
		elseif obj and obj.IS_GEM then
			local foundTarget = true
			if self.targetOnly then
				foundTarget = false
				for _, gemName in ipairs(self.targetOnly) do
					if obj.name == gemName then
						foundTarget = true
					end
				end
			end
			if foundTarget then
				if self.selectedGem then
					self:ClearSelectedGemOverlay(self.selectedGem)
					self.selectedGem = obj
					self:SetSelectedGemOverlay(self.selectedGem, self.player_id, true)
				else
					self:HideHint()
					self.selectedGem = obj
					self:SetSelectedGemOverlay(self.selectedGem, self.player_id, true)
				end
				AttachParticles(self, 'RingWhite', self:GemCoordToPixel(gx, gy))
				self.targetCallback(gx, gy)
				self:EndTargettingMode()
			else
				SoundSystem.Play("pq2audio/ui/illegal")
			end
		end
		
		return -- no further processing
	end
	
	if up then
		if self:OnDragEnd(obj, gx, gy) then
			return
		end
	else
		return self:OnDragStart(obj)
	end
	
	-- NOTE: its handy to have the object given to us by the event, but empty locations 
	-- are not game objects, so if the event:object is the world, we should have a look
	-- and see if the click was on an empty grid slot
	
	if obj == self then
		if gx > 0 and gx <= self.width and gy > 0 and gy <= self.height then
			if not self.grid[gx] or not self.grid[gx][gy] then
				-- hmm we are trying to swap with an empty gem
				if self.selectedGem then
					self:SendSwapRequest(self.selectedGem.x, self.selectedGem.y, gx, gy)
				end
			end
		end
		
		return
	end
	
	if obj and obj.IS_GEM and obj.def.canSwap then
		if self.selectedGem then
			if obj == self.selectedGem then		-- clicked the selected gem, deselect it
				self:ShowHint()
				self:ClearSelectedGemOverlay(self.selectedGem)
				self.selectedGem = nil
			elseif not self:IsAdjacent(self.selectedGem, obj) then	-- clicked a gem that is not adjacent, select it instead
				self:ClearSelectedGemOverlay(self.selectedGem)
				self.selectedGem = obj
				self:SetSelectedGemOverlay(self.selectedGem, self.player_id, true)
			else	-- clicked a different gem, swap the selected and targeted gems
				self.targetGem = obj
				self:SetSelectedGemOverlay(self.targetGem, self.player_id, false)
				
				self:SendSwapRequest(self.selectedGem.x, self.selectedGem.y, self.targetGem.x, self.targetGem.y)
			end
		else	-- clicked a new gem
			self:HideHint()
			self.selectedGem = obj
			self:SetSelectedGemOverlay(self.selectedGem, self.player_id, true)
		end
	end
end

function BattleGroundView:SendSwapRequest(x1, y1, x2, y2)
	assert(x1)
	assert(y1)
	assert(x2)
	assert(y2)

	local swapEvent = GameEventManager:Construct("SwapRequest")
	
	swapEvent:SetAttribute("source_x", x1)
	swapEvent:SetAttribute("source_y", y1)
	swapEvent:SetAttribute("target_x", x2)
	swapEvent:SetAttribute("target_y", y2)
	swapEvent:SetAttribute("request_id", self.swap_request_id)
	swapEvent:SetAttribute("client", self.activeClient)
	
	GameEventManager:Send(swapEvent)
	
	self.pending_swap_request = self.swap_request_id
	self.swap_request_id = self.swap_request_id + 1
	
	-- Disable the move text
	SCREENS[self.activeClient]:EndMove()
end

function BattleGroundView:MoveGemTo(gem, target_x, target_y)
	local target_x_pix, target_y_pix = self:GemCoordToPixel(target_x, target_y)
	local x_pix, y_pix = gem:GetX(), gem:GetY()
	
	self.grid[target_x][target_y] = gem
	
	gem.x = target_x
	gem.y = target_y
	
	local moveToTarget = MovementManager:Construct("Smooth")
	
	moveToTarget:SetAttribute("Duration", Constants.GEM_SWAP_TIME)
	moveToTarget:SetAttribute("StartX", x_pix)
	moveToTarget:SetAttribute("StartY", y_pix)
	moveToTarget:SetAttribute("EndX", target_x_pix)
	moveToTarget:SetAttribute("EndY", target_y_pix)
	
	gem:SetMovementController(moveToTarget)
end

function BattleGroundView:OnEventSwapNotify(event)
	self:RemoveHint()
	
	if self.pending_swap_request and self:IsLocalClient(event:GetAttribute("client")) then -- this is confirmation of a swap request that this client sent
		assert(event:GetAttribute("request_id") == self.pending_swap_request, string.format("%s to %s", tostring(event:GetAttribute("request_id")), tostring(self.pending_swap_request)))
		
		-- release the UI
		self.pending_swap_request = nil
	end
	
	self.allowMoves = false
	self.isMoving = true
	
	local sx, sy = event:GetAttribute("source_x"), event:GetAttribute("source_y")
	local tx, ty = event:GetAttribute("target_x"), event:GetAttribute("target_y")
	
	local gem1 = self.grid[sx][sy]
	local gem2 = self.grid[tx][ty]
	
	assert(self.grid[tx])
	assert(self.grid[sx])
	
	-- remove any old overlays
	self:ClearSelectedGemOverlay(self.selectedGem)	-- Do this in case the user has set up selected gem overlays
	self.selectedGem = nil
	
	if self.targetGem then
		self:ClearSelectedGemOverlay(self.targetGem)
		self.targetGem = nil
	end
	
	self:ClearSelectedGemOverlay(gem1)
	self:ClearSelectedGemOverlay(gem2)
	
	-- set the new overlays
	local player = event:GetAttribute("client")
	if gem1 then
		self:SetSelectedGemOverlay(gem1, player:GetAttribute("team"), true, true)
	end
	if gem2 then
		self:SetSelectedGemOverlay(gem2, player:GetAttribute("team"), false, true)
	end
	
	-- move gems
	if gem1 then
		self:MoveGemTo(gem1, tx, ty)
	else
		self.grid[tx][ty] = nil
	end
	
	if gem2 then
		self:MoveGemTo(gem2, sx, sy)
	else
		self.grid[sx][sy] = nil
	end
end

function BattleGroundView:OnEventSwapRejected(event)
	if self.pending_swap_request then
		SoundSystem.Play("pq2audio/ui/illegal")
		local sx_pix, sy_pix = self:GemCoordToPixel(event:GetAttribute("source_x"), event:GetAttribute("source_y"))
		local tx_pix, ty_pix = self:GemCoordToPixel(event:GetAttribute("target_x"), event:GetAttribute("target_y"))
		local x = (sx_pix + tx_pix) / 2
		local y = (sy_pix + ty_pix) / 2
		SCREENS.EffectsMenu:PlayFallingTextEffectAt(x, y, "[ILLEGAL_MOVE]")
		
		assert(event:GetAttribute("request_id") == self.pending_swap_request)
		assert(self:IsLocalClient(event:GetAttribute("client")))
		
		-- release the UI
		self:ShowHint()
		
		self.pending_swap_request = nil
		
		self:ClearSelectedGemOverlay(self.selectedGem)
		
		if self.selectedGem then
			self.x = self.selectedGem.x
			self.y = self.selectedGem.y
		end
		self.selectedGem = nil
		
		if self.targetGem then
			self:ClearSelectedGemOverlay(self.targetGem)
			self.targetGem = nil
		end
		
		if self:IsLocalClient(self.activeClient) then
			self:SetupSelection()
		end
	end
end

function BattleGroundView:GetIndexFromClient(client)
	assert(client)
	
	local clients = SCREENS.BattleGameMenu.clients
	for i, c in ipairs(clients) do
		if c == client then
			return i
		end
	end
	
	error("Client not found, index is undeterminable")
end

function BattleGroundView:OnEventSpellInvalidNotify(event)
	self:RemoveHint()
	local spellInfo = SPELLS[event:GetAttribute("spell")]
	--SCREENS.EffectsMenu:PlaySpellTextEffect(spellInfo.invalid_message) 			-- or
	SCREENS.EffectsMenu:PlayDefaultSpellTextEffect(spellInfo.invalid_message)
end

function BattleGroundView:OnEventSpellResistedNotify(event)
	self:RemoveHint()
	
	self.allowMoves = false
	
	SoundSystem.Play("pq2audio/board/spell_default")
	
	SoundSystem.Play("pq2audio/board/fx_resist")
	
	local spellInfo = SPELLS[event:GetAttribute("spell")]
	
	scheduler:AddChildThread(self, function ()
		local Spell = require("Spells/Spell")
		local heading, text, startTime = Spell:ShowStandardSpellHeader(spellInfo.title)
		
		Spell:Pause(Spell.SPELL_INTRO_PAUSE)
		Spell:ShowStandardEffectSubtitle("[SPELL_RESISTED]", Spell.SPELL_MIN_DURATION)
		Spell:RemoveStandardSpellHeader(heading, text, startTime)
	end, "SpellResistedNotify")
	
	if spellInfo.id == "CRUSHINGKILL" or spellInfo.id == "SUBJUGATE" then
		Achievements.AwardAchievement(1, "THATWASLUCKY")
	end
end

function BattleGroundView:OnEventSpellNotify(event)
	self:RemoveHint()
	
	self.allowMoves = false
	
	local spellInfo = SPELLS[event:GetAttribute("spell")]
	
	if spellInfo.Notify then
		scheduler:StartChildThread(self, function () spellInfo:Notify(self.grid, event) end, "Spell Notify")
	end

	-- sometimes spells are not as a result of the click of a button
	if event:GetAttribute("index") ~= -1 then
		local client = event:GetAttribute("client")
		local client_index = self:GetIndexFromClient(client)
		
		for i=1, 16 do
			local effectEvent = GameEventManager:Construct("SpellNotifyEffect")
			effectEvent:SetAttribute("effect_index", i)
			effectEvent:SetAttribute("widget_index", event:GetAttribute("index"))
			effectEvent:SetAttribute("client_index", client_index)
			effectEvent:SetAttribute("client", client)
			GameEventManager:SendDelayed(effectEvent, self, GetGameTime()+(i-1)*60)
		end
	end
	
	SoundSystem.Play("pq2audio/board/spell_default")
	
	if spellInfo.sound then
		SoundSystem.Play(string.format("pq2audio/sp/%s", spellInfo.sound))
	end
end

function BattleGroundView:OnEventPlayClientSpellEffect(event)
	local spellInfo = SPELLS[event:GetAttribute("spell")]
	local methodName = string.format("OnEventSpellNotifyEffect_%s", event:GetAttribute("effect"))
	
	if spellInfo[methodName] then
		spellInfo[methodName](spellInfo, event, self.grid)
	else	-- supply a default handler
		SCREENS.EffectsMenu:PlayDefaultSpellTextEffect(translate_text(event:GetAttribute("effect")))
	end
end

function BattleGroundView:OnEventItemActivationNotify(event)
	self:RemoveHint()
	
	local itemUsed = SCREENS[event:GetAttribute("client")]:GetItemFromIndex(event:GetAttribute("item"))
	--SCREENS.EffectsMenu:PlayItemActivationTextEffect(itemUsed:GetName())
    
    local client = event:GetAttribute("client")
    local client_index = self:GetIndexFromClient(client)
    local client_portrait = event:GetAttribute("portrait") .. "_A" 
	
	-- make sure any tooltips are closed before opening this
	close_custompopup_menu()
    
    local effect = GameObjectManager:ConstructLocal("BITM")
	effect:OnEventItemActivationNotify(event, client_index, client_portrait, SCREENS.EffectsMenu, client.hero)
    
    for i=1, 16 do
	    local effectEvent = GameEventManager:Construct("ItemActivationNotifyEffect")
	    effectEvent:SetAttribute("effect_index", i)
	    effectEvent:SetAttribute("widget_index", event:GetAttribute("index"))
	    effectEvent:SetAttribute("client_index", client_index)
	    effectEvent:SetAttribute("client", client)
	    GameEventManager:SendDelayed(effectEvent, self, GetGameTime()+(i-1)*60)
    end
    
    if self:IsLocalClient(client) then	--(self.activeClient) then
		SoundSystem.Play("pq2audio/board/attack_good")
	else
		SoundSystem.Play("pq2audio/board/attack_evil")
	end
end

function BattleGroundView:OnEventItemActivationFailedNotify(event)
	local reason = event:GetAttribute("reason")
	if type(reason) == "string" then
		SCREENS.EffectsMenu:PlaySpellTextEffect(reason)
	end
end

function BattleGroundView:OnEventSpellNotifyEffect(event)
	local effect = "NovaBarFilling"
	if event:GetAttribute("effect_index") == 16 then
		effect = "NovaBarFilled"
	end
	local widget = string.format("butt_spell%d", event:GetAttribute("widget_index"))
	
	self:ShowEffectOnWidget(event, effect, widget)
end

function BattleGroundView:OnEventItemActivationNotifyEffect(event)
	self.allowMoves = false
end

function BattleGroundView:ShowEffectOnWidget(event, effect, widget)
	local effect_index = event:GetAttribute("effect_index")
	local client_index = event:GetAttribute("client_index")
	local client = event:GetAttribute("client")
	if not client then
		return
	end
	
	local player_x = SCREENS[client]:GetX() - GetScreenOffset()
	local player_y = SCREENS[client]:GetY()
	local button_x = SCREENS[client]:get_widget_x(widget)
	local button_y = SCREENS[client]:get_widget_y(widget)
	local button_w = SCREENS[client]:get_widget_w(widget)
	local button_h = SCREENS[client]:get_widget_h(widget)
	
	local x = player_x + button_x			-- widget position in screen space
	x = x + (effect_index * (button_w/16))
	
	local y = player_y + button_y		-- widget position in screen space
	y = 768 - y								-- now to world space, where y goes up not down
	
	local targetObject = SCREENS.EffectsMenu:GetWorld()
	
	AttachParticles(targetObject, effect, x, y)
	AttachParticles(targetObject, effect, x, y - button_h)
end

function BattleGroundView:OnEventExplodeGem(event)
	assert(event)
	
	local x = event:GetAttribute("x")
	local y = event:GetAttribute("y")
	
	local gem = self.grid[x][y]
	
	if gem == self.selectedGem then
		self.selectedGem = nil
	end
	
	if gem then
		local particleEffects = Gems[gem.name].particleEffects
		if type(particleEffects) == "table" then
			for _, sys in ipairs(particleEffects) do
				AttachParticles(self, sys, self:GemCoordToPixel(x, y))
			end
		end
		
		self:DestroyGemAt(x, y)
	end
end

function BattleGroundView:OnEventCascadeNotify(event)
	assert(event)
	
	local left = event:GetAttribute("left")
	local right = event:GetAttribute("right")
	local top = event:GetAttribute("top")
	local bottom = event:GetAttribute("bottom")
	
	-- move all falling gems
	for i=1,event:NumAttributes("sx") do
		local sx = event:GetAttributeAt("sx", i)
		local sy = event:GetAttributeAt("sy", i)
		local tx = event:GetAttributeAt("tx", i)
		local ty = event:GetAttributeAt("ty", i)
		
		local startGemFalling = GameEventManager:Construct("StartGemFalling")
		
		startGemFalling:SetAttribute("sx", sx)
		startGemFalling:SetAttribute("sy", sy)
		startGemFalling:SetAttribute("tx", tx)
		startGemFalling:SetAttribute("ty", ty)
		startGemFalling:SetAttribute("gem", self.grid[sx][sy])
		
		local delay = ty - bottom
		
		GameEventManager:SendDelayed(startGemFalling, self, GetGameTime() + delay * DROP_DELAY + tx * DROP_XDELAY)
		
		-- the actual movement might be delayed, but the gems are moved into
		-- the correct place on the board IMMEDIATELY
		
		self:DestroyGemAt(tx, ty)
		self.grid[tx][ty] = self.grid[sx][sy]
		self.grid[tx][ty].x = tx
		self.grid[tx][ty].y = ty
		self.grid[sx][sy] = nil
	end
	
	-- create new gems
	for i=1,event:NumAttributes("newx") do
		local x = event:GetAttributeAt("newx", i)
		local y = event:GetAttributeAt("newy", i)
		local name = event:GetAttributeAt("newname", i)
		
		-- note: noattach is set to true, this is because the gems aren't supposed to appear immediately, they
		-- appear some time in the future after the gems in its place have cascaded.  It is the StartGemFalling
		-- event that adds them to the world.
		local gem = self:AddNewGemToBoard(x, y, name, true)
		
		local startGemFalling = GameEventManager:Construct("StartGemFalling")
		
		startGemFalling:SetAttribute("tx", x)
		startGemFalling:SetAttribute("ty", y)
		startGemFalling:SetAttribute("sx", x)
		startGemFalling:SetAttribute("sy", self.height+1)
		startGemFalling:SetAttribute("gem", gem)
		
		local delay = y - bottom
		
		GameEventManager:SendDelayed(startGemFalling, self, GetGameTime() + delay * DROP_DELAY + x * DROP_XDELAY)
	end
	
	-- play a sound to indicate this is a cascade
	local count = event:GetAttribute("count") - 1
	
	count = math.min(math.max(count, 1), #CASCADE_SOUNDS)
	
	SoundSystem.Play(CASCADE_SOUNDS[count])
	
	-- check heroic effort condition
	if count == HEROIC_EFFORT_COUNT then
		self:OnHeroicEffort(event)
	end
end

function BattleGroundView:OnHeroicEffort(event)
	if not self.dontShowHeroicEffort then
		self.dontShowHeroicEffort = true
		SoundSystem.Play("pq2audio/board/fx_heroic")
		SCREENS.EffectsMenu:PlayHeroicTextEffect("[HEROIC_GAINXP]", "[HEROIC_WILDCARD]", "[HEROIC_EXTRATURN]")
	end
end

function BattleGroundView:OnEventStartGemFalling(event)
	assert(event)
	
	local gem = event:GetAttribute("gem")
	
	if not gem then
		-- gem was destroyed while falling, slightly unusual!
		return
	end
	
	local sx = event:GetAttribute("sx")
	local sy = event:GetAttribute("sy")
	local tx = event:GetAttribute("tx")
	local ty = event:GetAttribute("ty")
	local sx_pix, sy_pix = self:GemCoordToPixel(sx, sy)
	local tx_pix, ty_pix = self:GemCoordToPixel(tx, ty)

	gem:SetPos(sx_pix, sy_pix)
	
	if gem:GetView():GetAlpha() < 1.0 then
		gem:GetView():SetAlpha(1.0)
	end
	
	local moveToTarget = MovementManager:Construct("Gravity")
	
	moveToTarget:SetAttribute("Accel", FALL_ACCELERATION)
	moveToTarget:SetAttribute("StartX", sx_pix)
	moveToTarget:SetAttribute("StartY", sy_pix)
	moveToTarget:SetAttribute("EndX", tx_pix)
	moveToTarget:SetAttribute("EndY", ty_pix)
	
	gem:SetMovementController(moveToTarget)
end

function BattleGroundView:OnEventStartTurn(event)
	assert(event)
	
	if not self.activeClient then
		-- first turn, make sure that the cursor is set correctly (normally
		-- it is set back to "defend" on the end turn, so it is not setup
		-- correctly initially - #3898)
		local cursor = Interface.GetCurrentCursor()
		if cursor then
			cursor:SetGraphic("curs_defend")
			cursor:SetActive(true)
		end
	end
	
	self.activeClient = event:GetAttribute("client")
	
	-- TUTORIAL TODO: Battle game begins

	if self:IsLocalClient(event:GetAttribute("client")) then
		local cursor = Interface.GetCurrentCursor()
		if cursor then
			cursor:SetGraphic("curs_attack")
			cursor:SetActive(true)
		end
	
		if (self.allowMoves) and (not self.isMoving) then
			SCREENS.BattleGameMenu:MoveToGrid()
		end
		
		if HeroManager:GetActiveHero() and self.activeClient.host and self.activeClient.host.grid then
			for x1, y1, x2, y2, matches in self.activeClient.host.grid:IterateValidMoves() do	
				self:HandleTutorials(matches, x1, y1, x2, y2)
			end
		end
		
		if _G.TUTORIALS_ON then
			for i, tutorial in ipairs(self.tutorials) do
				if (not IsOpen("IsOpen")) and tutorial.show then
					SCREENS.OppTutorialMenu:Open()
					SCREENS.OppTutorialMenu:ShowTutorial(i, tutorial.x1, tutorial.y1, tutorial.x2, tutorial.y2)
					--[[
					if tutorial.gem_1 and tutorial.gem_2 then
						self.input_gems = { tutorial.gem_1, tutorial.gem_2 }
					end
					]]--
					break
				end
			end
		end
		self.tutorials[TUTORIAL_MATCH5].show = false
		self.tutorials[TUTORIAL_MATCH4].show = false
		self.tutorials[TUTORIAL_MATCHSKULL].show = false
		self.tutorials[TUTORIAL_MATCHSKULL5].show = false
		self.tutorials[TUTORIAL_MATCHMANA].show = false
		self.tutorials[TUTORIAL_MATCHWILD].show = false
		self.tutorials[TUTORIAL_MATCHACTION].show = false
		self.tutorials.locked = false
	end
	
	if (self.localClients) and (#self.localClients > 1) then
		UserHandler.LockControlsToPlayer(event:GetAttribute("client"):GetAttribute("team"))
	end
end

function BattleGroundView:OnEventEndTurn(event)
	assert(event)
	
	SCREENS[event:GetAttribute("client")]:EndTurn()
	
	-- if we are in targetting mode, lets clear that out (#3350, #3352)
	if self.targetCallback then
		SCREENS[self.activeClient]:DisableTargettingLockdown()
	end	
	
	-- NOTE: Regarding #2628.  When a turn times out, there is still probably a gem selected
	-- and that causes issues with MoveToGrid (which, for some indeterminable reasons, resets
	-- the gem selection overlay).  
	-- 
	-- So I have moved the clearing of the currently selected gem to above the MoveToGrid ..
	-- and then I clear it again at the bottom :/  This is horrible but this is such a dodgy
	-- and fragile piece of code that I dare not attempt to fix it properly
	
	if self.selectedGem then
		self:ClearSelectedGemOverlay(self.selectedGem)
		self.selectedGem = nil
	end
	
	SCREENS.BattleGameMenu:MoveToGrid()
	
	self.allowMoves = false
	self.activeClient = nil
	
	if self:IsLocalClient(event:GetAttribute("client")) then
		local cursor = Interface.GetCurrentCursor()
		if cursor then
			cursor:SetGraphic("curs_defend")
			cursor:SetActive(true)
		end
	end

	if self.selectedGem then
		self:ClearSelectedGemOverlay(self.selectedGem)
		self.selectedGem = nil
	end
end

function BattleGroundView:OnEventStartMove(event)
	assert(event)
	
	self.allowMoves = self:IsLocalClient(self.activeClient)
	self.isMoving = false
	self.dontShowHeroicEffort = false
	
	if self:IsLocalClient(self.activeClient) then
		SCREENS.BattleGameMenu:MoveToGrid()
		self:SetupSelection()
		
		if not event:GetAttribute("bonus") then
			SoundSystem.Play("pq2audio/board/newturn")
		end
	else
		if not event:GetAttribute("bonus") then
			SoundSystem.Play("pq2audio/board/newturnx")
		end
	end
end

function BattleGroundView:OnEventSkullMatchNotify(event)
	local name = event:GetAttribute("name")
	local ignored = event:GetAttribute("ignored")		-- this is set if this gives damage and the opponent is stealthed
	local critical = event:GetAttribute("critical")
	
	local xAvg = 0
	local yAvg = 0
	for i, val in IterateCollection(event, "x") do
		xAvg = xAvg + val
	end
	for i, val in IterateCollection(event, "y") do
		yAvg = yAvg + val
	end
	xAvg = xAvg / event:NumAttributes("x")
	yAvg = yAvg / event:NumAttributes("y")
	local x, y = self:GemCoordToPixel(xAvg, yAvg)
	
	local count = event:GetAttribute("count")
	if count >= Constants.BONUS_TURN_COUNT then
		SoundSystem.Play("pq2audio/board/fx_extraturn")
		
		if count == 4 then
			SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[OFKIND_FOUR]", "[OFKIND_TAKETURN]")
		else
			SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[OFKIND_FIVE]", "[OFKIND_WILDCARD]")
		end
	end
	
	local manaGained = event:GetAttribute("score") * event:GetAttribute("bonus")

	if not ignored and manaGained > 0 then
		if critical then
			SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[CRITICAL_SKULL]")
			--manaGained = 2 * manaGained
		end
		SCREENS.EffectsMenu:PlayFallingTextEffectAt(x, y, string.format("+%d", manaGained), Gems[name].colourEffect)
	elseif ignored then
		SCREENS.EffectsMenu:PlayFallingTextEffectAt(x, y, "[STEALTH_SKULLMATCH]", Gems[name].colourEffect)
	end
end
	
function BattleGroundView:OnEventMatchNotify(event)
	local name = event:GetAttribute("name")
	
	local xAvg = 0
	local yAvg = 0
	for i, val in IterateCollection(event, "x") do
		xAvg = xAvg + val
	end
	for i, val in IterateCollection(event, "y") do
		yAvg = yAvg + val
	end
	xAvg = xAvg / event:NumAttributes("x")
	yAvg = yAvg / event:NumAttributes("y")
	local x, y = self:GemCoordToPixel(xAvg, yAvg)
	
	local xOffset = -30
	local yOffset = -20
	--For a vertical match I want it appearing offset right by ~20.
	if xAvg == event:GetAttributeAt("x", 1) then
		xOffset = xOffset + 20
	--For a horizontal match I want it appearing offset down by ~20
	elseif yAvg == event:GetAttributeAt("y", 1) then
		yOffset = yOffset - 20
	end
	
	local count = event:GetAttribute("count")
	if count >= Constants.BONUS_TURN_COUNT then
		SoundSystem.Play("pq2audio/board/fx_extraturn")
		
		if count == 4 then
			SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[OFKIND_FOUR]", "[OFKIND_TAKETURN]")
		else
			SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[OFKIND_FIVE]", "[OFKIND_WILDCARD]")
		end
	end
	
	local manaGained = event:GetAttribute("score") * event:GetAttribute("bonus")
	if manaGained > 0 then
		SCREENS.EffectsMenu:PlayFallingTextEffectAt(x + xOffset, y + yOffset, string.format("+%d", manaGained), Gems[name].colourEffect)
	end
	
	-- spawn gem fragments for the mana gained
	
	local infoMenu = SCREENS[event:GetAttribute("client")]
	
	for i=1,math.min(25, manaGained) do
		local fragment = GameObjectManager:ConstructLocal("BGMF")
		local target_widget = string.format("str_%s_value", string.lower(name))
		local target_x = infoMenu:get_widget_x(target_widget) + infoMenu:get_widget_w(target_widget) / 2 + infoMenu:GetX() - GetScreenOffset()
		local target_y = 768 - (infoMenu:get_widget_y(target_widget) + infoMenu:get_widget_h(target_widget) / 2 + infoMenu:GetY())
		local randomGem = math.random(event:NumAttributes("x"))
		local screenPt = SCREENS.BattleGameMenu:ConvertWorldCoordsToScreenCoords(self:GemCoordToPixel(event:GetAttributeAt("x", randomGem), event:GetAttributeAt("y", randomGem)))
		
		if infoMenu.stealthed then
			target_x = infoMenu:get_widget_x("icon_weapon1") + infoMenu:get_widget_w("icon_weapon1") / 2 + infoMenu:GetX() - GetScreenOffset()
			target_y = 768 - (infoMenu:get_widget_y("icon_weapon1") + infoMenu:get_widget_h("icon_weapon1") / 2 + infoMenu:GetY())
		end
		
		screenPt.x = screenPt.x - GetScreenOffset()					-- ConvertWorldCoordsToScreenCoords returns a position from the top left of the screen (MONITOR, not MENU) which is 0,0 regardless of widescreen mode
		local randOffset = self.gem_size/2
		screenPt.x = screenPt.x + math.random(-randOffset,randOffset)
		screenPt.y = screenPt.y + math.random(-randOffset,randOffset)
		
		fragment:SetInfo(Gems[name], self.gem_scale, self.gem_size)
		fragment:Fire(screenPt.x, 768 - screenPt.y, target_x, target_y, i * 150)
		
		SCREENS.EffectsMenu:GetWorld():AddChild(fragment)
	end
end

function BattleGroundView:RemoveHint()
	self:RemoveOverlay("hint1")
	self:RemoveOverlay("hint2")
end

function BattleGroundView:ShowHint()
	if (not self:IsLocalClient(self.activeClient)) then
		return	-- do not start scheduler if it is not the players turn
	end
	
	local hintdelay = Settings:Read("hintarrowdelay", 5000)
	
	if hintdelay == 0 then 
		return  -- hint arrow is disabled
	end
	
	scheduler:AddChildThread(self, function ()
		scheduler:Sleep(hintdelay)
		if self:CanShowHint() and self.hint1 and self.hint2 then
			self.hint1:SetAlpha(Constants.HINT_ARROW_ALPHA_VALUE)
			self.hint2:SetAlpha(Constants.HINT_ARROW_ALPHA_VALUE)
		end
	end, "BattleGroundView:ShowHint")
end

function BattleGroundView:HideHint()
	if self.hint1 then
		self.hint1:SetAlpha(0.0)
	end
	if self.hint2 then
		self.hint2:SetAlpha(0.0)
	end
end

function BattleGroundView:CanShowHint()
	if self.isMoving then
		return false
	end
	
	if Settings:Read("hintarrowdelay", 5000) == 0 then
		return false		-- #4108: ensure hint arrow is still turned on
	end
	
	if IsGamepadActive() then
		if not SCREENS.BattleGameMenu.selected then
			return true
		end
	else
		if not self.selectedGem then
			return true
		end
	end
	
	return false
end

function BattleGroundView:OnEventHintArrow(event)
	self:RemoveHint()
	
	local hintX1 = event:GetAttribute("x1")
	local hintY1 = event:GetAttribute("y1")
	local hintX2 = event:GetAttribute("x2")
	local hintY2 = event:GetAttribute("y2")
	local moveID = event:GetAttribute("moveID")
	LOGF("HintArrow - recieve coord values: x1=%d, y1=%d, x2=%d, y2=%d (%d)", hintX1, hintY1, hintX2, hintY2, moveID)
	
	local x1, y1 = self:GemCoordToPixel(event:GetAttribute("x1"), event:GetAttribute("y1"))
	y1 = y1 + math.round(self.gem_step / 2)
	local x2, y2 = self:GemCoordToPixel(event:GetAttribute("x2"), event:GetAttribute("y2"))
	y2 = y2 + math.round(self.gem_step / 2)
	self:AddOverlay("hint1", self.hint1, x1, y1)
	self:AddOverlay("hint2", self.hint2, x2, y2)
	
	if self:CanShowHint() then
		self:ShowHint()
	end
end

function BattleGroundView:SetupSelection()
	-- if gamepad connected then show the selection marker
	if IsGamepadActive() then
		if self.x and self.y and self.grid[self.x][self.y] then
			self.selectedGem = self.grid[self.x][self.y]
		elseif self.grid[math.floor(self.width / 2)][math.floor(self.height / 2)] then
			self.selectedGem = self.grid[math.floor(self.width / 2)][math.floor(self.height / 2)]
		else
			while not self.selectedGem do
				self.selectedGem = self.grid[math.random(1, self.width)][math.random(1, self.height)]
			end
		end
		
		self:SetSelectedGemOverlay(self.selectedGem, self.player_index, true)
	end
end

function BattleGroundView:Unselect()
	if self.selectedGem then
		self:ClearSelectedGemOverlay(self.selectedGem)
		self.selectedGem = nil
	end
end

function BattleGroundView:OnEventManaDrainNotify(event)
	SCREENS.EffectsMenu:PlayManaDrainTextEffect("[NO_MOVES]")
	SoundSystem.Play("pq2audio/board/fx_manadrain")
end

function BattleGroundView:OnEventMissMove(event)
	local reason = event:GetAttribute("reason")
	
	SCREENS.EffectsMenu:PlayMissMoveTextEffect(reason)
	
	if reason == "[STUN_EFFECT_MESSAGE]" then
		SoundSystem.Play("pq2audio/board/fx_stun")
	end
	
	if self.selectedGem then
		self:ClearSelectedGemOverlay(self.selectedGem)
		self.selectedGem = nil
	end
end

-- For now we ping the gem by using the particle effect that is played when
-- the gem is destroyed.  A better effect might be nice.
function BattleGroundView:OnEventPingGem(event)
	assert(event)
	
	local x = event:GetAttribute("x")
	local y = event:GetAttribute("y")
	
	local gem = self.grid[x][y]
	
	if gem then
		local particleEffects = Gems[gem.name].particleEffects
		if type(particleEffects) == "table" then
			for _, sys in ipairs(particleEffects) do
				AttachParticles(self, sys, self:GemCoordToPixel(x, y))
			end
		end
	end
end

function BattleGroundView:SetTargettingOnly(tableOfGemNames)
	if (not self.allowMoves) or (self.isMoving or self.pending_swap_request or not self:IsLocalClient(self.activeClient)) then
		return
	end
	
	self.targetOnly = tableOfGemNames
end

function BattleGroundView:IsInTargettingMode()
	if self.targetCallback then
		return true
	end
	
	return
end

function BattleGroundView:StartTargettingMode(callback, target_icon)
	if (not self.allowMoves) or (self.isMoving or self.pending_swap_request or not self:IsLocalClient(self.activeClient)) then
		return
	end
	
	if self.targetOnly then
		for x=1, self.width do
			for y=1, self.height do
				if self.grid[x] and self.grid[x][y] then
					self.grid[x][y]:GetView():SetAlpha(0.25)
					for _, gemName in ipairs(self.targetOnly) do
						if self.grid[x][y].name == gemName then
							self.grid[x][y]:GetView():SetAlpha(1.0)
						end
					end
				end
			end
		end
	end
	
	self.targetCallback = callback
	
	-- change cursor?
	local cursor = Interface.GetCurrentCursor()
	if cursor then
		cursor:SetGraphic("curs_select")
		cursor:SetActive(true)
	end
	
	self.targetIcon = GameObjectViewManager:Construct("GraphicAsset", "anim_target_selection")
	self.targetIcon:SetScale(self.temp_scale)
	self.targetIcon:SetSortingValue(Constants.HINT_ARROW_SORTING_VALUE)
	
	SCREENS.BattleGameMenu:MoveToGrid()
end

function BattleGroundView:EndTargettingMode(cancelled)
	if self.targetOnly then
		for x=1, self.width do
			for y=1, self.height do
				if self.grid[x] and self.grid[x][y] then
					self.grid[x][y]:GetView():SetAlpha(1.0)
				end
			end
		end
	end
	
	self.targetCallback = nil
	self.targetOnly = nil
	
	if self.selectedGem then
		self:ClearSelectedGemOverlay(self.selectedGem)	
		
		if cancelled then
			self:SetSelectedGemOverlay(self.selectedGem, self.player_id, true)
		else
			self.selectedGem = nil -- HAX? It seems to have fixed the crash...
		end
	end
	
	-- change cursor back?
	local cursor = Interface.GetCurrentCursor()
	if cursor then
		cursor:SetGraphic("curs_attack")
		cursor:SetActive(true)
	end
	
	self.targetIcon = nil
end

function BattleGroundView:GetNextDisplacement(x, y)
	assert(x)
	assert(y)
	
	if x < 0 then
		return x-1, 0
	elseif x > 0 then
		return x+1, 0
	elseif y < 0 then
		return 0, y-1
	elseif y > 0 then
		return 0, y+1
	else
		return 0, 0
	end
end

function BattleGroundView:ShowGamepadSelection(xDisp, yDisp)
	assert(xDisp)
	assert(yDisp)
	
	if (xDisp == 0) and (yDisp == 0) then
		return -- not moving anywhere - this triggers an infinite loop if left unchecked
	end
	
	if (not self.allowMoves) or (self.isMoving or self.pending_swap_request or not self:IsLocalClient(self.activeClient)) then
		return
	elseif not self.selectedGem then
		--assert(false, "Called ShowGamepadSelection on Invalid Gem")
		self:SetupSelection()
		return
--[[
		local displacement = 1
		local foundValidGem
		for x = -displacement, displacement do
			for y = -displacement, displacement do
				x = math.max(math.min(self.width,  x), 1)
				y = math.max(math.min(self.height, y), 1)
				local obj = self.grid[x][y]
				if obj and obj.IS_GEM and obj.def.canSwap then
					xDisp = x
					yDisp = y
					foundValidGem = true
					break
				end
			end
			if foundValidGem then
				break
			end
		end
]]--
	end
	
	LOGF("BattleGroundView:ShowGamepadSelection(x=%d, y=%d)", xDisp, yDisp)
	
	local x, y 
	if self.selectedGem then
		x, y = self.selectedGem.x, self.selectedGem.y
	else
		x, y = self.x, self.y
	end
	
	if x + xDisp < 1 or x + xDisp > self.width or y + yDisp < 1 or y + yDisp > self.height then
		return
	end
	
	local obj = self.grid[x + xDisp][y + yDisp]
	LOGF("BattleGroundView:ShowGamepadSelection: x=%d, y=%d, obj=%s", x + xDisp, y + yDisp, tostring(obj))
	
	if obj and obj.IS_GEM and obj.def.canSwap then
		if self.selectedGem then
			self:ClearSelectedGemOverlay(self.selectedGem)
		end
		self.selectedGem = obj
		self:SetSelectedGemOverlay(self.selectedGem, self.player_index, true)
	else
		self:ShowGamepadSelection(self:GetNextDisplacement(xDisp, yDisp))
	end
end

function BattleGroundView:DoGamepadSwap(xDisp, yDisp)
	assert(xDisp)
	assert(yDisp)
	
	if (not self.allowMoves) or (self.isMoving or self.pending_swap_request or not (self:IsLocalClient(self.activeClient)) or not self.selectedGem) then
		return
	end
	
	LOGF("BattleGroundView:DoGamepadSwap(x=%d, y=%d)", xDisp, yDisp)
	
	local x, y = self.selectedGem.x, self.selectedGem.y
	
	if x + xDisp < 1 or x + xDisp > self.width or y + yDisp < 1 or y + yDisp > self.height then
		SoundSystem.Play("pq2audio/ui/illegal")
		local sx_pix, sy_pix = self:GemCoordToPixel(x, y)
		local tx_pix, ty_pix = self:GemCoordToPixel(x + xDisp, y + yDisp)
		local ix = (sx_pix + tx_pix) / 2
		local iy = (sy_pix + ty_pix) / 2
		SCREENS.EffectsMenu:PlayFallingTextEffectAt(ix, iy, "[ILLEGAL_MOVE]")
		
		self:ShowHint()
		
		self:ClearSelectedGemOverlay(self.selectedGem)
		self.selectedGem = nil
		self.x = x
		self.y = y
		self:SetupSelection()
		
		return
	end
	
	local obj = self.grid[x + xDisp][y + yDisp]
	LOGF("BattleGroundView:DoGamepadSwap: x=%d, y=%d, obj=%s", x + xDisp, y + yDisp, tostring(obj))
	
	if obj and obj.IS_GEM and obj.def.canSwap then
		self.x = x + xDisp
		self.y = y + yDisp
		
		self.targetGem = obj
		self:SetSelectedGemOverlay(self.targetGem, self.player_index, false)
		
		self:SendSwapRequest(self.selectedGem.x, self.selectedGem.y, self.targetGem.x, self.targetGem.y)
	else
		self:SendSwapRequest(self.selectedGem.x, self.selectedGem.y, x + xDisp, y + yDisp)
	end
end

function BattleGroundView:DoGamepadTarget()
	if self.targetCallback then
		if self.selectedGem then
			local foundTarget = true
			if self.targetOnly then
				foundTarget = false
				for _, gemName in ipairs(self.targetOnly) do
					if self.selectedGem.name == gemName then
						foundTarget = true
					end
				end
			end
			if foundTarget then
				AttachParticles(self, 'RingWhite', self:GemCoordToPixel(self.selectedGem.x, self.selectedGem.y))
				self.targetCallback(self.selectedGem.x, self.selectedGem.y)
				self:EndTargettingMode()
			else
				SoundSystem.Play("pq2audio/ui/illegal")
			end
		end
	end
end

function BattleGroundView:OnEventEnterStealth(event)
	scheduler:AddChildThread(self, function ()
		scheduler:Sleep(2000)
		SCREENS.EffectsMenu:PlayRisingTextEffectAt("MIDDLE_X", "LOWER_Y", "[ENTER_STEALTH_HEADING]", "[ENTER_STEALTH_SUBTITLE]")
	end, "BattleGroundView:OnEventEnterStealth")
	SCREENS[event:GetAttribute("client")]:OnEventEnterStealth(event)
	SCREENS.BattleGameMenu:ShowStealthHalo()
end

function BattleGroundView:OnEventExitStealth(event)
	local detected = event:GetAttribute("detected")
	if detected then
		local detected_x = event:GetAttribute("detected_xpos")
		local detected_y = event:GetAttribute("detected_ypos")

		SCREENS.EffectsMenu:PlayRisingTextEffectAt(detected_x, detected_y, "[EXIT_STEALTH_HEADING]", "[EXIT_STEALTH_DETECTED]")
	end
	SCREENS[event:GetAttribute("client")]:OnEventExitStealth(event)
	
	SCREENS.BattleGameMenu:HideStealthHalo()
end

function BattleGroundView:OnEventStatusEffectIgnored(event)
	if event:GetAttribute("client"):GetAttribute("team") == 2 then
		SCREENS.EffectsMenu:PlayRisingTextEffectAt("RIGHT_X", "UPPER_Y", "[STEALTH_IGNORED]", event:GetAttribute("prettyName"), "[STEALTH_IGNORED_INFO]")
	else
		SCREENS.EffectsMenu:PlayRisingTextEffectAt("LEFT_X", "UPPER_Y", "[STEALTH_IGNORED]", event:GetAttribute("prettyName"), "[STEALTH_IGNORED_INFO]")
	end
end

function BattleGroundView:OnEventDamageManaIgnored(event)
	-- only want one of these going at a time
	if not self.damageManaIgnoredEffectActive then
		local offset = -32
		if event:GetAttribute("client"):GetAttribute("team") == 2 then
			offset = 32
		end
		SCREENS.EffectsMenu:PlayRisingTextEffectAt(256+offset, "UPPER_Y", "[STEALTH_MANA_IGNORED]", "[STEALTH_MANA_IGNORED_INFO]")
		self.damageManaIgnoredEffectActive = true
		scheduler:AddChildThread(self, function ()
			scheduler:Sleep(4000)
			self.damageManaIgnoredEffectActive = false
		end)
	end
end

function BattleGroundView:OnEventReanimationNotify(event)
	SCREENS.EffectsMenu:PlayRisingTextEffectAt("MIDDLE_X", 96, "[REANIMATION_HEADER]", "[REANIMATION_INFO]")
	
	self.reanimationCount = self.reanimationCount + 1
	
	if self.reanimationCount >= 3 then
		Achievements.AwardAchievement(1, "YOUONLYLIVETHRICE")
	end
end

function BattleGroundView:IsLocalClient(activeClient)
	if self.localClients then
		for index, client in pairs(self.localClients) do
			if client == activeClient then
				self.player_index = index
				return true
			end
		end
	else
		if self.client == activeClient then
			self.player_index = 1
			return true
		end
	end
	self.player_index = 2
	return false
end

function BattleGroundView:HandleTutorials(matches, grid_x1, grid_y1, grid_x2, grid_y2)
	LOGF("X1 = %s, Y1 = %s, X2 = %s, Y2 = %s", 
		tostring(grid_x1),
		tostring(grid_y1),
		tostring(grid_x2),
		tostring(grid_y2))


	-- convert co-ordinates from grid to screen
	if (not self.grid[grid_x1][grid_y1]) then
		return
	end
	local x1 =      self.grid[grid_x1][grid_y1]:GetX()  + SCREENS.BattleGameMenu:get_widget_abs_x("BattleGameMenuWorldIcon")
	local y1 = (SCREENS.BattleGameMenu:get_widget_h("BattleGameMenuWorldIcon") - self.grid[grid_x1][grid_y1]:GetY()) + SCREENS.BattleGameMenu:get_widget_abs_y("BattleGameMenuWorldIcon")
	local x2, y2
	if grid_x2 and grid_y2 and self.grid[grid_x2][grid_y2] then
		LOGF("X2 = %s, Y2 = %s", tostring(grid_x2), tostring(grid_y2))
		x2 =      self.grid[grid_x2][grid_y2]:GetX()  + SCREENS.BattleGameMenu:get_widget_abs_x("BattleGameMenuWorldIcon")
		y2 = (SCREENS.BattleGameMenu:get_widget_h("BattleGameMenuWorldIcon") - self.grid[grid_x2][grid_y2]:GetY()) + SCREENS.BattleGameMenu:get_widget_abs_y("BattleGameMenuWorldIcon")
	end
	
	local function SetShowTutorial(id)
		local hero = HeroManager:GetActiveHero()
		if hero and hero:GetAttributeAt("shown_tutorials", id) == 0 then
			self.tutorials[id].show = true
			self.tutorials[id].x1 = x1
			self.tutorials[id].y1 = y1
			self.tutorials[id].x2 = x2
			self.tutorials[id].y2 = y2
			self.tutorials[id].gem_1 = self.grid[grid_x1][grid_y1]
			self.tutorials[id].gem_2 = self.grid[grid_x2][grid_y2]
		end
	end
	
	for _, match in ipairs(matches) do
		if (match.count >= 5) then
			SetShowTutorial(TUTORIAL_MATCH5)
		end
		if (match.count == 4) then
			SetShowTutorial(TUTORIAL_MATCH4)
		end
		if (match.name == "Skull") or (match.name == "Skull5") then
			for _, gem_data in ipairs(match.gems) do
				if (gem_data.gem.name == "Skull") then
					SetShowTutorial(TUTORIAL_MATCHSKULL)
				elseif (gem_data.gem.name == "Skull5") then
					SetShowTutorial(TUTORIAL_MATCHSKULL5)
				end
			end
		end
		if (match.name == "Red") or
			(match.name == "Green") or
			(match.name == "Blue") or
			(match.name == "Yellow") then -- mana match
			SetShowTutorial(TUTORIAL_MATCHMANA)
			for _, gem_data in ipairs(match.gems) do
				if (string.sub(gem_data.gem.name, 1, 8) == "Wildcard") then
					SetShowTutorial(TUTORIAL_MATCHWILD)
				end
			end
		elseif (match.name == "Power") then
			SetShowTutorial(TUTORIAL_MATCHACTION)
		end
	end
end

function BattleGroundView:PlayMatchAudio(event)
	local total = event:GetAttribute("total_num")
	
	if total < 7 then
		SoundSystem.Play("pq2audio/board/gemmatch_0")
	elseif total < 13 then
		SoundSystem.Play("pq2audio/board/gemmatch_1")
	else
		SoundSystem.Play("pq2audio/board/gemmatch_2")
	end
	
	if event:GetAttribute("red_matched") then
		SoundSystem.Play("pq2audio/board/gemmana_red")
	end
	if event:GetAttribute("green_matched") then
		SoundSystem.Play("pq2audio/board/gemmana_green")
	end
	if event:GetAttribute("yellow_matched") then
		SoundSystem.Play("pq2audio/board/gemmana_yellow")
	end
	if event:GetAttribute("blue_matched") then
		SoundSystem.Play("pq2audio/board/gemmana_blue")
	end
	if event:GetAttribute("black_matched") then
		SoundSystem.Play("pq2audio/board/gemmana_black")
	end
	if event:GetAttribute("skull5_matched") then
		SoundSystem.Play("pq2audio/board/gemskull_1")
	elseif event:GetAttribute("skull_matched") then
		SoundSystem.Play("pq2audio/board/gemskull_0")
	elseif event:GetAttribute("wildcard_matched") then
		SoundSystem.Play("pq2audio/board/fx_wildcard")
	end
	
	local power_gems = event:GetAttribute("power_num")
	if power_gems > 4 then
		SoundSystem.Play("pq2audio/board/gempower_1")
	elseif power_gems > 0 then
		SoundSystem.Play("pq2audio/board/gempower_0")
	end
end

function BattleGroundView:GetGemUnderCursor()
	LOG("GetGemUnderCursor: %s %s", tostring(self.hoverGem), tostring(self.selectedGem))
	return self.hoverGem or self.selectedGem
end

function BattleGroundView:OnEventBattleStartNotify(event)
	local aPlayer = event:GetAttribute("activePlayer")
	local aPlayerLocal = aPlayer:GetAttribute("local")
	local aPlayerNPC = aPlayer:GetAttribute("npc")
	local aPlayerTeam = aPlayer:GetAttribute("team")
	
	local nPlayer = event:GetAttribute("nextPlayer")
	local nPlayerLocal = nPlayer:GetAttribute("local")
	local nPlayerNPC = nPlayer:GetAttribute("npc")
	local nPlayerTeam = nPlayer:GetAttribute("team")
	
	if (aPlayerLocal and not aPlayerNPC) and (nPlayerLocal and not nPlayerNPC) then
		SCREENS.EffectsMenu:PlayBattleStartTextEffect(string.format("[PLAYER%d_START]", aPlayerTeam))	-- local MP
	elseif aPlayerLocal and not aPlayerNPC then
		SCREENS.EffectsMenu:PlayBattleStartTextEffect("[YOU_START]")
	else
		SCREENS.EffectsMenu:PlayBattleStartTextEffect("[ENEMY_START]")
	end
end

function BattleGroundView:OnEventBattleEndNotify(event)
	local winningTeam = event:GetAttribute("winningTeam")
	
	local wPlayer = event:GetAttribute("winningPlayer")
	local wPlayerLocal = wPlayer:GetAttribute("local")
	local wPlayerNPC = wPlayer:GetAttribute("npc")
	local wPlayerTeam = wPlayer:GetAttribute("team")
	
	local lPlayer = event:GetAttribute("losingPlayer")
	local lPlayerLocal = lPlayer:GetAttribute("local")
	local lPlayerNPC = lPlayer:GetAttribute("npc")
	local lPlayerTeam = lPlayer:GetAttribute("team")
	
	if (wPlayerLocal and not wPlayerNPC) and (lPlayerLocal and not lPlayerNPC) then
		SCREENS.EffectsMenu:PlayGameWonTextEffect(string.format("[PLAYER%d_WON]", winningTeam))	-- local MP
	elseif wPlayerLocal and not wPlayerNPC then
		SCREENS.EffectsMenu:PlayGameWonTextEffect("[GAME_WON]")
	else
		SCREENS.EffectsMenu:PlayGameOverTextEffect("[GAME_OVER]")
	end
end

return ExportClass("BattleGroundView")
