--[[ 
HeroState :
	Retains an invariant snapshot of base and derived stats, that is used for implementing battle game actions and determining
	status effect results.
--]]

use_safeglobals()

class "HeroState" (GameObject)

HeroState.AttributeDescriptions = AttributeDescriptionList()

local MAX_SPELLS = 5

local validPools = {
	red = true,
	blue = true,
	green = true,
	yellow = true,
	black = true,
	power = true,
}

local poolTranslation = {
	red = "red",
	blue = "blue",
	green = "green",
	yellow = "yellow",
	black = "black",
	power = "power",

	Red = "red",
	Blue = "blue",
	Green = "green",
	Yellow = "yellow",
	Black = "black",
	Power = "power",
}

local pools = {
	"red", 
	"green", 
	"blue", 
	"yellow",
	"black",
	"power",
}

local uniqueIndex = 0

local function GetUniqueID()
	uniqueIndex = uniqueIndex + 1
	return uniqueIndex
end

local function IsValidPool(x)
	return validPools[x]
end

function HeroState:__init()
	super("BGHS")
end

-- Attribute initialization
function HeroState:Initialise(client, hero, rules)
	assert(client)
	assert(hero)
	assert(rules)

	self.client = client
	self.hero = hero
	
	self.gold = 0
	self.xp = 0
	
	-- A snapshot of the invariant state of the hero, factoring in all class, level, equipment components,
	-- we use these values as our base when calculating the results of status effects
	self.hero_invariant_stats	= self.hero:GetInvariantStats()
	
	-- [SC] TODO: ensure we arent unnecessarily duplicating data here, we only need to clone values if the 
	-- they represent a resource or quantity that fluctuates over the course of the battle (health, mana)
	self.health					= self.hero_invariant_stats.derived.health.maximum - self.hero_invariant_stats.derived.health.wounds
	self.max_health				= self.health
	
	self.red					= self.hero_invariant_stats.derived.mana.start.red
	self.green					= self.hero_invariant_stats.derived.mana.start.green
	self.blue					= self.hero_invariant_stats.derived.mana.start.blue
	self.yellow					= self.hero_invariant_stats.derived.mana.start.yellow
	self.black					= self.hero_invariant_stats.derived.mana.start.black
	self.power					= self.hero_invariant_stats.derived.mana.start.power
	
	self.actual_max_red			= self.hero_invariant_stats.derived.mana.maximum.red
	self.actual_max_green		= self.hero_invariant_stats.derived.mana.maximum.green
	self.actual_max_blue		= self.hero_invariant_stats.derived.mana.maximum.blue
	self.actual_max_yellow		= self.hero_invariant_stats.derived.mana.maximum.yellow
	self.actual_max_black		= self.hero_invariant_stats.derived.mana.maximum.black
	
	self.immunities = {}
	
	-- Statistics for umm... stuff (not to be confused with base stats and derived stats)
	self.stats = {}
	self.stats.best_stealth_damage = 0
	self.stats.total_stealth_damage = 0
	self.stats.total_stealth_damage_absorbed = 0
	self.stats.best_skull_damage = 0
	self.stats.total_skull_damage = 0
	self.stats.best_blocked_damage = 0
	self.stats.total_blocked_damage = 0
	self.stats.total_mana_used = 0
	self.stats.num_item_activations = 0
	self.stats.total_damage_taken = 0
	self.stats.total_damage_given = 0
	self.stats.best_attack_rating = 0
	self.stats.best_mana_gain = 0
	self.stats.num_bonus_moves = 0
	self.stats.num_bonus_moves_this_turn = 0
	self.stats.best_bonus_moves_streak = 0
	self.stats.num_turns = 0
	self.stats.spell_use = {}
	self.stats.spell_use[1] = 0
	self.stats.spell_use[2] = 0
	self.stats.spell_use[3] = 0
	self.stats.spell_use[4] = 0
	self.stats.spell_use[5] = 0
	
	self.spells = hero:GetSpells()
	self.spell_cost_modifier = 100		-- 100 is 100% (base spell value)
	self.damage = 0
	self.frenzyEndTime = nil
	self.statusEffects = {}
	
	-- Cap our pools to our maximums (in case we have some crazy bonuses to starting mana)
	for _, pool in pairs(pools) do
		local base_max_pool = string.format("base_max_%s", pool)
		if self[base_max_pool] and self[base_max_pool] < self[pool] then 
			self[pool] = self[base_max_pool]
		end
	end
	
	if self.host.gameOptions.replaceSpell then
		local data = self.host.gameOptions.replaceSpell[self.hero]
		if data then
			for i=MAX_SPELLS, 1, -1 do
				local spell = data[string.format("replace%d", i)]
				if spell then
					if spell == "None" then
						self.spells[i] = nil
					else
						self.spells[i] = spell
					end
				end
			end
		end
	end
	
	-- add any traits from spells
	for i, spell in pairs(self.spells) do
		local spellInfo = SPELLS[spell]
		
		if spellInfo.givesTrait then
			spellInfo:ApplyTrait(self, i)
		end
	end
	
	local initmsg = GameEventManager:Construct("ClientInfo")
	
	LOG("Sending client info package about: " .. tostring(client))
	
	initmsg:SetAttribute("client", client)
	initmsg:SetAttribute("portrait", hero:GetAttribute("portrait"))
	initmsg:SetAttribute("weapon_bonus_damage", self.hero_invariant_stats.derived.weapon_bonus)
	initmsg:SetAttribute("shield_bonus_defence", 0) -- [SC] TODO: remove, deprecated
	
	for i=1, MAX_SPELLS do
		if self.spells[i] then
			initmsg:PushAttribute("spells", self.spells[i])
		end
	end
	
	GameEventManager:Send(initmsg)
	
	-- Send information about each item
	self.mainHandItem = hero:GetAttribute("mainHandItem")
	self.offHandItem = hero:GetAttribute("offHandItem")
	
	if self.host.gameOptions.replaceItem then
		local data = self.host.gameOptions.replaceItem[self.hero]
		if data then
			local mainHand = data.replaceMainHand
			if mainHand then
				if mainHand == "Empty" then
					self.mainHandItem = nil
				else
					self.mainHandItem = CreateUniqueItem(mainHand, self.hero:GetAttribute("level"))
					self:AddChild(self.mainHandItem)
				end
			end
			
			local offHand = data.replaceOffHand
			if offHand then
				if offHand == "Empty" then
					self.offHandItem = nil
				else
					self.offHandItem = CreateUniqueItem(offHand, self.hero:GetAttribute("level"))
					self:AddChild(self.offHandItem)
				end
			end
		end
	end
	
	if self.mainHandItem then
		self:BroadcastItem(self.mainHandItem, self.client, Items.MAIN_HAND)
	end
	
	if self.offHandItem then
		self:BroadcastItem(self.offHandItem, self.client, Items.OFF_HAND)
	end
	
	for i, item in pairs(hero:GetPassiveItemsInOrder()) do
		self:BroadcastItem(item, self.client, i)
	end
	
	hero:ApplyTraitsToHeroState(self, client, rules)

	-- add any traits from the hero
	for i, name in IterateCollection(hero, "traits") do
		self:AddHeroTrait(client, name)
	end
end

function HeroState:OnBonusTurn()
	self.stats.num_bonus_moves = self.stats.num_bonus_moves + 1
	self.stats.num_bonus_moves_this_turn = self.stats.num_bonus_moves_this_turn + 1
end

function HeroState:IsImmuneToLesserStun()
	return self.immunities.LesserStun and self.immunities.LesserStun > 0
end

function HeroState:IsImmuneToStun()
	return self.immunities.Stun and self.immunities.Stun > 0
end

function HeroState:SetImmuneToLesserStun(immunity)
	self:SetImmunity("LesserStun", immunity)
end

function HeroState:SetImmuneToStun(immunity)
	self:SetImmunity("Stun", immunity)
end

function HeroState:SetImmunity(kind, immunity)
	LOGF("HeroState:SetImmunity(%s, %s)", tostring(kind), tostring(immunity))
	if immunity then
		self.immunities[kind] = (self.immunities[kind] or 0) + 1
	else
		self.immunities[kind] = self.immunities[kind] - 1
	end
end

function HeroState:OnWeaponUsed(weapon, damage)
	for status in pairs(self.statusEffects) do
		if status.OnWeaponUsed then
			status:OnWeaponUsed(weapon, damage)
		end
	end
end

function HeroState:OnItemActivated(item)
	if not item:IsUnavoidableToUse() then
		self.stats.num_item_activations = self.stats.num_item_activations + 1
	end
end

function HeroState:BroadcastItem(item, client, index)
	local itemInfoEvent = GameEventManager:Construct("ItemInfo")
	
	itemInfoEvent:SetAttribute("client", client)
	itemInfoEvent:SetAttribute("slot", index)
	itemInfoEvent:SetAttribute("baseType", item:GetAttribute("baseType"))
	itemInfoEvent:SetAttribute("race", item:GetAttribute("race"))
	itemInfoEvent:SetAttribute("rarity", item:GetAttribute("rarity"))
	itemInfoEvent:SetAttribute("unique", item:GetAttribute("unique"))
	
	GameEventManager:Send(itemInfoEvent)
end

function HeroState:IsDead()
	return self.dead
end

function HeroState:IsFrenzied()
	return self.frenzyEndTime and (GetGameTime() < self.frenzyEndTime)
end

function HeroState:SetFrenzied(seconds, callback)
	assert(type(seconds) == "number")
	
	self.frenzyEndTime = GetGameTime() + seconds * 1000
	self.frenzyCallback = callback
	self.host:AddTurnTimerToTurn(seconds * 1000, true)
end

function HeroState:ResetFrenzied(turnCount)
	local callback = self.frenzyCallback
	
	self.frenzyCallback = nil
	self.frenzyEndTime = nil
	
	return callback and callback(self, turnCount) 
end

function HeroState:IsStealthed()
	return self.stealthMode
end

function HeroState:GetStealthModeItemCostModifier(itemObject)
	if self.stealthMode and (not itemObject:IsStealthy()) then
		return self.stealthMode.itemCostModifier / 100
	else
		return 1
	end
end

function HeroState:GetStealthModeDamageBonus()
	if self.stealthMode then
		return self.stealthMode.damageBonus
	else
		return 0
	end
end

function HeroState:EnterStealthMode(params)
	assert(self.stealthMode == nil, "already in stealth mode")
	assert(params)
	assert(params.matchBonus)
	self.stealthMode = params
	self.stealthMode.damageBonus = 0
	self.stealthMode.turnNumber = 0
	self.stealthMode.itemCostModifier = 0
	self.stealthMode.manaMutateStatusEffect = self:AddTrait(self.client, "ManaMutate", "[MANAMUTATE_STEALTH]", math.huge, "Stealth", 0, 1)
	
	local enterEvent = GameEventManager:Construct("EnterStealth")
	enterEvent:SetAttribute("client", self.client)
	enterEvent:SetAttribute("itemCostModifier", self.stealthMode.itemCostModifier)
	GameEventManager:Send(enterEvent)
end

function HeroState:ExitStealthMode(detected_x, detected_y)
	assert(self.stealthMode, "not in stealth mode")
	local detected = false
	if detected_x or detected_y then
		detected = true
	end

	-- cancel any PressurePoint spells added to the other player
	self.host.nextClient.state:RemoveStatusEffectsFrom("PressurePoints")

	self:RemoveStatusEffect(self.stealthMode.manaMutateStatusEffect)
	self.stealthMode = nil

	local enterEvent = GameEventManager:Construct("ExitStealth")
	enterEvent:SetAttribute("client", self.client)
	enterEvent:SetAttribute("detected_xpos", detected_x or 0)
	enterEvent:SetAttribute("detected_ypos", detected_y or 0)
	enterEvent:SetAttribute("detected", detected)
	GameEventManager:Send(enterEvent)
end

function HeroState:OnStartTurn()
	if self.stealthMode then
		self.stealthMode.turnNumber = self.stealthMode.turnNumber + 1
	end
	self.stats.num_bonus_moves_this_turn = 0
	self.stats.num_turns = self.stats.num_turns + 1
end

local function GetAverageGridPos(gems)
	local xAvg = 0
	local yAvg = 0
	for i, gem in ipairs(gems) do
		xAvg = xAvg + gem.x
		yAvg = yAvg + gem.y
	end
	xAvg = xAvg / #gems
	yAvg = yAvg / #gems
	
	return xAvg, yAvg
end

-- HeroState: Queries
-- Each query returns the status effect modified value and the original hero invariant value

-- HeroState: Health - modifications cached in local
function HeroState:GetHealth()
	return self.health, self.hero_invariant_stats.derived.health.maximum - self.hero_invariant_stats.derived.health.wounds
end

-- HeroState: MaxHealth - no modifications
function HeroState:GetMaxHealth()
	return self.hero_invariant_stats.derived.health.maximum - self.hero_invariant_stats.derived.health.wounds, self.hero_invariant_stats.derived.health.maximum
end

-- HeroState: Mana - modifications cached in local
function HeroState:GetManaAmount(pool)
	local translated_pool = poolTranslation[pool]
	return assert( self[translated_pool] ), self.hero_invariant_stats.derived.mana.maximum[translated_pool]
end

-- HeroState: MaxMana - modified by "GetPoolManaMax" status effects
function HeroState:GetMaxMana(pool)
	local translated_pool = poolTranslation[pool]
	local invariant_max = self.hero_invariant_stats.derived.mana.maximum[translated_pool]
	local mutator = string.format( "Get%s%sManaMax", string.upper( string.sub( translated_pool, 1, 1 ) ), string.sub( translated_pool, 2 ) )
	return MutateValueWithStatusEffects( mutator, self.statusEffects, invariant_max ), invariant_max
end

-- HeroState: Attack (skull bonus damage) - modified by "GetAttackRating" status effects
function HeroState:GetAttack()
	local invariant_attack = self.hero_invariant_stats.derived.match_bonus.skull
	return math.max( 0, MutateValueWithStatusEffects( "GetAttackRating", self.statusEffects, invariant_attack ) ), invariant_attack
end

-- HeroState: Defence (% chance to half received damage) - modified by "GetDefenseRating" status effects
function HeroState:GetDefence()
	local invariant_defence = self.hero_invariant_stats.derived.defence_rating
	return math.max( 0, MutateValueWithStatusEffects( "GetDefenseRating", self.statusEffects, invariant_defence ) ), invariant_defence
end

-- HeroState: WeaponBonusDamage - modified by "GetDamageBonus" status effects
function HeroState:GetWeaponBonusDamage(weapon_base_damage)
	local invariant_damage = self.hero_invariant_stats.derived.weapon_bonus
	return MutateValueWithStatusEffects( "GetDamageBonus", self.statusEffects, invariant_damage, weapon_base_damage ), invariant_damage
end

-- HeroState: WeaponCritical - modified by "GetWeaponCritical" status effects
function HeroState:GetWeaponCriticalChance()
	local invariant_critical = self.hero_invariant_stats.derived.weapon_critical
	return MutateValueWithStatusEffects( "GetWeaponCritical", self.statusEffects, invariant_critical ), invariant_critical
end

function HeroState:IsWeaponCritical()
	return math.random(100) <= self:GetWeaponCriticalChance()
end

-- HeroState: ShieldCritical - modified by "GetShieldCritical" status effects
function HeroState:GetShieldCriticalChance()
	local invariant_critical = self.hero_invariant_stats.derived.shield_critical
	return MutateValueWithStatusEffects( "GetShieldCritical", self.statusEffects, invariant_critical ), invariant_critical
end

function HeroState:IsShieldCritical()
	return math.random(100) <= self:GetShieldCriticalChance()
end

-- HeroState: SpellPenetration - modified by "GetSpellPenetration" status effects
function HeroState:GetSpellPenetration()
	local invariant_penetration = self.hero_invariant_stats.derived.spell_penetration
	return MutateValueWithStatusEffects( "GetSpellPenetration", self.statusEffects, invariant_penetration ), invariant_penetration
end

-- HeroState: SpellResist - modified by "GetSpellResistance" status effects
function HeroState:GetSpellResistChance()
	local invariant_resist = self.hero_invariant_stats.derived.spell_resist
	return MutateValueWithStatusEffects( "GetSpellResistance", self.statusEffects, invariant_resist ), invariant_resist
end

function HeroState:IsSpellResisted(spellPenetration)
	--local myRoll = math.random(100)
	--local myResistance = self:GetSpellResistChance()
	--LOGF("HeroState:IsSpellResisted roll=%d resistance=%d",myRoll,myResistance)
	--return myRoll <= myResistance
	return math.random(100) <= self:GetSpellResistChance() - spellPenetration
end

-- HeroState: BlockingEffectiveness - modified by "GetBlockingEffectiveness" status effects
function HeroState:GetBlockingEffectiveness()
	local invariant_blocking = self.hero_invariant_stats.derived.blocking_effectiveness
	return math.min(0.75, MutateValueWithStatusEffects( "GetBlockingEffectiveness", self.statusEffects, invariant_blocking )), invariant_blocking
end

function HeroState:MutateMatch(match, opposingState)
	assert(match)
	assert(opposingState)
	
	if match.count > 2 then
		-- we only count match3+ matches, not individual destruction of gems, otherwise we get massive stacking of bonuses
		-- from spells/effects that destroy a lot of gems individually
		
		local match_name = ""
		
		if Gems[match.name].givesManaProxy then
			match_name = string.lower( Gems[match.name].givesManaProxy )
		else
			match_name = string.lower( match.name )
		end
		
		local additive_bonus = 0
		
		if match_name == "skull" or match_name == "skull5" then
			-- special case for attack rating
			additive_bonus = self:GetAttack()
		else
			additive_bonus = self.hero_invariant_stats.derived.match_bonus[match_name] or 0
		end
		
		print(string.format("HeroState:MutateMatch - name = %s score = %s  addition = %s", tostring(match_name), tostring(match.score), tostring(additive_bonus)))
		
		match.score = match.score + additive_bonus
	end
	
	for effect in pairs(self.statusEffects) do
		effect:MutateMatch(match, true)
	end
	
	if self.stealthMode then
		-- matches buff up your stealth bonus!
		self:ApplyStealthModeDamageBonus(1, self.stealthMode.matchBonus)
	end
end

function HeroState:MutateOpponentMatch(match, opposingState)
	assert(match)
	assert(opposingState)
	
	for effect in pairs(self.statusEffects) do
		effect:MutateOpponentMatch(match, false)
	end
end

function HeroState:ApplyStealthModeDamageBonus(multiplier, bonus)
	assert(multiplier)
	assert(bonus)
	self.stealthMode.damageBonus = self.stealthMode.damageBonus * multiplier + bonus
	local updateEvent = GameEventManager:Construct("ChangeStealthBonus")
	updateEvent:SetAttribute("client", self.client)
	updateEvent:SetAttribute("bonus", self.stealthMode.damageBonus)
	GameEventManager:Send(updateEvent)
end

function HeroState:GetActualRedManaMax()
	return MutateValueWithStatusEffects("GetRedManaMax", self.statusEffects, self.hero_invariant_stats.derived.mana.maximum.red)
end

function HeroState:GetActualGreenManaMax()
	return MutateValueWithStatusEffects("GetGreenManaMax", self.statusEffects, self.hero_invariant_stats.derived.mana.maximum.green)
end

function HeroState:GetActualBlueManaMax()
	return MutateValueWithStatusEffects("GetBlueManaMax", self.statusEffects, self.hero_invariant_stats.derived.mana.maximum.blue)
end

function HeroState:GetActualYellowManaMax()
	return MutateValueWithStatusEffects("GetYellowManaMax", self.statusEffects, self.hero_invariant_stats.derived.mana.maximum.yellow)
end

function HeroState:GetActualBlackManaMax()
	return MutateValueWithStatusEffects("GetBlackManaMax", self.statusEffects, self.hero_invariant_stats.derived.mana.maximum.black)
end

function HeroState:ApplyMagicalItemEffects(when, params, ...)
	local mainHandItem = self.mainHandItem
	local offHandItem = self.offHandItem
	
	if mainHandItem then
		mainHandItem:ApplyMagicalEffects(when, params, ...)
	end
	
	if offHandItem then
		offHandItem:ApplyMagicalEffects(when, params, ...)
	end
	
	for i, item in ipairs(self.hero:GetPassiveItems()) do
		item:ApplyMagicalEffects(when, params, ...)
	end
end

function HeroState:AddHeroTrait(client, name)
	require(string.format("Traits.%s", name)):Apply(self, client)
end

function HeroState:AddHiddenTrait(...)
	return self:AddStatusEffectInternal(true, false, ...)
end

function HeroState:AddStatusEffect(...)
	return self:AddStatusEffectInternal(false, true, ...)
end

function HeroState:AddTrait(...)
	return self:AddStatusEffectInternal(true, true, ...)
end

function HeroState:AddStatusEffectInternal(trait, visible, from, name, prettyName, turns, reason, ...)
	assert(from)
	assert(name)
	assert(prettyName)
	assert(turns)
	assert(type(reason) == "string", string.format("reason parameter should be a string, it is a %s", type(reason)))
	
	-- check for immunities
	local effectDef = STATUS[name]
	
	if effectDef.immunities then
		for _, immunity in ipairs(effectDef.immunities) do
			if self.immunities[immunity] and self.immunities[immunity] > 0 then
				-- player is immune to this status effect, forget about it
				return
			end
		end
	end
	
	-- check for a similar status effect from the same source
	local existingEffect = self:FindStatusEffectFromSource(name, prettyName)
	
	if existingEffect then
		if (turns == math.huge) or (existingEffect.turns and turns >= existingEffect.turns) then
			self:RemoveStatusEffect(existingEffect)
		else
			LOGF("%s:AddStatusEffect(%s,%s,%s,%s) - IGNORED", tostring(self), tostring(from), name, tostring(turns), tostring(reason))
			return -- the existing status effect has more turns left, keep it
		end
	end
	
	LOGF("%s:AddStatusEffect(%s,%s,%s,%s)", tostring(self), tostring(from), name, tostring(turns), tostring(reason))

	local effect = CreateStatusEffect(name)
	
	effect.enemy = self.host:GetOpposingClient(self.client).state
	effect.host = self.host
	effect.target = self
	effect.source = from
	if turns ~= math.huge then
		effect.turns = turns
	end
	effect.name = name
	effect.prettyName = prettyName
	effect.reason = reason
	effect.unique = GetUniqueID()
	effect.trait = trait
	effect.visible = visible
	effect:Apply(...)
	
	if effect.OnAdd then
		effect.OnAdd(effect)
	end
	
	self.statusEffects[effect] = effect
	
	-- custom callback per status effect
	local method = "OnAddStatusEffect" .. name
	
	if self[method] then
		self[method](self, effect)
	end
	
	-- notify views
	if effect.visible then
		local newstatus = GameEventManager:Construct("AddStatusEffect")
		
		if turns ~= math.huge then
			newstatus:SetAttribute("turns", turns)
		end
		newstatus:SetAttribute("client", self.client)
		newstatus:SetAttribute("name", name)
		newstatus:SetAttribute("prettyName", prettyName)
		newstatus:SetAttribute("reason", reason)
		newstatus:SetAttribute("id", effect.unique)
		
		if effect.SaveParams then
			effect:SaveParams(newstatus)
		end
		
		GameEventManager:Send(newstatus)
	end
	
	self:Broadcast()
	
	return effect
end

function HeroState:FindStatusEffect(name, reason)
	for _, effect in pairs(self.statusEffects) do
		if effect.name == name and effect.reason == reason then
			return true
		end
	end
	return false
end

function HeroState:FindStatusEffectFromSource(statusEffectName, source)
	for _, effect in pairs(self.statusEffects) do
		if effect.name == statusEffectName and effect.prettyName == source then
			return effect
		end
	end
end

function HeroState:FindStatusEffectType(name)
	for _, effect in pairs(self.statusEffects) do
		if effect.name == name then
			return true
		end
	end
	return false
end

function HeroState:GetBeneficialStatusEffects()
	local effects = {}
	for effect in pairs(self.statusEffects) do
		if not effect.trait and effect:IsBeneficial() then
			table.insert(effects, effect)
		end
	end
	return effects
end

function HeroState:GetNegativeStatusEffects()
	local effects = {}
	for effect in pairs(self.statusEffects) do
		if not effect.trait and not effect:IsBeneficial() then
			table.insert(effects, effect)
		end
	end
	return effects
end

function HeroState:GetAllStatusEffects()
	local effects = {}
	for effect in pairs(self.statusEffects) do
		if not effect.trait then
			table.insert(effects, effect)
		end
	end
	return effects
end

function HeroState:RemoveAllStatusEffectsOfName(name)
	local statusEffects = table.clone(self.statusEffects)
	
	for effect in pairs(self.statusEffects) do
		if effect.name == name and not effect.trait then
			self:RemoveStatusEffect(effect)
		end
	end
end

function HeroState:RemoveAllStatusEffects()
	local statusEffects = table.clone(self.statusEffects)
	
	for effect in pairs(self.statusEffects) do
		if not effect.trait then
			self:RemoveStatusEffect(effect)
		end
	end
end

function HeroState:RemoveAllNegativeStatusEffects()
	local statusEffects = self:GetNegativeStatusEffects()
	
	for effect in pairs(self.statusEffects) do
		if not effect.trait and not effect:IsBeneficial() then
			self:RemoveStatusEffect(effect)
		end
	end
end


function HeroState:RemoveStatusEffectsFrom(source)
	local statusEffects = table.clone(self.statusEffects)
	
	for effect in pairs(self.statusEffects) do
		if effect.reason == source then
			self:RemoveStatusEffect(effect)
		end
	end
end

function HeroState:RemoveStatusEffect(effect)
	assert(effect)
	assert(self.statusEffects[effect])
	
	LOGF("%s:RemoveStatusEffect(%s,%s)", tostring(self), tostring(effect.prettyName), tostring(effect.source))
	
	self.statusEffects[effect] = nil
	
	-- custom callback per status effect
	local method = "OnRemoveStatusEffect" .. effect.name
	
	if self[method] then
		self[method](self, effect)
	end
	
	if effect.OnRemove then
		effect:OnRemove(self)
	end

	-- notify views
	if effect.visible then
		local deletestatus = GameEventManager:Construct("RemoveStatusEffect")
		
		deletestatus:SetAttribute("client", self.client)
		deletestatus:SetAttribute("name", effect.name)
		deletestatus:SetAttribute("id", effect.unique)
		
		GameEventManager:Send(deletestatus)
	end
	
	self:Broadcast()
end

function HeroState:UpdateStatusEffects()
	-- iterate over a copy because a status effect can add/remove another status
	-- effect while updating
	local statusEffects = table.clone(self.statusEffects)
	
	for effect in pairs(statusEffects) do
		if effect.turns and not effect.trait then
			effect.turns = effect.turns - 1
			
			if effect.turns <= 0 then
				self:RemoveStatusEffect(effect)
			end
		end
	end
end

function HeroState:OnEventDamageMana(event)
	local amount = event:GetAttribute("amount")
	local pool = event:GetAttribute("pool")
	self:DamageManaImmediately(pool, amount)
end

function HeroState:OnEventRestoreMana(event)
	local amount = event:GetAttribute("amount")
	local pool = event:GetAttribute("pool")
	local pool_max = "actual_max_" .. pool
	
	assert(IsValidPool(pool), string.format("RestoreMana event is restoring to an unknown pool: %s", pool))
	
	self[pool] = self[pool] + amount
	
	if self[pool_max] and self[pool] > self[pool_max] then	-- Power has no max at the moment, will this be fixed?
		self[pool] = self[pool_max]
	elseif self[pool] < 0 then
		self[pool] = 0
	end
	
	self:Broadcast()
end

function HeroState:GetAbsorbedDamage(damage)
	return MutateValueWithStatusEffects("GetAbsorbedDamage", self.statusEffects, 0, assert(damage))
end

function HeroState:Kill(reason, info)
	self.health = 0
	self.dead = true
	
	self:Broadcast()
end

function HeroState:DoDeathCheck()
	local result, hitpoints
	
	for statusEffect in pairs(self.statusEffects) do
		result, hitpoints = statusEffect:HandleDeath()
		if result then
			break
		end
	end
	
	if result == "reanimate" then 
		self.health = hitpoints
		
		local notify = GameEventManager:Construct("ReanimationNotify")
		
		notify:SetAttribute("client", self.client)
		notify:SetAttribute("amount", hitpoints)
		
		GameEventManager:Send(notify)
	else
		self.dead = true
	end
	
	return self.health
end

function HeroState:OnEventDamageHealth(damageHealthEvent)
	local amount = damageHealthEvent:GetAttribute("amount")
	local immediate = damageHealthEvent:GetAttribute("immediate")
	
	if immediate == "" then
		-- NOTE: damage is queued up and delivered all at once at the end of a turn, this
		-- way the "block" calculation is performed once per turn.
		local event = GameEventManager:Construct("AccumulateDamageNotify")
		
		event:SetAttribute("client", self.client)
		event:SetAttribute("amount", amount)
		event:SetAttribute("stealthed", toboolean(self:IsStealthed()))
		
		GameEventManager:Send(event)
		
		self.damage = self.damage + amount
	else
		local absorbed = self:GetAbsorbedDamage(amount)
	
		if absorbed > 0 then
			self.stats.total_stealth_damage_absorbed = self.stats.total_stealth_damage_absorbed + absorbed
			-- some kind of notify
		end
		
		local event = GameEventManager:Construct("ImmediateDamageNotify")
		
		event:SetAttribute("client", self.client)
		event:SetAttribute("source", damageHealthEvent:GetAttribute("immediate"))
		event:SetAttribute("amount", amount)
		event:SetAttribute("absorbed", absorbed)
		
		GameEventManager:Send(event)
	
		-- damage is applied immediately
		self.health = self.health - (amount - absorbed)
		self.health = math.max(0, self.health)
		self.damage = 0
		
		if self.health <= 0 then
			self.health = self:DoDeathCheck()
		end

		self:Broadcast()
	end
		
	-- Some status effects apply effects when damage is recieved
	table.foreach(self.statusEffects, function (f) f:HandleDamageRecieved() end)
end

function HeroState:ApplyDamagePreProcessing()
	if self.damage > 0 then
		local original_amount = self.damage
		local adjusted_amount = self.damage
		local blocked_amount = 0
		local stealthed = self:IsStealthed()
		
		local stats = self.host:GetOpposingClient(self.client).state.stats
	
		stats.total_skull_damage = stats.total_skull_damage + original_amount
		stats.best_skull_damage = math.max(stats.best_skull_damage, original_amount)
		
		
		if math.random(100) <= self:GetDefence() then
			--print( string.format( "HeroState:ApplyDamagePreProcessing - Apply Blocking( damage = %d, blocking = %f )", original_amount, self:GetBlockingEffectiveness() ) )
			
			-- we always block atleast 1 damage on successful defence
			blocked_amount = math.max( 1, math.ceil( original_amount * self:GetBlockingEffectiveness() ) )
			adjusted_amount = original_amount - blocked_amount
			
			-- some status effects do damage when the player blocks damage
			for statusEffect in pairs(self.statusEffects) do
				statusEffect:HandleBlock(blocked_amount)
			end
			
			-- update stats
			self.stats.total_blocked_damage = self.stats.total_blocked_damage + blocked_amount
			self.stats.best_blocked_damage = math.max(self.stats.best_blocked_damage, blocked_amount)
		end
		
		LOGF("NICK: total damage: %d, adjusted damage: %d, blocked: %d", original_amount, adjusted_amount, blocked_amount)
		
		local event = GameEventManager:Construct("ApplyAccumulateDamageNotify")
		
		event:SetAttribute("client", self.client)
		event:SetAttribute("opponent", self.host:GetOpposingClient(self.client))
		event:SetAttribute("amount", original_amount)
		event:SetAttribute("blocked", blocked_amount)
		event:SetAttribute("stealthed", toboolean(stealthed))
		
		GameEventManager:Send(event)

		if stealthed then
			self.damage = 0
			
			-- in stealth mode, we sent the notify, but no actual damage gets applied
			-- because a player in stealth takes no damage.
			return 0, 0
		end
		
		return original_amount, adjusted_amount
	end
	
	return 0, 0
end

function HeroState:ApplyDamagePostProcessing(original_amount, adjusted_amount)
	assert(original_amount)
	assert(adjusted_amount)

	if self.damage > 0 then
		local absorbed = self:GetAbsorbedDamage(adjusted_amount)
	
		if absorbed > 0 then
			-- some kind of notify
			self.stats.total_stealth_damage_absorbed = self.stats.total_stealth_damage_absorbed + absorbed
			LOGF("NICKM: total damage %d, damage absorbed %d", adjusted_amount, absorbed)
		end
		
		if absorbed > adjusted_amount then
			absorbed = adjusted_amount
		end
		
		if absorbed > 0 then	-- [NEOBEN]
			local event = GameEventManager:Construct("AbsorbedDamageNotify")
			
			event:SetAttribute("client", self.client)
			event:SetAttribute("amount", absorbed)
			event:SetAttribute("stealthed", toboolean(self:IsStealthed()))
			
			GameEventManager:Send(event)
		end
		
		local final_damage = (adjusted_amount - absorbed)
		
		self.stats.total_damage_taken = self.stats.total_damage_taken + final_damage
		
		-- Some status effects can reflect damage
		local reflected_damage = MutateValueWithStatusEffects("GetReflectedDamage", self.statusEffects, 0, final_damage)
		
		LOGF("reflect_damage is %s", tostring(reflected_damage))
		
		if reflected_damage > 0 then
			-- Ensure that we do not reflect more damage than was inflicted, otherwise
			-- we have bugs like #1721.
			
			if reflected_damage > final_damage then
				reflected_damage = final_damage
			end
			
			self.host:GetOpposingClient(self.client).state:DamageHealth(self, reflected_damage, "[REFLECTDAMAGE_NOTIFY]")
		end
		
		self.health = self.health - final_damage
		self.health = math.max(0, self.health)
		self.damage = 0

		if self.health <= 0 then
			self.health = self:DoDeathCheck()
		end

		if not self.dead then
			for statusEffect in pairs(self.statusEffects) do
				statusEffect:OnDamageApplied(adjusted_amount)
			end
		end
		
		self:Broadcast()
	end
end

function HeroState:OnEventRestoreHealth(event)
	local amount = event:GetAttribute("amount")
	local actual_amount = amount
	
	self.health = self.health + amount
	
	if self.health > self.max_health then
		actual_amount = amount - (self.health - self.max_health)
		
		self.health = self.max_health
	end
	
	if actual_amount > 0 then	-- [NEOBEN]
		local notify = GameEventManager:Construct("ImmediateRestoreNotify")
		notify:SetAttribute("amount", actual_amount)
		notify:SetAttribute("client", self.client)
		GameEventManager:Send(notify)
	end
	
	self:Broadcast()
end

function HeroState:OnEventGivePower(event)
	self.power = self.power + event:GetAttribute("amount")
	self:Broadcast()
end

function HeroState:OnEventGiveBattleGold(event)
	self.gold = self.gold + event:GetAttribute("amount")
	self:Broadcast()
end

function HeroState:OnEventGiveBattleExperience(event)
	self.xp = self.xp + event:GetAttribute("amount")
	self:Broadcast()
end

function HeroState:OnEventManaDrain(event)
	self.red = 0
	self.green = 0
	self.blue = 0
	self.yellow = 0
	self.black = 0
	
	self:NotifyEffectsManaIsDrained("red")
	self:NotifyEffectsManaIsDrained("green")
	self:NotifyEffectsManaIsDrained("blue")
	self:NotifyEffectsManaIsDrained("yellow")
	self:NotifyEffectsManaIsDrained("black")
		
	self:Broadcast()
end

function HeroState:Broadcast()
	local clientInfo = GameEventManager:Construct("ClientSnapshot")
	
	clientInfo:SetAttribute("client", self.client)
	clientInfo:SetAttribute("health", self.health)
	clientInfo:SetAttribute("name", self.hero:GetName())	--GetAttribute("name"))
	clientInfo:SetAttribute("level", self.hero:GetAttribute("level"))
	clientInfo:SetAttribute("desc_tag", self.hero:GetDescriptionTag())
	clientInfo:SetAttribute("class", self.hero:GetAttribute("class"))
	clientInfo:SetAttribute("max_health", self.max_health)
	clientInfo:SetAttribute("red", self.red)
	clientInfo:SetAttribute("green", self.green)
	clientInfo:SetAttribute("blue", self.blue)
	clientInfo:SetAttribute("yellow", self.yellow)
	clientInfo:SetAttribute("black", self.black)
	clientInfo:SetAttribute("base_max_red", self.hero_invariant_stats.derived.mana.maximum.red)
	clientInfo:SetAttribute("base_max_green", self.hero_invariant_stats.derived.mana.maximum.green)
	clientInfo:SetAttribute("base_max_blue", self.hero_invariant_stats.derived.mana.maximum.blue)
	clientInfo:SetAttribute("base_max_yellow", self.hero_invariant_stats.derived.mana.maximum.yellow)
	clientInfo:SetAttribute("base_max_black", self.hero_invariant_stats.derived.mana.maximum.black)
	clientInfo:SetAttribute("actual_max_red", self.actual_max_red)
	clientInfo:SetAttribute("actual_max_green", self.actual_max_green)
	clientInfo:SetAttribute("actual_max_blue", self.actual_max_blue)
	clientInfo:SetAttribute("actual_max_yellow", self.actual_max_yellow)
	clientInfo:SetAttribute("actual_max_black", self.actual_max_black)
	clientInfo:SetAttribute("power", self.power)
	clientInfo:SetAttribute("gold", self.gold)
	clientInfo:SetAttribute("xp", self.xp)
	clientInfo:SetAttribute("base_attack_rating", self.hero_invariant_stats.derived.match_bonus.skull)
	clientInfo:SetAttribute("base_defence_rating", self.hero_invariant_stats.derived.defence_rating)
	local attack = self:GetAttack()
	clientInfo:SetAttribute("actual_attack_rating", attack )
	local defence = self:GetDefence()
	clientInfo:SetAttribute("actual_defence_rating", defence )
	clientInfo:SetAttribute("base_spell_cost_modifier", self.spell_cost_modifier)
	clientInfo:SetAttribute("actual_spell_cost_modifier", self:GetActualSpellCostModifier())
	-- [SCF] Added stats
	local value1, modified1 = self.hero:GetStat("strength")
	local value2, modified2 = self.hero:GetStat("agility")
	local value3, modified3 = self.hero:GetStat("stamina")
	local value4, modified4 = self.hero:GetStat("intelligence")
	local value5, modified5 = self.hero:GetStat("morale")
	clientInfo:SetAttribute("strength", value1)
	clientInfo:SetAttribute("agility", value2)
	clientInfo:SetAttribute("stamina", value3)
	clientInfo:SetAttribute("intelligence", value4)
	clientInfo:SetAttribute("morale", value5)
	local value6, modified6 = self:GetWeaponBonusDamage(0)
	local value7, modified7 = self:GetWeaponCriticalChance()
	local value8, modified8 = self:GetShieldCriticalChance()
	local value9, modified9 = self:GetSpellPenetration()
	local value10, modified10 = self:GetSpellResistChance()
	local value11, modified11 = self:GetBlockingEffectiveness()
	clientInfo:SetAttribute("weapon_bonus_damage", value6)
	clientInfo:SetAttribute("weapon_critical_chance", value7)
	clientInfo:SetAttribute("shield_critical_chance", value8)
	clientInfo:SetAttribute("spell_penetration", value9)
	clientInfo:SetAttribute("spell_resist_chance", value10)
	clientInfo:SetAttribute("blocking_effectiveness", value11)
	
	GameEventManager:Send(clientInfo)
end

function HeroState:GetActualSpellCostModifier()
	return MutateValueWithStatusEffects("GetSpellCostModifier", self.statusEffects, self.spell_cost_modifier)
end

function HeroState:ApplyItemMagicalEffects(event, params, ...)
	self.hero:ApplyItemMagicalEffects(event, params, ...)
end

function HeroState:RestoreHealth(caster, amount)
	assert(caster)
	assert(amount)

	local restoreHealth = GameEventManager:Construct("RestoreHealth")
	
	restoreHealth:SetAttribute("amount", amount)
	restoreHealth:SetAttribute("healer", caster)
	
	GameEventManager:Send(restoreHealth, self)
end

function HeroState:DamageHealth(caster, amount, immediate, reason)
	assert(caster)
	assert(amount)
	
	if immediate then
		assert(type(immediate) == "string")
	end
	
	if reason then
		assert(type(reason) == "string")
	end
	
	local damageHealth = GameEventManager:Construct("DamageHealth")
	
	damageHealth:SetAttribute("amount", amount)
	damageHealth:SetAttribute("source", caster)
	
	if immediate then
		damageHealth:SetAttribute("immediate", immediate)
	end
	
	if reason then
		damageHealth:SetAttribute("reason", reason)
	end
	
	GameEventManager:Send(damageHealth, self)
end

function HeroState:RestoreMana(caster, pool, amount)
	assert(caster)
	assert(amount)
	assert(pool)

	local restoreMana = GameEventManager:Construct("RestoreMana")
	
	restoreMana:SetAttribute("amount", amount)
	restoreMana:SetAttribute("pool", poolTranslation[pool])
	restoreMana:SetAttribute("healer", caster)
	
	GameEventManager:Send(restoreMana, self)
end

function HeroState:DamageMana(caster, pool, amount)
	assert(caster)
	assert(amount)
	
	local damageMana = GameEventManager:Construct("DamageMana")
	
	damageMana:SetAttribute("amount", amount)
	damageMana:SetAttribute("pool", poolTranslation[pool])
	damageMana:SetAttribute("source", caster)
	
	GameEventManager:Send(damageMana, self)
	
	return math.min(self[poolTranslation[pool]], amount)
end

function HeroState:DamageManaImmediately(pool, amount)
	assert(amount)
	assert(pool)
	assert(IsValidPool(pool), string.format("DamageMana event is draining an unknown pool: %s", pool))
	local pool_max = "actual_max_" .. pool
	self[pool] = self[pool] - amount
	if self[pool] < 0 then
		self[pool] = 0
	end
	
	if self[pool] == 0 then
		self:NotifyEffectsManaIsDrained(pool)
	end
	
	self:Broadcast()
end

function HeroState:NotifyEffectsManaIsDrained(pool)
	-- notify all running effects that the pool has been depleted (in case they need to stop or something)
	for _, effect in pairs(self.statusEffects) do
		if effect.NotifyManaPoolIsDepleted then
			effect:NotifyManaPoolIsDepleted(pool)
		end
	end
end

-- Drains from sources, gives to target, AT MOST "amount" specified
function HeroState:LeechHealth(caster, from, amount_in)
	assert(caster)
	assert(from)
	assert(type(amount_in) == "number")
	
	local amount
	
	if amount_in > self.health then
		amount = self.health
	else
		amount = amount_in
	end

	from:DamageHealth(caster, amount)
	self:RestoreHealth(caster, amount)
end

-- Drains from sources, gives to target, AT MOST "amount" from the pool specified
function HeroState:LeechMana(caster, from, pool_in, amount_in)
	assert(poolTranslation[pool_in])
	assert(caster)
	assert(from)
	assert(type(amount_in) == "number")
	
	local amount
	local pool = poolTranslation[pool_in]
	
	if amount_in > self[pool] then
		amount = self[pool]
	else
		amount = amount_in
	end

	from:DamageMana(caster, pool, amount)
	self:RestoreMana(caster, pool, amount)
end

function HeroState:OnAddStatusEffectModifyManaLimits(effect)
	self.actual_max_red = self:GetActualRedManaMax()
	self.actual_max_green = self:GetActualGreenManaMax()
	self.actual_max_blue = self:GetActualBlueManaMax()
	self.actual_max_yellow = self:GetActualYellowManaMax()
	self.actual_max_black = self:GetActualBlackManaMax()
	
	if self.red > self.actual_max_red then
		self.red = self.actual_max_red
	end
	
	if self.green > self.actual_max_green then
		self.green = self.actual_max_green
	end
	
	if self.blue > self.actual_max_blue then
		self.blue = self.actual_max_blue
	end
	
	if self.yellow > self.actual_max_yellow then
		self.yellow = self.actual_max_yellow
	end
	
	if self.black > self.actual_max_black then
		self.black = self.actual_max_black
	end
end

function HeroState:OnRemoveStatusEffectModifyManaLimits(effect)
	self.actual_max_red = self:GetActualRedManaMax()
	self.actual_max_green = self:GetActualGreenManaMax()
	self.actual_max_blue = self:GetActualBlueManaMax()
	self.actual_max_yellow = self:GetActualYellowManaMax()
	self.actual_max_black = self:GetActualBlackManaMax()
	
	if self.red > self.actual_max_red then
		self.red = self.actual_max_red
	end
	
	if self.green > self.actual_max_green then
		self.green = self.actual_max_green
	end
	
	if self.blue > self.actual_max_blue then
		self.blue = self.actual_max_blue
	end
	
	if self.yellow > self.actual_max_yellow then
		self.yellow = self.actual_max_yellow
	end
	
	if self.black > self.actual_max_black then
		self.black = self.actual_max_black
	end
end

function HeroState:SendStats(opponent)
	local stats = GameEventManager:Construct("BattleStats")
	
	stats:SetAttribute("client", self.client)
	stats:SetAttribute("opponent", assert(opponent))
	--stats:SetAttribute("best_stealth_damage", self.stats.best_stealth_damage)
	--stats:SetAttribute("total_stealth_damage", self.stats.total_stealth_damage)
	stats:SetAttribute("best_skull_damage", self.stats.best_skull_damage)
	stats:SetAttribute("total_skull_damage", self.stats.total_skull_damage)
	stats:SetAttribute("best_blocked_damage", self.stats.best_blocked_damage)
	stats:SetAttribute("total_blocked_damage", self.stats.total_blocked_damage)
	stats:SetAttribute("total_mana_used", self.stats.total_mana_used)
	stats:SetAttribute("num_bonus_moves", self.stats.num_bonus_moves)
	stats:SetAttribute("best_bonus_moves_streak", self.stats.best_bonus_moves_streak)
	stats:SetAttribute("spell1_use", self.stats.spell_use[1])
	stats:SetAttribute("spell2_use", self.stats.spell_use[2])
	stats:SetAttribute("spell3_use", self.stats.spell_use[3])
	stats:SetAttribute("spell4_use", self.stats.spell_use[4])
	stats:SetAttribute("spell5_use", self.stats.spell_use[5])
	stats:SetAttribute("total_stealth_absorbed_damage", self.stats.total_stealth_damage_absorbed)
	stats:SetAttribute("num_item_activations", self.stats.num_item_activations)
	stats:SetAttribute("total_damage_taken", self.stats.total_damage_taken)
	stats:SetAttribute("total_damage_given", self.stats.total_damage_given)
	stats:SetAttribute("best_attack_rating", self.stats.best_attack_rating)
	stats:SetAttribute("best_mana_gain", self.stats.best_mana_gain)
	stats:SetAttribute("best_damage", self.stats.best_skull_damage)
	stats:SetAttribute("num_turns", self.stats.num_turns)
	
	
	LOGF("STATS: sending the stats collected from a player")
	table.log(self.stats)
	
	GameEventManager:Send(stats)
end

return ExportClass("HeroState")
