
--[[

Not surprisingly, this manages heroes:

]]

local DEBUG_LOG = LOG -- uncomment to enable logging
--local DEBUG_LOG = function() end

local DEFAULT_MAX_NUMBER_OF_SAVES = 256

local HeroSaveManager = {}

--	Returns an "instance" of the HeroSaveManager
function HeroSaveManager:Create()
	return HeroSaveManager._Construct({})
end

--	Creates an instance of the HeroSaveManager, copying all the desired values/functions across
function HeroSaveManager:_Construct()
    for k,v in pairs(HeroSaveManager) do
        self[k] = v
    end
	
	self:ResetSaving()
	self:SetMaxNumberOfSaves(DEFAULT_MAX_NUMBER_OF_SAVES)
	
    return self
end

--	Destroys all currently loaded saves, and nils self variables, should be called before discarding an Instance of the HeroSaveManager
function HeroSaveManager:Destroy()
	self:ClearHeroes()
	self:ResetActiveHero()
	self.saveHeroEnabled = nil
end

function HeroSaveManager:ResetSaving()
	self:SetHeroSaveEnabled(true)
	self:SetSaveSystemEnabled(true)
end

function HeroSaveManager:GetMaxNumberOfSaves()
	return self.maxSaves
end

function HeroSaveManager:SetMaxNumberOfSaves(value)
	self.maxSaves = value
end

function HeroSaveManager:IsSaveEnabled()
	return (self.saveHeroEnabled and self.saveSystemEnabled)
end

function HeroSaveManager:IsHeroSaveEnabled()
	return self.saveHeroEnabled
end

function HeroSaveManager:SetHeroSaveEnabled(value)
	self.saveHeroEnabled = value
end

function HeroSaveManager:IsSaveSystemEnabled()
	return self.saveSystemEnabled
end

function HeroSaveManager:SetSaveSystemEnabled(value)
	self.saveSystemEnabled = value
end

function HeroSaveManager:HasDeviceChangedSinceLoad()
	return self.deviceChangedSinceLoad
end

function HeroSaveManager:SetDeviceChangedSinceLoad(value)
	self.deviceChangedSinceLoad = value
end

function HeroSaveManager:AddNewHero(newHero)
	self.heroList = self.heroList or {}
	local heroData = {
		hero = newHero
	}
	table.insert(self.heroList, heroData)
	return heroData
end

function HeroSaveManager:GetActiveHero()
	if self.activeHero then
		return self.activeHero.hero
	else
		return nil
	end
end

function HeroSaveManager:GetActiveHeroSaveHandle()
	if self.activeHero then
		return self.activeHero.saveHandle
	else
		return nil
	end
end

function HeroSaveManager:ResetActiveHero()
	self.activeHero = nil
end

-------------------------------------------------------------------------------
-- ReloadActiveHero
--	Deserializes the player data again from the save handle.
function HeroSaveManager:ReloadActiveHero()
	LOG("[HeroSaveManager] ReloadActiveHero")
	local heroData = self.activeHero
	
	assert(heroData, "No data at the specified index")
	assert(heroData.saveHandle, "No save handle for this position. This really should not happen!")

	if heroData.hero then
		GameObjectManager:Destroy(heroData.hero)
		heroData.hero = nil
	end
	
	local savedData = heroData.saveHandle:Deserialize()
	
	for _, object in pairs(savedData) do
		if object.isHero then
			heroData.hero = object
			heroData.saveHandle:SetName(object:GetAttribute("name"))
			break
		end
	end
end

function HeroSaveManager:SetActiveHero(hero)
	local index
	if (type(hero) == "number") then
		index = hero
	else
		assert(hero.isHero)
		for k, heroData in ipairs(self.heroList) do
			if heroData.hero and heroData.hero == hero then
				index = k
				break
			end
		end
	end

	assert(index)
	self.activeHero = self.heroList[index]
end

function HeroSaveManager:SaveActiveHero(ContinueCallback, FailCallback)
	assert(ContinueCallback, "[HeroSaveManager] SaveActiveHero called without a ContinueCallback, you must specify a ContinueCallback to be called when the save succeeds.")
	assert(FailCallback, "[HeroSaveManager] SaveActiveHero called without a FailCallback, you must specify a FailCallback that will be called in the case of save failure.")
	if not (self:IsHeroSaveEnabled() and self:IsSaveSystemEnabled()) then
		DEBUG_LOG("[HeroSaveManager:SaveActiveHero] Saving currently disabled. Just calling ContinueCallback")
		ContinueCallback()
	else
		DEBUG_LOG("[HeroSaveManager:SaveActiveHero] Saving currently enabled, attempting save.")
		-- Check the hero has a save handle, and if not, create one
		local heroData = self.activeHero
		if not heroData.saveHandle then
			heroData.saveHandle = self:CreateSaveGame(heroData.hero:GetName(), _G.SAVE_GAME_VERSION)
		end

		self:AssociateDevice(heroData.saveHandle)

		local save_success = heroData.saveHandle:Save({heroData.hero})
		if save_success then
			ContinueCallback()
		else
			self:HandleLastSaveError(heroData.saveHandle, ContinueCallback, FailCallback)
		end
	end
end

-- Removes the hero from the save list, the destruction and clean up of the hero then becomes the responsibility of the caller
function HeroSaveManager:ReleaseHero(targetHero)
	assert(self.heroList,"[HeroSaveManager:ReleaseHero] Error, no heroes currently loaded, nothing to release.")
	local targetIndex
	for index, heroData in pairs(self.heroList) do
		if targetHero == heroData.hero then
			targetIndex = index
			break
		end
	end
	
	table.remove(self.heroList,targetIndex)
end

-- Removes the hero from the save list, the destruction and clean up of the hero then becomes the responsibility of the caller
function HeroSaveManager:ReleaseHeroByIndex(targetIndex)
	assert(self.heroList,"[HeroSaveManager:ReleaseHero] Error, no heroes currently loaded, nothing to release.")
	assert(targetIndex and type(targetIndex) == "number" and targetIndex > 0 and targetIndex <= #self.heroList, string.format("Invalid index (%s) provided, must be a number between 1 and number of heroes (%d)", tostring(targetIndex), #self.heroList))
	self.heroList[targetIndex] = nil
end

-------------------------------------------------------------------------------
-- RetrieveHero
--	Retrieves the hero object from savedData.
function HeroSaveManager:RetrieveHero(targetIndex)
	LOG("[HeroSaveManager] RetrieveHero")
	assert(self.heroList, "Hero list has not been loaded.")
	
	local heroData = self.heroList[targetIndex]
	assert(heroData, "No data at the specified index")
	assert(heroData.saveHandle, "No save handle for this position. This really should not happen!")
	assert(not heroData.hero, "Hero has already been retrieved!")
	
	local load_succeeded = heroData.saveHandle:Load() -- table of loaded objects
	
	if load_succeeded then
		local savedData = heroData.saveHandle:Deserialize()
		
		for _, object in pairs(savedData) do
			if object.isHero then
				heroData.hero = object
				heroData.saveHandle:SetName(object:GetAttribute("name"))
				break
			end
		end
	else
		local lastError = SaveGameManager.GetLastError()
		if not heroData.saveHandle:IsValid() then
			-- Error was caused by pulling the device while it was enumerating, reset the bastard!
			_G.ResetToSplashWithError("[ERROR]", "[DEVICE_LOST]")
		else
			heroData.corrupt = true
		end
	end
end

-------------------------------------------------------------------------------
-- Destroys all the loaded heroes
function HeroSaveManager:ClearHeroes()
	LOG("[HeroSaveManager:ClearHeroes]")
	if self.heroList then
		for _, heroData in pairs(self.heroList) do
			if heroData.hero then
				LOG("HeroObject being destroyed")
				GameObjectManager:Destroy(heroData.hero)
			end
		end
		self.heroList = nil
	else
		DEBUG_LOG("[HeroSaveManager:ClearHeroes] Warning: Attempted to ClearHeroes when the list is already empty")
	end
	self.savesInvalidated = false
	self:SetDeviceChangedSinceLoad(false)
	self.activeHero = nil
end


-------------------------------------------------------------------------------
--
function HeroSaveManager:ClearNonActiveHeroes()
	LOG("[HeroSaveManager:ClearNonActiveHeroes]")
	assert(self.activeHero)
	
	if self.heroList then
		for _, heroData in pairs(self.heroList) do
			if self.activeHero ~= heroData and heroData.hero then
				LOG("HeroObject being destroyed")
				GameObjectManager:Destroy(heroData.hero)
			end
		end
	else
		DEBUG_LOG("[HeroSaveManager:ClearNonActiveHeroes] Warning: Attempted to ClearHeroes when the list is already empty")
	end
	
	self.heroList = { self.activeHero }
end

function HeroSaveManager:GetHeroList()
	assert(self.heroList, "[HeroSaveManager] Error: Hero list is not currently loaded, you need to call LoadHeroes first.")
	assert(not self.savesInvalidated, "[HeroSaveManager] Error: Saves have been invalidated, you need to check HeroSaveManager:AreSavesInvalid() before calling GetHeroList, and reload if AreSavesInvalid is true")
	return self.heroList
end

function HeroSaveManager:GetNumHeroes()
	if self:IsSaveSystemEnabled() and self.heroList then
		return #(self.heroList)
	else
		return 0
	end
end

function HeroSaveManager:AreSavesInvalid()
	return self.savesInvalidated
end

function HeroSaveManager.EnumerateSaveGames()
	-- overridden in the platform versions
end

function HeroSaveManager:CreateSaveGame(saveName, version)
	-- overridden in the platform versions
end

function HeroSaveManager:DeleteHero(index, ContinueCallback, ShutdownCallback)
	local saveHandle = self.heroList[index].saveHandle
	LOGF("[HeroSaveManager] DeleteHero at index=%s, saveHandle:GetName()=%s", tostring(index), tostring(saveHandle:GetName()))
	local function HandleResult(delete_success)
		if delete_success then
			local heroData = table.remove(self.heroList, index)
			if heroData.hero then
				GameObjectManager:Destroy(heroData.hero)
			end
		end
		ContinueCallback(delete_success)
	end

	if saveHandle and saveHandle:GetDevice() ~= 0 then -- check the save handle has a valid device associated with it, otherwise just delete the hero and drop the handle.
		self:DeleteSave(saveHandle, HandleResult, ShutdownCallback)
	else
		HandleResult(true)
	end
end

function HeroSaveManager:DeleteSave(saveHandle, ContinueCallback, ShutdownCallback)
	-- overridden in the platform versions
end

function HeroSaveManager:AssociateDevice(saveHandle)
	-- overridden in the platform versions
end

-- make the HeroSaveManager a global briefly so the platform file can add to it, then remove it from the global table
assert(not _G.HeroSaveManager, "_G.HeroSaveManager is currently set, and will be overwritten!")
_G.HeroSaveManager = HeroSaveManager
dofile("Assets/Scripts/Hero/HeroSaveManagerPlatform.lua")
_G.HeroSaveManager = nil

return HeroSaveManager
