
function Hero:GetNextMostImportantQuest()
	local runningQuests = GetRunningQuests(self)
	for i, questState in ipairs(runningQuests) do
		if questState:GetQuestData().type == "main" then
			return questState:GetAttribute("index")
		end
	end
	for i, questState in ipairs(runningQuests) do
		return questState:GetAttribute("index")
	end
end

function Hero:SetRequireSave(toSave)
	if not toSave then
		self.requireSave = false
	else
		if IsOpen("WorldMapMenu") then
			self.requireSave = true
		else
			SCREENS.SaveMenu:Display()
		end
	end
end

function Hero:GetRequireSave()
	return self.requireSave
end

function Hero:PushQuestFeedback(item)
    table.insert(self.questMessages, item)
end

function Hero:PopQuestFeedback()
	if IsOpen("WorldMapMenu") then
		SCREENS.WorldMapMenu.hero:DisableMovement()
	end
	
	for i, message in ipairs(self.questMessages) do	-- deal with level up first
		if message.title == "[LEVELUP_HEAD]" then
			return table.remove(self.questMessages, i)
		end
	end
	
    return table.remove(self.questMessages, 1)
end

function Hero:HasQuestFeedback()
	return next(self.questMessages)
end

function Hero:HasOrHasCompletedQuest(questID)
	return IsQuestRunning(self, questID) or IsQuestCompleted(self, questID)
end

function Hero:OnEventUpdateQuestUI(event)
	LOG("NICK: OnEventUpdateQuestUI")

	-- if we dont have an active quest, lets find a quest to activate
	if not self:GetActiveQuest() then
		local questIndex = self:GetNextMostImportantQuest()
		if questIndex then
			self:SetActiveQuest(questIndex)
		end
	end
	
	self:PushQuestFeedback({
		modal = false,
		action = function (view, node)
			view:UpdateGraphVisibility(self)				-- quest changes can make nodes become visible
			view:UpdateQuestEncounters(self)				-- quest changes can make new monsters spawn
			view:UpdateQuestNPCs(self)						-- quest changes can make new npcs spawn
			view:UpdateQuestIcons(self)						-- quest changes can change quest start/objective locations inside a room
			view:UpdateQuestBarriers(self)					-- quest changes can make barriers change states
			view:UpdateQuestPortals(self)					-- quest changes can change the portals
			view:UpdateQuestOverlays(self)					-- quest changes can change the overlays
			view:UpdateQuestDirections(self)				-- quest changes can change the direction hints!
			view:UpdateCompass(self)						-- quest changes can change quest start/objective locations outside a room
		end,
	})
end

function Hero:OnQuestEventStartQuest(questState)
	local index = questState:GetAttribute("index")
	local activeQuest = self:GetActiveQuest()
	if not activeQuest then
		self:SetActiveQuest(index)
	elseif GetQuestFromIndex(activeQuest).type == "side" and questState:GetQuestData().type == "main" then
		self:SetActiveQuest(index)
	end
	local questData = questState:GetQuestData()
    self:PushQuestFeedback({
		title = "[QUEST_NEW]",
		subtitle = questData.name_tag,
		type = questData.start_show,
		modal = questData.start_show == "big",
		index = index,
		action = function()
			SoundSystem.Play("pq2audio/ui/questnew")
		end
    })
end

function Hero:OnQuestEventCompleteQuest(questState)
	local index = questState:GetAttribute("index")
	-- if this was the active quest, pick another quest
	if self:GetAttribute("activeQuest") == index then
		local nextQuest = self:GetNextMostImportantQuest()
		if nextQuest then
			self:SetActiveQuest(nextQuest)
		else
			self:SetActiveQuest(0)
		end
	end

	local questData = questState:GetQuestData()
    self:PushQuestFeedback({
		title = "[QUEST_COMPLETED]",
		subtitle = questData.name_tag,
		type = questData.complete_show,
		modal = questData.complete_show == "big",
		action = function ()
			SoundSystem.Play("pq2audio/ui/questdone")
			
			self:SetRequireSave(true)
			
			local questHandlers = require("Quests.QuestHandlers")
			local questHandler = questHandlers[questState:GetAttribute("index")]
			
			if questHandler and questHandler.OnComplete then
				questHandler.OnComplete()
			end
		end		
    })
end

function Hero:OnQuestActionSetMusicPlayerState(action, index)
	assert(action)
	assert(action.state)
	
	self:PushQuestFeedback({
		action = function ()
			MusicManager:SetState(action.state)
		end,
	})
end

function Hero:OnQuestActionSetActiveQuest(action, index)
	assert(action)
	assert(action.quest)
	
	self:SetActiveQuest(action.quest)
end

function Hero:OnEventQuestNewObjective(questState, startedStateIndex)
	local stateData = questState:GetQuestData().states[startedStateIndex]
	self:PushQuestFeedback({
		title = "[QUEST_NEW_OBJECTIVE]",
		subtitle = stateData.log,
		type = stateData.start_show,
		modal = stateData.start_show == "big",
		action = function ()
			SoundSystem.Play("pq2audio/ui/questobjective")
		end
    })
end

function Hero:OnEventQuestObjectiveComplete(questState, objective, nextStateIndex)
	local stateData = questState:GetQuestData().states[startedStateIndex]
    self:PushQuestFeedback({
		title = "[QUEST_OBJECTIVE_COMPLETE]",
		subtitle = stateData.log,
		type = stateData.complete_show,
		modal = stateData.complete_show == "big",
		action = function ()
			SoundSystem.Play("pq2audio/ui/questtask")
		end
    })
end

function Hero:OnQuestActionShowMessageBox(action, quest)
	assert(action.text)
	assert(action.icon)
	assert(quest)
	self:PushQuestFeedback({
		modal = true,
		action = function ()
			SCREENS.QuestFeedbackMenu:SetupAndOpen(action.title or GetQuestName(quest), action.text, action.icon, function ()
			end)
		end
	})
end

function Hero:OnQuestActionGiveExperience(action)
	assert(action.experience)
	self:PushQuestFeedback({
		title    = translate_text("[QUEST_REWARD]"),
		subtitle = substitute(translate_text("[QUEST_XP_GAINED_X]"), action.experience),
		type = action.show,
		modal = action.show == "big",
		action = function ()
			SoundSystem.Play("pq2audio/ui/questtask")
		end
	})
	self:AwardExperience(action.experience)
end

function Hero:OnQuestActionGiveGold(action)
	assert(action.gold)
	self:PushQuestFeedback({
		title    = translate_text("[QUEST_REWARD]"),
		subtitle = substitute(translate_text("[QUEST_GOLD_GAINED_X]"), action.gold),
		type = action.show,
		modal = action.show == "big",
		action = function ()
			SoundSystem.Play("pq2audio/ui/questtask")
		end
	})
	self:AwardGold(action.gold)
end

function Hero:OnQuestActionGiveTradeItem(action)
	assert(action.resource)
	assert(action.amount)
	self:AddCommodity(action.resource, action.amount)
	self:PushQuestFeedback({
		title = "[QUEST_TRADEITEM_GIVE]",
		subtitle = substitute(translate_text("[QUEST_TRADEITEM_GIVE_X]"), translate_text(string.format("[%s_NAME]", TRADEITEMS[action.resource].id)), tostring(action.amount)),
		type = action.show,
		icon = TRADEITEMS[action.resource].icon,
		modal = action.show == "big",
		action = function ()
			SoundSystem.Play("pq2audio/ui/questobjective")
		end
	})
end

function Hero:OnQuestActionTakeTradeItem(action)
	assert(action.resource)
	assert(action.amount)
	self:RemoveCommodity(action.resource, action.amount)
	self:PushQuestFeedback({
		title = "[QUEST_TRADEITEM_TAKE]",
		subtitle = substitute(translate_text("[QUEST_TRADEITEM_TAKE_X]"), translate_text(string.format("[%s_NAME]", TRADEITEMS[action.resource].id)), tostring(action.amount)),
		type = action.show,
		icon = TRADEITEMS[action.resource].icon,
		modal = action.show == "big",
		action = function ()
			SoundSystem.Play("pq2audio/ui/questobjective")
		end
	})
end

function Hero:OnQuestActionGiveFriend(action)
	assert(action.friend)
	if not CollectionContainsAttribute(self, "friends", action.friend) then
		self:PushAttribute("friends", action.friend)
		
		local friend = import("Friends." .. action.friend)
		self:PushQuestFeedback({
			title = "[QUEST_FRIEND_GIVE]",
			subtitle = friend.texttag,
			type = action.show,
			modal = action.show == "big",
			action = function()
				SoundSystem.Play("pq2audio/ui/questobjective")
			end
		})
	end
end

function Hero:OnQuestActionTakeFriend(action)
	assert(action.friend)
	if CollectionContainsAttribute(self, "friends", action.friend) then
		self:EraseAttribute("friends", action.friend)
		
		local friendsList = import("Friends/Friends")
		local friend
		for _, friendData in ipairs(friendsList) do
			if (friendData.name == action.friend) then
				friend = friendData
				break
			end
		end
		
		assert(friend, string.format("Unable to find matching friend data for friend '%s'", tostring(action.friend)))
		
		self:PushQuestFeedback({
			title = "[QUEST_FRIEND_TAKE]",
			subtitle = friend.texttag,
			type = action.show,
			modal = action.show == "big",
			action = function()
				SoundSystem.Play("pq2audio/ui/questobjective")
			end
	    })
	end
end

function Hero:OnQuestActionAwardAchievement(action)
	local count = action.count
	
	if count == -1 then
		count = nil
	end
	
	Achievements.AwardAchievement(1, assert(action.achievement), count)
	
	-- There is one quest achievement (SAVEVERLOREN) that we can trigger in the demo
	if _G.IsDemo() then
		Achievements.DisplayNewDemoAchievements()
	end
end

function Hero:OnQuestActionGiveRandomItem(action)
	-- implement me
end

function Hero:OnQuestActionShowConversation(action)
	assert(action.conversation)
	self:PushQuestFeedback({
		modal = true,
		action = function ()
			open_conversation_menu(
				"Assets/Conversations/" .. action.conversation .. ".xml",
				self:GetName(), 
				self:GetConversationPortrait(), 
				self:GetGenderForConversation(), 
				"", 
				"",
				"", 
				function () end, 
				x or GetScreenWidth()*0.5,
				y or GetScreenHeight()*0.5
			)
		end
	})
end

function Hero:OnQuestActionShowCutscene(action)
	add_asset_file("Assets/AssetsNIS.xml")
	assert(action.cutscene)
	self:PushQuestFeedback({
		modal = true,
		action = function ()
			local hero_voice = 0
			if self:GetAttribute("sex") == "M" then
				hero_voice = 1
			end
			SCREENS.WorldMapMenu:SuspendAmbientSound()
			SoundSystem.Load("amb_bank")
			SoundSystem.SetParameter( string.format( "pq2audio/amb/amb_%s", string.lower( action.cutscene ) ), "hero_gender", hero_voice )
			MusicManager:Stop()
			open_nis_menu(action.cutscene, function ()
				LOG("NIS completed")
				MusicManager:Resume()
				SCREENS.WorldMapMenu:ResumeAmbientSound()
			end)
		end
	})
end

function Hero:OnQuestActionStartQuest(action)
	assert(action.quest)
	StartQuest(self, action.quest)
end

function Hero:OnQuestActionGiveQuestItem(action)
	assert(action.questItem)
	if not CollectionContainsAttribute(self, "questItems", action.questItem) then
		self:PushAttribute("questItems", action.questItem)
		
		local questItem = import("QuestItems." .. action.questItem)
		self:PushQuestFeedback({
			title = "[QUEST_ITEM_GIVEN]",
			subtitle = string.format("[ITEM_QUEST_%s]", questItem.texttag),
			type = action.show,
			icon = questItem.icon,
			modal = action.show == "big",
			action = function()
				SoundSystem.Play("pq2audio/ui/questobjective")
			end
	    })
	end
end

function Hero:OnQuestActionTakeQuestItem(action)
	assert(action.questItem)
	if CollectionContainsAttribute(self, "questItems", action.questItem) then
		self:EraseAttribute("questItems", action.questItem)
		
		local questItem = import("QuestItems." .. action.questItem)
		self:PushQuestFeedback({
			title = "[QUEST_ITEM_TAKEN]",
			subtitle = string.format("[ITEM_QUEST_%s]", questItem.texttag),
			type = action.show,
			icon = questItem.icon,
			modal = action.show == "big",
			action = function()
				SoundSystem.Play("pq2audio/ui/questobjective")
			end
	    })
	end
end

function Hero:OnQuestActionHighlightUI(action)
	assert(action.widget)
	self:PushQuestFeedback({
		modal = false,
		action = function (world)
			world:HighlightUI(action.widget)
		end
	})
end

function Hero:OnQuestActionGiveItemChoice(action)
	assert(action.items)
	self:PushQuestFeedback({
		modal = true,
		action = function ()
			SCREENS.QuestRewardsMenu:Open()
			SCREENS.QuestRewardsMenu:SetData(self, action.items)
		end
	})
end

function Hero:OnQuestActionTeleportHero(action)
	assert(action.room)
	assert(action.node)
	assert(action.exitFacingDirection)
	
	self:PushQuestFeedback({
		modal = true,
		action = function ()
			SoundSystem.Play("pq2audio/ui/teleport")
			SCREENS.BlankLoadingMenu:SetupAndOpen(action.room, function ()
				SCREENS.WorldMapMenu:SetRoomTransitionFlag()
				SCREENS.WorldMapMenu:Close()
				
				StartWorldMapAt(action.room, action.node, action.exitFacingDirection)()
				
				self:SetRequireSave(true)
				SCREENS.BlankLoadingMenu:Close()
			end)
		end
	})
end

function Hero:OnQuestActionStartUnlockGame(action)
	local id = self:GetAttribute("currentNode")
	local game = action.minigame
	local difficulty = action.difficulty
	local hero1 = self
	
	local function EndActivity(win)
		if win then
-- This is what a WorldMapNode does when finishing an unlock minigame, I am not
-- sure what parts of this we will need to do, lets wait for tickets.
--			local hero = HeroManager:GetActiveHero()
--			hero:SetDoorUnlocked(self.activeBarrier.lock_id)
--			self.activeBarrier = nil
--			self:RemoveNodeMarker()
--			self:SetupDoor(hero)
--			self:SetNodeToGreen()
--			self:GetParent():UpdateQuestDirections(hero)
--			
--			SCREENS.SaveMenu:Open()
		end
		
		if win then
			OnQuestEvent(hero1, "DoorUnlocked", {
				location = id,
				minigame = action.minigame,
				difficulty = action.difficulty,
				battleground = action.battleground,
			})
		else
			OnQuestEvent(hero1, "DoorNotUnlocked", {
				location = id,
				minigame = action.minigame,
				difficulty = action.difficulty,
				battleground = action.battleground,
			})
		end
		
		hero1:SetRequireSave(true)
	end
	
	local nextScreen = SCREENS[string.format("%sGameMenu", game)]
	nextScreen:SetDifficulty(difficulty)
	nextScreen:SetEndMinigame(EndActivity)
	nextScreen:SetGameOptions({conversation=action.conversation})
	nextScreen:Open()
end

function Hero:OnQuestActionStartSearchGame(action)
	local id = self:GetAttribute("currentNode")
	local difficulty = action.difficulty
	local hero1 = self
	
	local function EndActivity(win)
		if win then
			OnQuestEvent(hero1, "RoomSearched", {
				location = id,
				difficulty = action.difficulty,
				battleground = action.battleground,
			})
		else
			OnQuestEvent(hero1, "RoomNotSearched", {
				location = id,
				difficulty = action.difficulty,
				battleground = action.battleground,
			})
		end
		
		hero1:SetRequireSave(true)
	end
	
	local nextScreen = SCREENS["SearchGameMenu"]
	nextScreen:SetDifficulty(difficulty)
	nextScreen:SetEndMinigame(EndActivity)
	nextScreen:SetGameOptions({conversation=action.conversation})
	nextScreen:Open()
end

function Hero:OnQuestActionStartDisarmGame(action)
	local id = self:GetAttribute("currentNode")
	local difficulty = action.difficulty
	local hero1 = self
	
	local function EndActivity(win)
		if win then
			OnQuestEvent(hero1, "TrapDisarmed", {
				location = id,
				difficulty = action.difficulty,
				battleground = action.battleground,
			})
		else
			OnQuestEvent(hero1, "TrapNotDisarmed", {
				location = id,
				difficulty = action.difficulty,
				battleground = action.battleground,
			})
		end
		
		hero1:SetRequireSave(true)
	end
	
	local nextScreen = SCREENS["DisarmGameMenu"]
	nextScreen:SetDifficulty(difficulty)
	nextScreen:SetEndMinigame(EndActivity)
	nextScreen:SetBoard(action.battleground)
	nextScreen:SetGameOptions({conversation=action.conversation})
	nextScreen:Open()
end

function Hero:OnQuestActionStartTrainingGame(action)
	local id = self:GetAttribute("currentNode")
	local difficulty = action.difficulty
	local hero1 = self
	
	local function EndActivity(win)
		if win then
			OnQuestEvent(hero1, "SpellTrained", {
				location = id,
				battleground = action.battleground,
			})
		else
			OnQuestEvent(hero1, "SpellNotTrained", {
				location = id,
				battleground = action.battleground,
			})
		end
		
		hero1:SetRequireSave(true)
	end
	
	local nextScreen = SCREENS["CaptureGameMenu"]
	nextScreen:SetDifficulty(difficulty)
	nextScreen:SetEndMinigame(EndActivity)
	nextScreen:SetBoard(action.battleground)
	nextScreen:SetGameOptions({conversation=action.conversation})
	nextScreen:Open()
end

function Hero:OnQuestActionStartLootGame(action)
	local id = self:GetAttribute("currentNode")
	local difficulty = action.difficulty
	local hero1 = self
	
	local function EndActivity(win)
		if win then
			OnQuestEvent(hero1, "LootCollected", {
				location = id,
				difficulty = action.difficulty,
				battleground = action.battleground,
			})
		else
			OnQuestEvent(hero1, "LootNotCollected", {
				location = id,
				difficulty = action.difficulty,
				battleground = action.battleground,
			})
		end
		
		hero1:SetRequireSave(true)
	end
	
	local nextScreen = SCREENS["HoardGameMenu"]
	nextScreen:SetDifficulty(difficulty)
	nextScreen:SetEndMinigame(EndActivity)
	nextScreen:SetBoard(action.battleground)
	nextScreen:SetGameOptions({conversation=action.conversation})
	nextScreen:Open()
end

function Hero:OnQuestActionStartEncounter(action)
	local id = self:GetAttribute("currentNode")
	local hero1 = self
	local hero2 = self:GetSessionDataForEnemy(id, action.monsterNumber or 0, action.monster)
	local lastNode = SCREENS.WorldMapMenu.hero.lastNode:GetAttribute("id")
	
	local function OnBattleResult(res)
		if res.winningTeam == 1 then
			OnQuestEvent(hero1, "MonsterTypeKilled", {
				monster = action.monster,
				id
			})
			OnQuestEvent(hero1, "MonsterKilled", {
				monster = action.monster,
				battleground = action.battleground,
				location = id,
			})
		elseif res.winningTeam == 2 then
			OnQuestEvent(hero1, "MonsterNotKilled", {
				monster = action.monster,
				battleground = action.battleground,
				location = id,
			})
		end
	end
	
	local function OnBattleFinished(result)
		-- Restore presence
		Presence:SetPresenceMode("SINGLEPLAYER")
	
		-- open up the world map menu to put the player back in the world
		SCREENS.WorldMapMenu = WorldMapMenu()
		SCREENS.WorldMapMenu:Load(hero1:GetAttribute("currentLocation"), hero1:GetAttribute("currentNode"), hero1:GetAttribute("currentDir"))
		SCREENS.WorldMapMenu:Open()
		
		if result and result.winningTeam == 2 then
			local worldmaphero = SCREENS.WorldMapMenu.hero
			worldmaphero:MoveTo(lastNode)
			local location = worldmaphero.pendingTarget	-- HAX!	-- see ticket #3712
			if location then
				hero1:SetAttribute("currentNode", location:GetAttribute("id"))
			end
		end
		
		-- save game
		self:SetRequireSave(true)
	end
	
	local function OnFlee()
		-- Restore presence
		Presence:SetPresenceMode("SINGLEPLAYER")
	
		if not IsOpen("WorldMapMenu") then
			SCREENS.WorldMapMenu = WorldMapMenu()
			SCREENS.WorldMapMenu:Load(hero1:GetAttribute("currentLocation"), hero1:GetAttribute("currentNode"), hero1:GetAttribute("currentDir"))
			SCREENS.WorldMapMenu:Open()
		end
		
		local worldmaphero = SCREENS.WorldMapMenu.hero
		worldmaphero:MoveTo(lastNode)
		local location = worldmaphero.pendingTarget	-- HAX!	-- see ticket #3712
		if location then
			hero1:SetAttribute("currentNode", location:GetAttribute("id"))
		end
	end
	
	local function OnFight(args)
		Presence:SetPresenceMode("SINGLEPLAYER_BATTLE")
		Presence:SetPresenceContext("MONSTER", string.upper(args.monsterHero.def.name))
	end
	
	local params = {}
	
	params.heroes = {}
	params.heroes[1] = {hero1}
	params.heroes[2] = {hero2}
	params.grid_layout = string.format("Grids.%s", action.battleground)
	
	local grid_layout = require(params.grid_layout)
	
	params.grid_width = grid_layout.grid_width
	params.grid_height = grid_layout.grid_height
	
	
	if action.battleRules then
		ConvertBattleRulesToGameOptions(action.battleRules, params)
	end
	
	params.gameOptions = params.gameOptions or {}
	params.gameOptions.conversation = action.conversation
	
	if action.challenge then
		SCREENS.ChallengeMenu:SetupAndOpen({
			subtitle = action.challengeText, 
			transitionFrom = "WorldMapMenu",
			playerHero = hero1,
			monsterHero = hero2, 
			onResult = OnBattleResult,
			onFinished = OnBattleFinished, 
			onFlee = OnFlee, 
			onFight = OnFight,
			params = params
		})
	else
		SCREENS.PreBattleMenu:SetupAndOpen({
			transitionFrom = "WorldMapMenu",
			playerHero = hero1,
			monsterHero = hero2,
			onResult = OnBattleResult,
			onFinished = OnBattleFinished,
			onFlee = OnFlee,
			onFight = OnFight,
			params = params
		})
	end
end

function Hero:CheckQuestConditionQuestAtState(quest_state, condition)
	assert(condition.quest)
	assert(condition.state)
	if IsQuestRunning(self, condition.quest) then
		return IsQuestAtState(self, condition.quest, condition.state)
	end
end

function Hero:CheckQuestConditionQuestCompleted(quest_state, condition)
	assert(condition.quest)
	return IsQuestCompleted(self, condition.quest)
end

function Hero:CheckQuestConditionHasQuestItem(quest_state, condition)
	assert(condition.item)
	return CollectionContainsAttribute(self, "questItems", condition.item)
end

function Hero:CheckQuestConditionHasTradeItem(quest_state, condition)
	assert(condition.item)
	assert(condition.count)
	return self:GetCommodityAmount(condition.item) >= condition.count
end
