
--[[

	An EquippableItem is any item that can be equipped on a monster or player.  This includes
	all swords, shields, armour, rings, amulets, etc.
	
	These kind of items all exhibit the following traits:
	    Base Type - "longsword", "dagger", etc
		Race - "orcish"
		Rarity - "standard", "legendary", etc
	
]]

use_safeglobals()

local RARITY_COLOURS = 
{
	Normal 			= {192, 192, 192, 255,			255, 255, 255, 255},
	Magical 		= {51, 204, 0, 255,				255, 255, 255, 255},
	Master 			= {51, 51, 255, 255,			255, 255, 255, 255},
	Renowned 		= {102, 0, 204, 255,			255, 255, 255, 255},
	Legendary 		= {255, 153, 0, 255,			255, 255, 255, 255},
}

local RARITY_OVERLAYS = 
{
	Normal			= "img_itemrarity_normal",
	Magical			= "img_itemrarity_magical",
	Master			= "img_itemrarity_master",
	Renowned		= "img_itemrarity_renowned",
	Legendary		= "img_itemrarity_legendary",
}

local RACE_ELEMENTS =
{
	BlackIron = "Darkness",
	DarkElven = "Prisms",
	Dwarven   = "Earth",
	Demonic   = "Fire",
	Elven     = "Water",
	Orcish    = "Combat", -- possibly temp
	Standard  = "Air",
}

local RACE_STATS =
{
	Orcish	  = "Agility",
	BlackIron = "Strength",
	DarkElven = "Intelligence",
	Dwarven   = "Stamina",
	Elven     = "Morale",
	Demonic   = "All",
}

local RARITY_JEWELLERY =
{
	Normal    = "Lesser",
	Magical   = "Lesser",
	Renowned  = "Normal",
	Master    = "Normal",
	Legendary = "Greater",
}

class "EquippableItem" (require("Items.Item"))

local ItemConfiguration = require("Items/ItemConfiguration")

EquippableItem.AttributeDescriptions = AttributeDescriptionList()
EquippableItem.AttributeDescriptions:AddAttribute('string', 'baseType', {serialize=1, default="DAGGER"})
EquippableItem.AttributeDescriptions:AddAttribute('string', 'race', {serialize=1, default="STANDARD"})
EquippableItem.AttributeDescriptions:AddAttribute('string', 'rarity', {serialize=1, default="NORMAL"})
EquippableItem.AttributeDescriptions:AddAttribute('string', 'unique', {serialize=1, default=""})
EquippableItem.IS_EQUIPPABLE = true

function EquippableItem:__init()
	super("ITEQ")
end

function EquippableItem:IsCarriedInHand()
	return self:IsWeapon() or self:IsShield() or self:IsPotion()
end

function EquippableItem:IsWeapon()
	return self:ApplyModifiers("IsWeapon", ITEM_BASE_TYPES_WEAPONS[self:GetAttribute("baseType")])
end

function EquippableItem:IsHelm()
	return self:ApplyModifiers("IsHelm", ITEM_BASE_TYPES_HELMS[self:GetAttribute("baseType")])
end

function EquippableItem:IsArmour()
	return self:ApplyModifiers("IsArmour", ITEM_BASE_TYPES_ARMOURS[self:GetAttribute("baseType")])
end

function EquippableItem:IsJewellery()
	return self:ApplyModifiers("IsJewellery", ITEM_BASE_TYPES_JEWELLERY[self:GetAttribute("baseType")])
end

function EquippableItem:IsAmulet()
	return self:ApplyModifiers("IsAmulet", ITEM_BASE_TYPES_AMULETS[self:GetAttribute("baseType")])
end

function EquippableItem:IsNecklace()
	return self:ApplyModifiers("IsNecklace", ITEM_BASE_TYPES_NECKLACES[self:GetAttribute("baseType")])
end

function EquippableItem:IsPendant()
	return self:ApplyModifiers("IsPendant", ITEM_BASE_TYPES_PENDANTS[self:GetAttribute("baseType")])
end

function EquippableItem:IsShoes()
	return self:ApplyModifiers("IsShoes", ITEM_BASE_TYPES_BOOTS[self:GetAttribute("baseType")])
end

function EquippableItem:IsShield()
	return self:ApplyModifiers("IsShield", ITEM_BASE_TYPES_SHIELDS[self:GetAttribute("baseType")])
end

function EquippableItem:IsPotion()
	return self:ApplyModifiers("IsPotion", ITEM_BASE_TYPES_POTIONS[self:GetAttribute("baseType")])
end

function EquippableItem:IsHealthPotion()
	return self:ApplyModifiers("IsHealthPotion", ITEM_BASE_TYPES_HEALTHPOTIONS[self:GetAttribute("baseType")])
end

function EquippableItem:IsManaPotion()
	return self:ApplyModifiers("IsManaPotion", ITEM_BASE_TYPES_MANAPOTIONS[self:GetAttribute("baseType")])
end

function EquippableItem:IsManaAccelerant()
	return self:ApplyModifiers("ManaAccelerant", ITEM_BASE_TYPES_MANAACCELERANTS[self:GetAttribute("baseType")])
end

function EquippableItem:IsPoison()
	return self:ApplyModifiers("IsPoison", ITEM_BASE_TYPES_POISONS[self:GetAttribute("baseType")])
end

function EquippableItem:IsCustomItem()
	return self:ApplyModifiers("IsCustomItem", false)
end

function EquippableItem:IsLootable()
    return self:GetBaseType() ~= "Custom"
end

function EquippableItem:IsUnavoidableToUse()
	return self:ApplyModifiers("IsUnavoidableToUse", false)
end

function EquippableItem:GetComponents()
	return self:ApplyModifiers("GetComponents", {hide=1, wood=2, metal=1})
end

function EquippableItem:InitItem(baseType, race, rarity)
	LOGF(string.format("InitItem(%s[%s], %s[%s], %s[%s])", tostring(baseType), type(baseType), tostring(race), type(race), tostring(rarity), type(rarity)))

	assert(type(baseType) == "string")
	assert(type(race) == "string")
	assert(type(rarity) == "string")
	
	self:SetAttribute("baseType", baseType)
	self:SetAttribute("race", race)
	self:SetAttribute("rarity", rarity)
	
	self.modifiers = {
		require("Items.BaseType" .. baseType),
		require("Items.Race" .. race),
		require("Items.Rarity" .. rarity),
		require(string.format("Items.%s.Rarity%s", race, rarity)),
	}
end

function EquippableItem:InitUniqueItem(modifier)
	self:SetAttribute("baseType", "Custom")
	self:SetAttribute("race", "Standard")
	self:SetAttribute("rarity", "Normal")
	self:SetAttribute("unique", modifier)
	
	self.modifiers = {
		require("Items.Custom" .. modifier),
	}
end

function EquippableItem:Fixup(savegame)
	if #self:GetAttribute("unique") > 0 then
		self.modifiers = {
			require("Items.Custom" .. self:GetAttribute("unique")),
		}
	else
		self.modifiers = {
			require("Items.BaseType" .. self:GetAttribute("baseType")),
			require("Items.Race" .. self:GetAttribute("race")),
			require("Items.Rarity" .. self:GetAttribute("rarity")),
			require(string.format("Items.%s.Rarity%s", self:GetAttribute("race"), self:GetAttribute("rarity"))),
		}
	end

	return Item.Fixup(self, savegame)
end

local function Mutate(mod, method, ...)
	if mod[method] then
		return mod[method](...)
	else
		return unpack({...}, 2)
	end
end

function EquippableItem:ApplyModifiers(named, ...)
	local mutated = {...}
	
	for i, mod in ipairs(self.modifiers) do
		mutated = {Mutate(mod, named, self, unpack(mutated))}
	end
	
	return unpack(mutated)
end

--[[ BASETYPE_SLOT_LIMITS
--   "either": 	single handed, can be held in either hand
--   "main": 	single handed, must be held in main hand
--   "off": 	off handed, must be held in off hand
--   "both": 	double handed, must be held in both hands
--   "armour":	assigned into armour slot
--   "helm":	assigned into helm slot
--   "boots":	assigned into boot slot
--   "misc":	assigned into jewellery slot
--]]

function EquippableItem:IsEquippableInMainHand()
	local slot_limit = ItemConfiguration.BASETYPE_SLOT_LIMITS[self:GetBaseType()]
	if slot_limit then
		if slot_limit == "main" or slot_limit == "either" or slot_limit == "both" then
			return true
		end
	else	-- custom item, has no restrictions
		return true
	end
end

function EquippableItem:IsEquippableInOffHand()
	local slot_limit = ItemConfiguration.BASETYPE_SLOT_LIMITS[self:GetBaseType()]
	if slot_limit then
		if slot_limit == "off" or slot_limit == "either" or slot_limit == "both" then
			return true
		end
	else	-- custom item, has no restrictions
		return true
	end
end

function EquippableItem:IsTwoHanded()
	local slot_limit = ItemConfiguration.BASETYPE_SLOT_LIMITS[self:GetBaseType()]
	if slot_limit then
		if slot_limit == "both" then
			return true
		end
	else	-- custom item, use function
		return self:ApplyModifiers("IsTwoHanded", false)
	end
end

function EquippableItem:GetIconAsset()
	return self:ApplyModifiers("EnumerateIcon", "img_item_missing")
end

function EquippableItem:GetIconOverlayAsset()
	return RARITY_OVERLAYS[self:ApplyModifiers("GetRarity", self:GetAttribute("rarity"))]
end

function EquippableItem:GetRarity()
	return self:ApplyModifiers("GetRarity", self:GetAttribute("rarity"))
end

function EquippableItem:GetNextRarity()
	local rarities = ITEM_RARITY
	local curr_rarity = self:GetRarity()
	local curr_rarity_num = 0
	for i=1, #rarities do
		if rarities[i] == curr_rarity then
			curr_rarity_num = i
			break
		end
	end
	return rarities[curr_rarity_num+1] -- will be nil if already at top level
end

function EquippableItem:GetRace()
	return self:ApplyModifiers("GetRace", self:GetAttribute("race"))
end

function EquippableItem:GetBaseType()
	return self:ApplyModifiers("GetBaseType", self:GetAttribute("baseType"))
end

function EquippableItem:GetBaseTypeName()
	return translate_text(string.format("[ITEM_BASE_TYPE_%s]", string.upper(self:GetBaseType())))
end

function EquippableItem:GetActivationSound()
	return self:ApplyModifiers("GetActivationSound", "atk_axe")
end

-- This returns the animation to use in the item activation combat
-- banner. Probably one of these:
--        img_combatbanner_sword (this is the default)
--        img_combatbanner_peck
--        img_combatbanner_bite
--        (etc...)
-- Each base type overrides this.  Each custom type should too.
function EquippableItem:GetActivationAnimation()
	return self:ApplyModifiers("GetActivationAnimation", "sword_attack")
end

function EquippableItem:GetActivationValue()
	return self:ApplyModifiers("MutateActivationValue", 0)
end

function EquippableItem:GetActivationTurns()
	return self:ApplyModifiers("MutateActivationTurns", 0)
end

function EquippableItem:GetGoldValue()
	local base = ItemConfiguration.BASETYPE_VALUES[self:GetBaseType()] or 0
	local race = ItemConfiguration.RACE_COSTS_MULTIPLIER[self:GetRace()] or 0
	local rare = ItemConfiguration.RARITY_COSTS_MULTIPLIER[self:GetRarity()] or 0
	local value = base * race * rare
	return math.floor(value)
end

function EquippableItem:GetSellValue()
	local value = self:GetGoldValue()
	return math.max(1, math.floor((0.25 * value) + 0.5))
end

function EquippableItem:GetActionPowerBase()
	return math.max(1, self:ApplyModifiers("MutateActionPowerBase", 0))
end

function EquippableItem:GetActionPower(hero)
	return self:GetActionPowerBase()
end            

function EquippableItem:GetDamageBase()
	return self:ApplyModifiers("MutateDamageBase", 0, 0)
end

function EquippableItem:GetAttackBonus()
	return self:ApplyModifiers("MutateAttackBonus", 0)
end

function EquippableItem:GetDefenseBonus()
	return self:ApplyModifiers("MutateDefenseBonus", 0)
end

function EquippableItem:GetWeaponCriticalBonus()
	return self:ApplyModifiers("MutateWeaponCriticalBonus", 0)
end

function EquippableItem:GetShieldCriticalBonus()
	return self:ApplyModifiers("MutateShieldCriticalBonus", 0)
end

function EquippableItem:GetBlockingEffectivenessBonus()
	return self:ApplyModifiers("MutateBlockingEffectivenessBonus", 0)
end

function EquippableItem:GetSpellResistanceBonus()
	return self:ApplyModifiers("MutateSpellResistanceBonus", 0)
end

function EquippableItem:GetSpellPenetrationBonus()
	return self:ApplyModifiers("MutateSpellPenetrationBonus", 0)
end

function EquippableItem:GetPassiveHealthBuff()
	return self:ApplyModifiers("GetPassiveHealthBuff", 0)
end

function EquippableItem:GetStartingManaBuff(pool)
	local functions = {
		green = "GetStartingGreenManaBuff",
		red = "GetStartingRedManaBuff",
		blue = "GetStartingBlueManaBuff",
		yellow = "GetStartingYellowManaBuff",
		black = "GetStartingBlackManaBuff",
		power = "GetStartingPowerManaBuff",
	}
	
	return self:ApplyModifiers(functions[pool], 0)
end

function EquippableItem:GetMaxManaBuff(pool)
	local functions = {
		green = "GetMaxGreenManaBuff",
		red = "GetMaxRedManaBuff",
		blue = "GetMaxBlueManaBuff",
		yellow = "GetMaxYellowManaBuff",
		black = "GetMaxBlackManaBuff",
	}
	
	return self:ApplyModifiers(functions[pool], 0)
end

function EquippableItem:GetStatBuff(stat)
	local functions = {
		strength = "GetStrengthBuff",
		agility  = "GetAgilityBuff",
		stamina = "GetStaminaBuff",
		intelligence = "GetIntelligenceBuff",
		morale = "GetMoraleBuff",
	}
	
	return self:ApplyModifiers(functions[stat], 0)
end

function EquippableItem:GetActiveDefenseBonus(hero)
	-- [SC] TODO: deprecated, heroes no longer get a shield bonus defence from any stat or status effect.
	--local shieldBonus = 0
	--if hero then
	--	shieldBonus = hero:GetShieldBonusDefence()
	--end
	return self:ApplyModifiers("MutateActiveDefenseBonus", 0, 0)
end

function EquippableItem:GetDamage(hero)
	local base = self:GetDamageBase()
	local bonus = hero:GetWeaponBonusDamage()
	--print(string.format("EquippableItem:GetDamage - base = %d  bonus = %d", base, bonus))
	return base + bonus
end

function EquippableItem:EnumerateMagicalEffects()
	return self:ApplyModifiers("EnumerateMagicalEffects", {}, self)
end

function EquippableItem:ApplyMagicalEffects(when, params, ...)
	for i, effect in ipairs(self:EnumerateMagicalEffects()) do
		if when == effect.when then
			self:ApplyEffect(effect, params, ...)
		end
	end
end

function EquippableItem:ApplyEffect(effect, params, ...)
	effect.method(self, effect, params, ...)
end

function EquippableItem:GetName()
	return self:ApplyModifiers("GetName", require("Items.ItemNameStitching"):GetItemName(self))
end

function EquippableItem:GetType()
	return self:ApplyModifiers("GetType", require("Items.ItemNameStitching"):GetItemType(self))
end

function EquippableItem:GetNameFormatString()
	local baseFormatString
	local rarity = self:ApplyModifiers("GetRarity", self:GetAttribute("rarity"))
	
	baseFormatString = string.format("{%03d,%03d,%03d,%03d}%%s{%03d,%03d,%03d,%03d}", unpack(RARITY_COLOURS[rarity]))
	
	return self:ApplyModifiers("GetNameFormatString", baseFormatString)
end

function EquippableItem:IsUseable(state)
	assert(state, "must pass hero state to IsUseable")
	return state.power >= self:GetActionPower(state.hero)
end

function EquippableItem:DebugPrint(hero)
	assert(hero)
	
	LOGF("%s: %s: %s", self:GetAttribute("baseType"), self:GetAttribute("race"), self:GetAttribute("rarity"))
	LOGF("   Name: %s", self:GetName())
	LOGF("   AP cost: %d (%d)", self:GetActionPowerBase(), self:GetActionPower(hero))
	if self:IsLootable() then
		LOGF("   Gold value: %d", self:GetGoldValue())
		LOGF("   Sell value: %d", self:GetSellValue())
	end
	if self:IsWeapon() then
		LOGF("   Damage: %d (%d)", self:GetDamageBase(), self:GetDamage(hero))
	elseif self:IsShield() then
		LOGF("   Active Defense: %d (%d)", self:GetActiveDefenseBonus(), self:GetActiveDefenseBonus(hero))
	elseif self:IsArmour() or self:IsHelm() or self:IsShoes() then
		LOGF("   Passive Defense: %d", self:GetDefenseBonus())
	end
	LOGF("   Icon: %s", self:GetIconAsset())
	LOGF("   Carried? %s", tostring(self:IsCarriedInHand()))
	LOGF("   Lootable? %s", tostring(self:IsLootable()))
	
	local effects = self:EnumerateMagicalEffects()
	
	if next(effects) then
		for index, effect in ipairs(effects) do
			LOGF("   Effect: %s (%s)", effect.name or "nil", effect.from or "nil")
		end
	end
end

local DEFAULT_WIDTH = 320
local DEFAULT_EXTRA_WIDTH = 16

function _G.GetQuestItemToolTipFormatAndData(item)
	local format = ""
	local data   = ""
	
	local item_data = require("QuestItems/"..tostring(item))
	
	local item_name = item_data.texttag
	local popup_width = math.max(DEFAULT_WIDTH, get_text_length("font_uitext", item_name)+DEFAULT_EXTRA_WIDTH+DEFAULT_EXTRA_WIDTH)
	
	format, data = TooltipHelpers.AddTitle(format, data, 0, "font_uitext", translate_text(string.format("[ITEM_QUEST_%s]", item_name)))
	format, data = TooltipHelpers.AddIcon(format, data, 0, 5, popup_width, 30, "img_TextBackground_Purple")
	format, data = TooltipHelpers.AddIcon(format, data, 13, 40, 64, 64, item_data.icon)
	
	format, data = TooltipHelpers.AddTextBlock(format, data, 88, 38, 180, 32, "font_tooltip", "left", "vcenter", "[QUESTITEM]")
	
	-- terminate the strings
	format = format .. "//"
	data   = data   .. "//"
	
	return format, data
end

function EquippableItem:GetToolTipFormatAndData(hero, final_string, use_buy_price)
	return self:GetToolTipFormatAndDataInternal({
		weaponBonusDamage = hero:GetWeaponBonusDamage(),
		shieldBonusDefense = 0, --shieldBonusDefense = hero:GetShieldBonusDefence(), -- [SC] TODO: deprecated, heroes no longer get a shield bonus defence from any stat or status effect.
		heroClass = hero:GetAttribute("class"),
	}, final_string, use_buy_price)
end

function EquippableItem:GetToolTipFormatAndDataWithParams(params, final_string, use_buy_price)
	return self:GetToolTipFormatAndDataInternal(params, final_string, use_buy_price)
end

function EquippableItem:GetToolTipFormatAndDataInternal(params, final_string, use_buy_price)
	local format = ""
	local data   = ""
	local curr_y = 31 -- for formatting the textbox
	
	
	local item_name = self:GetName()
	local popup_width = math.max(DEFAULT_WIDTH, get_text_length("font_uitext", item_name)+DEFAULT_EXTRA_WIDTH)
	
	-- Title
	format, data = TooltipHelpers.AddTitle(format, data, 0, "font_uitext", item_name)
	format, data = TooltipHelpers.AddIcon(format, data, 0, 5, popup_width, 30, "img_TextBackground_Purple")
	
	-- Icon
	format, data = TooltipHelpers.AddIcon(format, data, 13, 40, 64, 64, self:GetIconOverlayAsset())
	format, data = TooltipHelpers.AddIcon(format, data, 13, 40, 64, 64, self:GetIconAsset())
	
	-- Value
	if self:IsLootable() then	-- [NEOBEN] SCF, did you forget to add in this check, or did you mean for all items (inlcuding customs) to show a gold value?
		-- EquippableItem data shows shop price and items sell for 1/4 of the shop price. The inventory etc always show the
		-- value to the hero ie the sell price. So:
		local value = self:GetGoldValue()
		if not use_buy_price then 
			value = self:GetSellValue()
		end
		format, data = TooltipHelpers.AddIcon(format, data, 90, 71, 32, 32, "img_levelup_gold")
		format, data = TooltipHelpers.AddTextBlock(format, data, 128, 70, 180, 32, "font_uitext", "left", "vcenter", tostring(value))
	end
	
	-- Type 
	if self:IsWeapon() then
		if self:IsTwoHanded() then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[2-HANDED]")))
		else
			local main_hand = self:IsEquippableInMainHand()
			local off_hand = self:IsEquippableInOffHand()
			if main_hand and off_hand then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[1-HANDED]")))
			elseif main_hand then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[MAIN_HAND]")))
			elseif off_hand then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[OFF_HAND]")))
			end
		end
		format, data = self:GetToolTipFormatAndDataInternal_Weapon(format, data, params, popup_width)
		
	elseif self:IsCustomItem() then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[1-HANDED]")))
				format, data = self:GetToolTipFormatAndDataInternal_Custom(format, data, params, popup_width)
	elseif self:IsShield() then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[OFF_HAND]")))
				format, data = self:GetToolTipFormatAndDataInternal_Shield(format, data, params, popup_width)
	elseif self:IsArmour() then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[ARMOUR]")))
				format, data = self:GetToolTipFormatAndDataInternal_Armor(format, data, params, popup_width)
	elseif self:IsHelm() then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[HELM]")))
				format, data = self:GetToolTipFormatAndDataInternal_Armor(format, data, params, popup_width)
	elseif self:IsShoes() then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[SHOES]")))
				format, data = self:GetToolTipFormatAndDataInternal_Armor(format, data, params, popup_width)
	elseif self:IsPotion() then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[1-HANDED]")))
				format, data = self:GetToolTipFormatAndDataInternal_Potion(format, data, params, popup_width)
	elseif self:IsJewellery() then
				format, data = TooltipHelpers.AddTextBlock(format, data, 90, 38, 180, 32, "font_tooltip", "left", "vcenter", tostring(translate_text("[JEWELLERY]")))
				format, data = self:GetToolTipFormatAndDataInternal_Jewellery(format, data, params, popup_width)
	end


	if final_string and (format ~= "") then
		format = format .. "//"
		data   = data   .. "//"
	end
	
	
	return format, data, curr_y
end
	
	
	--[[
	
	format, data = TooltipHelpers.AddTitle(format, data, 0, "font_uitext", self:GetName())
	format, data = TooltipHelpers.AddIcon(format, data, 2, 5, 288, 24, "img_TextBackground_Purple")
	format, data = TooltipHelpers.AddIcon(format, data, 13, 33, 64, 64, self:GetIconOverlayAsset())
	format, data = TooltipHelpers.AddIcon(format, data, 13, 33, 64, 64, self:GetIconAsset())
	
	local function DisplayClassRestriction(y_pos)
		local restrict_string = ""
		local item_classes = ItemConfiguration.BASETYPE_CLASS_LIMITS[self:GetAttribute("baseType")]
		local restrict_equip = false
		if (type(item_classes) == "string") then
			if item_classes == "None" then
				restrict_string = substitute(translate_text("[TOOLTIP_CLASSES]"), translate_text("[ALL_CLASSES]"))
				restrict_equip = true
			else
				restrict_string = substitute(translate_text("[TOOLTIP_CLASSES]"), translate_text(string.format("[CLASS_%s]", string.upper(item_classes))))
				if params.heroClass then
					restrict_equip = (params.heroClass == item_classes)
				else
					restrict_equip = true
				end
			end
		elseif (type(item_classes) == "table") then
			if (#item_classes == 1) then
				local item_class = item_classes[1]
				restrict_string = substitute(translate_text("[TOOLTIP_CLASSES]"), translate_text(string.format("[CLASS_%s]", string.upper(item_class))))
				if params.heroClass then
					restrict_equip = (params.heroClass == item_class)
				else
					restrict_equip = true
				end
			else
				local valid_class = ""
				for i, class in ipairs(item_classes) do
					if i == 1 then
						valid_class = string.format("%s", translate_text(string.format("[%s_SHORT]", string.upper(class))))
					else
						valid_class = string.format("%s, %s", valid_class, translate_text(string.format("[%s_SHORT]", string.upper(class))))
					end
					if (params.heroClass == nil) or (params.heroClass == class) then
						restrict_equip = true
					end
				end
				restrict_string = substitute(translate_text("[TOOLTIP_CLASSES]"), valid_class)
			end
		else
			restrict_string = translate_text("[CUSTOM_ITEM]")
			restrict_equip = true
		end
		
		if not restrict_equip then
			restrict_string = string.format("{255,050,050,255}%s{255,255,255,255}", restrict_string)
		end
		local width = math.max(200, 20 + get_text_length("font_uitext", restrict_string))
		format, data = TooltipHelpers.AddTextBlock(format, data, 88, y_pos, width, 20, "font_uitext", "left", "top", restrict_string)
		y_pos = y_pos + 23
		
		return y_pos
	end
	
	-- class restrictions
	curr_y = DisplayClassRestriction(curr_y)

	if self:IsWeapon() then
		-- hands
		local hand_string = ""
		
		if self:IsTwoHanded() then
			hand_string = translate_text("[2-HANDED]")
			--format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", translate_text("[2-HANDED]"))
		else
			local main_hand = self:IsEquippableInMainHand()
			local off_hand = self:IsEquippableInOffHand()
			if main_hand and off_hand then
				hand_string = translate_text("[1-HANDED]")
				--format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", translate_text("[1-HANDED]"))
			elseif main_hand then
				hand_string = translate_text("[MAIN_HAND]")
				--format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", translate_text("[MAIN_HAND]"))
			elseif off_hand then
				hand_string = translate_text("[OFF_HAND]")
				--format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", translate_text("[OFF_HAND]"))
			end
		end
		
		if (self.equippable == -1) and (not self:CanClassEquip(params.heroClass)) then
			hand_string = string.format("{255,050,050,255}%s{255,255,255,255}", hand_string)
		end
		
		format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", hand_string)
		
		curr_y = curr_y + 23
	elseif self:IsCustomItem() then
		format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y - 23, 200, 20, "font_uitext", "left", "top", "[CUSTOM_ITEM]")
	elseif self:IsShield() then
		--
	elseif self:IsArmour() or self:IsHelm() or self:IsShoes() then
		-- slot
		if self:IsArmour() then
			format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", "[ARMOUR]")
		elseif self:IsHelm() then
			format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", "[HELM]")
		elseif self:IsShoes() then
			format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", "[SHOES]")
		end

		curr_y = curr_y + 23
		-- armour type
		local base = self:GetAttribute("baseType")
		local armour_type= ""
		local ARMOUR_TYPES =
		{
			Hide = "Hide",
			Cloth = "Cloth",
			PaddedCloth = "Cloth",
			Leather = "Leather",
			HardenedLeather = "HardenedLeather",
			Chainmail = "Chain",
			Scalemail = "Scale", 
			Plate = "Plate",
			Platemail = "Plate",
		}
		
		for material, filter in pairs(ARMOUR_TYPES) do
			if (base == string.format("%sRobes", material)) or
				(base == string.format("%sArmour", material)) or
				(base == string.format("%sBoots", material)) or
				(base == string.format("%sHelmet", material)) then
				armour_type = filter
				if (params.heroEquip) and (not params.heroEquip[armour_type]) then
					armour_type = string.format("{255,050,050,255}%s{255,255,255,255}", translate_text(string.format("[%s]", string.upper(armour_type))))
				else
					armour_type = translate_text(string.format("[%s]", string.upper(armour_type)))
				end
			end
		end
			
		format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", armour_type)
		
		curr_y = curr_y + 23
	elseif self:IsPotion() then
		format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", "[POTION]")
		curr_y = curr_y + 23
	elseif self:IsJewellery() then
		format, data = TooltipHelpers.AddTextBlock(format, data, 88, curr_y, 200, 20, "font_uitext", "left", "top", "[JEWELLERY]")
		curr_y = curr_y + 23
	end	
	if self:IsLootable() then
		format, data = TooltipHelpers.AddIcon(format, data, 8, 100, 26, 26, "img_levelup_gold")
		format, data = TooltipHelpers.AddTextBlock(format, data, 39, 96, 100, 26, "font_uitext", "left", "vcenter", tostring(self:GetGoldValue()))
	end

	if self:IsWeapon() then
		format, data = TooltipHelpers.AddIcon(format, data, 6, 125, 32, 32, "img_gem_power_32")
		format, data = TooltipHelpers.AddTextBlock(format, data, 45, 128, 32, 24, "font_uitext", "left", "vcenter", tostring(self:GetActionPower()))
		format, data = TooltipHelpers.AddIcon(format, data, 180, 125, 32, 32, "img_gem_skull_32")
		format, data = TooltipHelpers.AddTextBlock(format, data, 219, 128, 32, 24, "font_uitext", "left", "vcenter", tostring(self:GetDamageBase()+params.weaponBonusDamage))
		
		format, data = TooltipHelpers.AddIcon(format, data, 2, 129, 288, 24, "img_TextBackground_Purple")
	elseif self:IsCustomItem() then
		format, data = TooltipHelpers.AddIcon(format, data, 6, 125, 32, 32, "img_gem_power_32")
		format, data = TooltipHelpers.AddTextBlock(format, data, 45, 128, 32, 24, "font_uitext", "left", "vcenter", tostring(self:GetActionPower()))
		
		format, data = TooltipHelpers.AddIcon(format, data, 2, 129, 288, 24, "img_TextBackground_Purple")
	elseif self:IsShield() then
		local amount, turns = self:GetActiveDefenseBonus()
		format, data = TooltipHelpers.AddTextBlock(format, data, 45, 128, 32, 24, "font_uitext", "left", "vcenter", tostring(self:GetActionPower()))
		format, data = TooltipHelpers.AddIcon(format, data, 6, 125, 32, 32, "img_gem_power_32")
		format, data = TooltipHelpers.AddTextBlock(format, data, 139, 128, 32, 24, "font_uitext", "left", "vcenter", tostring(turns))
		format, data = TooltipHelpers.AddIcon(format, data, 100, 125, 32, 32, "img_hour_glass")
		format, data = TooltipHelpers.AddTextBlock(format, data, 223, 122, 32, 36, "font_uitext", "center", "vcenter", tostring(amount+params.shieldBonusDefense))
		format, data = TooltipHelpers.AddIcon(format, data, 194, 122, 32, 36, "img_DefenseRating_Icon")
		
		
		format, data = TooltipHelpers.AddIcon(format, data, 2, 129, 288, 24, "img_TextBackground_Purple")
	elseif self:IsArmour() or self:IsHelm() or self:IsShoes() then
		format, data = TooltipHelpers.AddTextBlock(format, data, 136, 123, 32, 32, "font_uitext", "left", "vcenter", tostring(self:GetDefenseBonus()))
		format, data = TooltipHelpers.AddIcon(format, data, 100, 122, 32, 36, "img_DefenseRating_Icon")
		
		format, data = TooltipHelpers.AddIcon(format, data, 2, 129, 288, 24, "img_TextBackground_Purple")
	elseif self:IsPotion() then
		format, data = TooltipHelpers.AddIcon(format, data, 6, 125, 32, 32, "img_gem_power_32")
		format, data = TooltipHelpers.AddTextBlock(format, data, 45, 128, 32, 24, "font_uitext", "left", "vcenter", tostring(self:GetActionPower()))
		
		local amount, turns = self:GetBonus()
		if amount ~= 0 then
			format, data = TooltipHelpers.AddIcon(format, data, 180, 127, 27, 27, self:GetBonusIcon())
			format, data = TooltipHelpers.AddTextBlock(format, data, 219, 128, 32, 24, "font_uitext", "left", "vcenter", tostring(self:GetBonus()))
		end
		
		if turns ~= 0 then
			format, data = TooltipHelpers.AddTextBlock(format, data, 139, 128, 32, 24, "font_uitext", "left", "vcenter", tostring(turns))
			format, data = TooltipHelpers.AddIcon(format, data, 100, 125, 32, 32, "img_hour_glass")
		end
		format, data = TooltipHelpers.AddIcon(format, data, 2, 129, 288, 24, "img_TextBackground_Purple")
	end
	
	local effects = table.join(self:EnumerateMagicalEffects(), self:EnumerateTooltipEffects())
	local effectsCategories = { }
	for _, effect in ipairs(effects) do
		if effect.category then
			if not effectsCategories[effect.category] then
				effectsCategories[effect.category] = { }
			end
			table.insert(effectsCategories[effect.category], effect)
		else -- a safety clause to display unhandled effects under 'No Category, No Effect'
			if not effectsCategories["No Category"] then
				effectsCategories["No Category"] = { }
			end
			table.insert(effectsCategories["No Category"], effect)
		end
	end
	
	if self:IsPotion() then
		curr_y = 159
	else
		curr_y = 156
	end
	
	local function DisplayEffectsTable(tbl, name)
		if #tbl > 0 then
			if not self:IsPotion() then -- potions do not show effect category names as they only ever have 'On Strike' (Use)
				format, data = TooltipHelpers.AddTextBlock(format, data, 0, curr_y, 264, 24, "font_tooltip", "center", "vcenter", name)
				format, data = TooltipHelpers.AddIcon(format, data, 2, curr_y, 288, 24, "img_TextBackground_Red")
				curr_y = curr_y + 33
			end
			
			for _, effect in ipairs(tbl) do
				format, data = TooltipHelpers.AddIcon(format, data, 14, curr_y, 32, 32, effect.category_icon or "img_missing")
				if effect.category_name then
					local effectString = string.format("%s: {150,150,150,255}%s{255,255,255,255}", translate_text(effect.category_title or "[BLANK]"), effect.category_name)
					local effectString_height = WidgetHelpers.get_string_height(effectString, "font_tooltip", 207)
					format, data = TooltipHelpers.AddFormattedHelpBlock(format, data, 51, curr_y, "font_tooltip", "left", "vcenter", effectString)
					curr_y = curr_y + effectString_height + 5
				else
					format, data = TooltipHelpers.AddFormattedHelpBlock(format, data, 51, curr_y, "font_tooltip", "left", "vcenter", "No effect")
					curr_y = curr_y + 45
				end
			end
		end
	end
	
	for cat_name, effects_list in pairs(effectsCategories) do
		DisplayEffectsTable(effects_list, cat_name)
	end
	
	if final_string and (format ~= "") then
		format = format .. "//"
		data   = data   .. "//"
	end
	
	
	return format, data, curr_y
end
	]]

function EquippableItem:GetToolTipFormatAndDataInternal_SpecialAbility(format, data, params, popup_width)

	local myEffects = self:EnumerateMagicalEffects()
	
	if myEffects and #myEffects > 0 then

		format, data = TooltipHelpers.AddTitle(format, data,   180, "font_uitext", translate_text("[ABILITY_SECTION]"))
		format, data = TooltipHelpers.AddIcon(format, data, 0, 185, popup_width, 30, "img_TextBackground_Red")
		
		-- Table of effects
		for i, effect in ipairs(myEffects) do
			format, data = TooltipHelpers.AddFormattedHelpBlock(format, data, 0, "auto", "font_uitext", "center", "vcenter", effect.category_name)
		end
		
	end
	
	return format, data
end
function EquippableItem:GetToolTipFormatAndDataInternal_ClassRestriction(format, data, params, popup_width)

	local restr_string
	
	local item_classes = ItemConfiguration.BASETYPE_CLASS_LIMITS[self:GetAttribute("baseType")]
	local restrict_equip = false
	if (type(item_classes) == "string") then
		if item_classes == "None" then
			restr_string = nil
		else
			local text_tag = string.format("[CLASS_RESTRICTION_%s]",string.upper(item_classes))
			restr_string = translate_text(text_tag)
		end
	end
	
	if restr_string then
		format, data = TooltipHelpers.AddFormattedHelpBlock(format, data, 0, "auto", "font_tooltip", "center", "vcenter", restr_string)	
	end
		
	return format, data
end

function EquippableItem:GetToolTipFormatAndDataInternal_Custom(format, data, params, popup_width)

	-- Section title
	format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[WEAPON_SECTION]"))
	format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
	
	-- AP & Dmg
	local ap  = self:GetActionPower()
	local dmg = self:GetActivationValue()
	
	if dmg > 0 then
		local x1 = math.floor(popup_width/6)
		local x2 = math.floor(4*popup_width/6)
		format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_action")
		format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(ap))
		format, data = TooltipHelpers.AddIcon(format, data, x2, 146, 32, 32, "img_popup_damage")
		format, data = TooltipHelpers.AddTextBlock(format, data, x2+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(dmg))
	else
		local x1 = math.floor(5*popup_width/12)
		format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_action")
		format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(ap))
	end
	
	-- Special abilities
	format,data = self:GetToolTipFormatAndDataInternal_SpecialAbility(format, data, params, popup_width)
	format,data = self:GetToolTipFormatAndDataInternal_ClassRestriction(format, data, params, popup_width)
	
	return format, data
end

function EquippableItem:GetToolTipFormatAndDataInternal_Weapon(format, data, params, popup_width)

	-- Section title
	format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[WEAPON_SECTION]"))
	format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
	
	-- AP & Dmg
	local ap  = self:GetActionPower()
	local dmg = self:GetDamageBase()+params.weaponBonusDamage
	local x1 = math.floor(popup_width/6)
	local x2 = math.floor(4*popup_width/6)
	format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_action")
	format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(ap))
	format, data = TooltipHelpers.AddIcon(format, data, x2, 146, 32, 32, "img_popup_damage")
	format, data = TooltipHelpers.AddTextBlock(format, data, x2+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(dmg))
	
	-- Special abilities
	format,data = self:GetToolTipFormatAndDataInternal_SpecialAbility(format, data, params, popup_width)
	format,data = self:GetToolTipFormatAndDataInternal_ClassRestriction(format, data, params, popup_width)
	
	return format, data
end

function EquippableItem:GetToolTipFormatAndDataInternal_Shield(format, data, params, popup_width)

	-- Section title
	format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[SHIELD_SECTION]"))
	format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
	
	-- AP & Dmg
	local ap  = self:GetActionPower()
	local def, turns = self:GetActiveDefenseBonus()
	local x1 = math.floor(popup_width/10)
	local x2 = math.floor(4*popup_width/10)
	local x3 = math.floor(7*popup_width/10)
	format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_action")
	format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(ap))
	format, data = TooltipHelpers.AddIcon(format, data, x2, 146, 32, 32, "img_spellpopup_time")
	format, data = TooltipHelpers.AddTextBlock(format, data, x2+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(turns))
	format, data = TooltipHelpers.AddIcon(format, data, x3, 146, 32, 32, "img_popup_defense")
	format, data = TooltipHelpers.AddTextBlock(format, data, x3+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(def+params.shieldBonusDefense))
	
	-- Special abilities
	format,data = self:GetToolTipFormatAndDataInternal_SpecialAbility(format, data, params, popup_width)
	format,data = self:GetToolTipFormatAndDataInternal_ClassRestriction(format, data, params, popup_width)
	
	return format, data
end

function EquippableItem:GetToolTipFormatAndDataInternal_Armor(format, data, params, popup_width)

	-- Section title
	format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[ARMOR_SECTION]"))
	format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
	
	-- Def
	local def = self:GetDefenseBonus()
	local x1 = math.floor(popup_width/2 - 32)
	format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_defense")
	format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(def))
	
	-- Special abilities
	format,data = self:GetToolTipFormatAndDataInternal_SpecialAbility(format, data, params, popup_width)
	format,data = self:GetToolTipFormatAndDataInternal_ClassRestriction(format, data, params, popup_width)
	
	return format, data
end

function EquippableItem:GetToolTipFormatAndDataInternal_Potion(format, data, params, popup_width)
	if self:IsPoison() then
		-- Section title
		format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[POISON_SECTION]"))
		format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
		
		local ap  = self:GetActionPower()
		local dmg = self:GetActivationValue()
		local turns = self:GetActivationTurns()
		local x1 = math.floor(popup_width/10)
		local x2 = math.floor(4*popup_width/10)
		local x3 = math.floor(7*popup_width/10)
		format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_action")
		format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(ap))
		format, data = TooltipHelpers.AddIcon(format, data, x2, 146, 32, 32, "img_spellpopup_time")
		format, data = TooltipHelpers.AddTextBlock(format, data, x2+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(turns))
		format, data = TooltipHelpers.AddIcon(format, data, x3, 146, 32, 32, "img_popup_damage")
		format, data = TooltipHelpers.AddTextBlock(format, data, x3+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(dmg))
		
	elseif self:IsHealthPotion() then
		-- Section title
		format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[POTION]"))
		format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
		
		local ap  = self:GetActionPower()
		local heal = self:GetActivationValue()
		local x1 = math.floor(popup_width/6)
		local x2 = math.floor(4*popup_width/6)
		format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_action")
		format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(ap))
		format, data = TooltipHelpers.AddIcon(format, data, x2, 146, 32, 32, "img_popup_life")
		format, data = TooltipHelpers.AddTextBlock(format, data, x2+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(heal))
		
	elseif self:IsManaPotion() then
		-- Section title
		format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[POTION]"))
		format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
		
		local ap  = self:GetActionPower()
		local mana = self:GetActivationValue()
		local x1 = math.floor(popup_width/6)
		local x2 = math.floor(4*popup_width/6)
		format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_action")
		format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(ap))
		format, data = TooltipHelpers.AddIcon(format, data, x2, 146, 32, 32, self:GetBonusIcon())
		format, data = TooltipHelpers.AddTextBlock(format, data, x2+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(mana))
		
	elseif self:IsManaAccelerant() then
		-- Section title
		format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[POTION]"))
		format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
		
		local ap  = self:GetActionPower()
		local mana = self:GetActivationValue()
		local turns = self:GetActivationTurns()
		local x1 = math.floor(popup_width/10)
		local x2 = math.floor(4*popup_width/10)
		local x3 = math.floor(7*popup_width/10)
		format, data = TooltipHelpers.AddIcon(format, data, x1, 146, 32, 32, "img_popup_action")
		format, data = TooltipHelpers.AddTextBlock(format, data, x1+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(ap))
		format, data = TooltipHelpers.AddIcon(format, data, x2, 146, 32, 32, "img_spellpopup_time")
		format, data = TooltipHelpers.AddTextBlock(format, data, x2+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(turns))
		format, data = TooltipHelpers.AddIcon(format, data, x3, 146, 32, 32, self:GetBonusIcon())
		format, data = TooltipHelpers.AddTextBlock(format, data, x3+36, 143, 64, 32, "font_uitext", "left", "vcenter", tostring(mana))
	end
	
	-- Special abilities
	format,data = self:GetToolTipFormatAndDataInternal_SpecialAbility(format, data, params, popup_width)
	format,data = self:GetToolTipFormatAndDataInternal_ClassRestriction(format, data, params, popup_width)
	
	return format, data
end

function EquippableItem:GetToolTipFormatAndDataInternal_Jewellery(format, data, params, popup_width)
	-- List all passive powers
	local tEffects = self:EnumerateTooltipEffects()
	
	if type(tEffects) == "table" then
		table.log(tEffects)
	end
	
	local y = 110
	
	if tEffects and #tEffects > 0 then
		-- Section title
		format, data = TooltipHelpers.AddTitle(format, data,   105, "font_uitext", translate_text("[JEWELLERY_SECTION]"))
		format, data = TooltipHelpers.AddIcon(format, data, 0, 110, popup_width, 30, "img_TextBackground_Red")
		
		-- Table of effects
		for i, effect in ipairs(tEffects) do
			y = y + 30
			format, data = TooltipHelpers.AddFormattedHelpBlock(format, data, 0, y, "font_uitext", "center", "vcenter", effect.category_name)
			
			LOGF("~ category_name: %s", effect.category_name)
		end
		
		y = y + 35
	end
	
	-- Special abilities (NOTE: no class restrictions on jewellery)
	local mEffects = self:EnumerateMagicalEffects()
	
	if type(mEffects) == "table" then
		table.log(mEffects)
	end
	
	if mEffects and #mEffects > 0 then
		format, data = TooltipHelpers.AddTitle(format, data,   y-5, "font_uitext", translate_text("[ABILITY_SECTION]"))
		format, data = TooltipHelpers.AddIcon(format, data, 0, y, popup_width, 30, "img_TextBackground_Red")
		
		-- Table of effects
		for i, effect in ipairs(mEffects) do
			format, data = TooltipHelpers.AddFormattedHelpBlock(format, data, 0, "auto", "font_uitext", "center", "vcenter", effect.category_name)
		end
	end
	
	return format, data
end

function EquippableItem:IsStealthy()
	return self:ApplyModifiers("IsStealthy", false)
end

function EquippableItem:IsUsableOnlyInStealth()
	return self:ApplyModifiers("IsUsableOnlyInStealth", false)
end

function EquippableItem:CanClassEquip(class)
	assert(class and (type(class) == "string"), "You must provide a class as a string")
	local baseType = self:GetBaseType()
	
	local restrictions = ItemConfiguration.BASETYPE_CLASS_LIMITS[baseType]
	
	if baseType ~= "Custom" then
		assert(restrictions, string.format("No data exists for baseType: %s",baseType))
	end
	
	if type(restrictions) == "table" then
		for _, classRestriction in ipairs(restrictions) do
			if (classRestriction == class) then
				return true
			end
		end
	elseif (restrictions == "None") or (restrictions == class) then -- assume its a string
		return true
	elseif not restrictions then -- this was a custom base type, custom base types have no class restrictions
		return true
	end
	
	return false
end

function EquippableItem:GetEquipCommands(hero, additional_callback)
	if hero:IsItemEquipped(self) then
		return { { tooltip="[UNEQUIP]", callback = function(params)
			SoundSystem.Play("pq2audio/ui/invswap")
			hero:UnequipItem(params.item)
			additional_callback()
		end } }
	end
	if self:IsCarriedInHand() then
		if self:IsTwoHanded() then
			return { { tooltip="[EQUIP_TWOHANDS]", callback = function(params)
				if params.item.equippable == 1 then
					SoundSystem.Play("pq2audio/ui/invswap")
					hero:EquipItem(params.item, "bothHands")
					additional_callback()
				elseif params.item.equippable == -1 then
					open_message_menu("[EQUIP_FAIL_HEAD]", "[EQUIP_FAIL_DESC_UNEQUIPPABLE]", function() end)
				end
			end} }
		else
			local return_table = { }
			if self:IsEquippableInMainHand() then
				table.insert(return_table, { tooltip = "[EQUIP_MAIN]", callback = function(params)
					if params.item.equippable == 1 then
						SoundSystem.Play("pq2audio/ui/invswap")
						hero:EquipItem(params.item, "mainHand")
						additional_callback()
					elseif params.item.equippable == -1 then
					open_message_menu("[EQUIP_FAIL_HEAD]", "[EQUIP_FAIL_DESC_UNEQUIPPABLE]", function() end)
				end
				end })
			end
			if self:IsEquippableInOffHand() then
				table.insert(return_table, { tooltip="[EQUIP_OFFHAND]", callback=function(params)
					if params.item.equippable == 1 then
						SoundSystem.Play("pq2audio/ui/invswap")
						hero:EquipItem(params.item, "offHand")
						additional_callback()
					elseif params.item.equippable == -1 then
						open_message_menu("[EQUIP_FAIL_HEAD]", "[EQUIP_FAIL_DESC_UNEQUIPPABLE]", function() end)
					end
				end })
			end
			return return_table
		end
	elseif self:IsHelm() then
		return {{ tooltip = "[EQUIP_HELM]", callback = function(params)
			if params.item.equippable == 1 then
				SoundSystem.Play("pq2audio/ui/invswap")
				hero:EquipItem(params.item, "helm")
				additional_callback()
			elseif params.item.equippable == -1 then
				open_message_menu("[EQUIP_FAIL_HEAD]", "[EQUIP_FAIL_DESC_UNEQUIPPABLE]", function() end)
			end
		end }}
	elseif self:IsArmour() then
		return {{ tooltip = "[EQUIP_ARMOUR]", callback = function(params)
			if params.item.equippable == 1 then
				SoundSystem.Play("pq2audio/ui/invswap")
				hero:EquipItem(params.item, "armour")
				additional_callback()
			elseif params.item.equippable == -1 then
				open_message_menu("[EQUIP_FAIL_HEAD]", "[EQUIP_FAIL_DESC_UNEQUIPPABLE]", function() end)
			end
		end }}
	elseif self:IsShoes() then
		return {{ tooltip = "[EQUIP_SHOES]", callback = function(params)
			if params.item.equippable == 1 then
				SoundSystem.Play("pq2audio/ui/invswap")
				hero:EquipItem(params.item, "feet")
				additional_callback()
			elseif params.item.equippable == -1 then
				open_message_menu("[EQUIP_FAIL_HEAD]", "[EQUIP_FAIL_DESC_UNEQUIPPABLE]", function() end)
			end
		end }}
	elseif self:IsJewellery() then
		return {{ tooltip = "[EQUIP_MISC]", callback = function(params)
			if params.item.equippable == 1 then
				SoundSystem.Play("pq2audio/ui/invswap")
				hero:EquipItem(params.item, "jewellery")
				additional_callback()
			elseif params.item.equippable == -1 then
				open_message_menu("[EQUIP_FAIL_HEAD]", "[EQUIP_FAIL_DESC_UNEQUIPPABLE]", function() end)
			end
		end }}
	end
	error("Unequippable Item")
end

function EquippableItem:GetBonusIcon()
	-- gets the icon of the effect of this potion
	return self:ApplyModifiers("GetBonusIcon", "img_missing")
end

function EquippableItem:EnumerateTooltipEffects()
	return self:ApplyModifiers("EnumerateTooltipEffects", {}, self)
end

-- Any platform specific stuff in here
dofile("Assets/Scripts/Items/EquippableItemPlatform.lua")

return EquippableItem
