-- The BASH mini game

use_safeglobals()

local Grid = require("Battlegrounds/Grid")
local ProbabilityTable = require("Utils/ProbabilityTable")
local Constants = require("Battlegrounds/BattleGroundConstants")
local SpinHelpers = require("Utils/SpinHelpers")
local PuzzleHelpers = require("Puzzles/PuzzleHelpers")
local Configuration = require("Puzzles/PuzzleConfiguration").BASH

class "Bash" (GameObject)

require("Utils/DraggingSupport"):AddToClass(Bash)

Bash.AttributeDescriptions = AttributeDescriptionList()

local SHAKE = {
	"screenShake_small",
	"screenShake_medium",
	"screenShake_large",
	"screenShake_huge"
}

local LARGE_GEM_SIZE 	= Constants.PUZZLE_GEM_SIZE
local LARGE_GEM_SCALE	= Constants.PUZZLE_GEM_SCALE
local MATCH4_TURN_BONUS = Configuration.MATCH4_TURN_BONUS
local MATCH5_TURN_BONUS = Configuration.MATCH5_TURN_BONUS
local DIFFICULTY_SETTINGS = Configuration.DIFFICULTY_SETTINGS

local screen = SCREENS.BashGameMenu

local function EndMinigame(endMinigame, result)
	if endMinigame then
		CallScreenSequencer("BashGameMenu", function() endMinigame(result) end)
	else
		CallScreenSequencer("BashGameMenu", "MinigameSelectionMenu")
	end
end

function Bash:__init(clid)
	super("PBSH")
	
	self.max_turns = 20
	self.gem_step = Constants.PUZZLE_GEM_STEP
	self.isMoving = false	-- this will be set to true the very instance a move or spell is approved, and unset when the board is in a steady state
	
	self.hint1 = GameObjectViewManager:Construct("GraphicAsset", "anim_hint")
	self.hint1:SetScale(LARGE_GEM_SCALE)
	self.hint1:SetSortingValue(Constants.HINT_ARROW_SORTING_VALUE)
	self.hint1:SetAlpha(Constants.HINT_ARROW_ALPHA_VALUE)
	
	self.hint2 = GameObjectViewManager:Construct("GraphicAsset", "anim_hint")
	self.hint2:SetScale(LARGE_GEM_SCALE)
	self.hint2:SetSortingValue(Constants.HINT_ARROW_SORTING_VALUE)
	self.hint2:SetAlpha(Constants.HINT_ARROW_ALPHA_VALUE)
	
	self.select1 = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_g")
	self.select1:SetScale(LARGE_GEM_SCALE)
	self.select1:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self.select2 = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_g")
	self.select2:SetScale(LARGE_GEM_SCALE)
	self.select2:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)

	self.selectFast = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_b")
	self.selectFast:SetScale(LARGE_GEM_SCALE)
	self.selectFast:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self:SetCursorInteract(false)
	self:OnDragInit(Constants.GEM_SELECTION_SORTING_VALUE, LARGE_GEM_SCALE)
end

function Bash:PreDestroy()
	if leakdetect then
		leakdetect.PreDestroyHook(self)
	end
	scheduler:StopAllThreadsForObject(self)
end

function Bash:GemCoordToPixel(x, y)
	return math.round((x - 1) * self.gem_step + (self.gem_step / 2)), math.round(y * self.gem_step - (self.gem_step / 2))
end

function Bash:GemPixelToCoord(px, py)
	return math.round((px - (self.gem_step / 2)) / self.gem_step + 1), math.round((py + (self.gem_step / 2)) / self.gem_step)
end

function Bash:Initialise(layout, turnsBonus, gemProbabilityTable, difficultyLevel, endMinigame)
	assert(layout)
	assert(gemProbabilityTable)
	assert(difficultyLevel)
	assert(turnsBonus)
	
	self.width = layout.grid_width
	self.height = layout.grid_height
	self.gemProbabilityTable = gemProbabilityTable
	self.state = table.clone(DIFFICULTY_SETTINGS[difficultyLevel])
	self.endMinigame = endMinigame
	
	self.health_max = self.state.health_left
	screen:UpdateHealth(self.state.health_left, self.health_max)
	
	self.max_turns = self.max_turns + turnsBonus
	self:AddToTurnsLeft(0)
	
	self.grid = Grid:Construct(self.width, self.height, self.grid)
	
	for row = self.height, 1, -1 do
		if layout[row] then
			local y = (1 + self.height) - row
			for x = 1, self.width do
				local gem = layout[row][x]
				if gem and gem ~= " " then
					self.grid[x][y] = self:CreateGem(Gems:GetNameFromSymbol(gem), x, y, true)
				end
			end
		end
	end
	
	PuzzleHelpers.FillGems(self)
	
	self:SetupHintAndSelection()
	
	scheduler:AddChildThread(self, function ()
		scheduler:Sleep(1000)
		SCREENS.EffectsMenu:PlayDefaultSpellTextEffect("[BASH_TEXT]")
	end, "Bash:Initialise")
end

function Bash:RefreshUI()
	screen:UpdateHealth(self.state.health_left, self.health_max)
	screen:UpdateUI(self.turns_left, self.max_turns)
end

function Bash:CreateGem(name, x, y, visible)
	assert(x)
	assert(y)
	
	if not name then
		name = self.gemProbabilityTable:RandomItem()
	end
	
	if name == "Wildcard" then
		if not self.wildcardProbabilityTable then
			self.wildcardProbabilityTable = ProbabilityTable:Construct(Constants.WILDCARD_DIST)
		end
		name = "Wildcard" .. self.wildcardProbabilityTable:RandomItem()
	end
	
	local gem = GameObjectManager:ConstructLocal("BGEM")
	gem:SetInfo(Gems[name], LARGE_GEM_SCALE, LARGE_GEM_SIZE)
	gem:SetPos(self:GemCoordToPixel(x, y))
	gem.x = x
	gem.y = y
	
	if not visible then
		gem:GetView():SetAlpha(0.0)
	end
	
	self:AddChild(gem)
	
	return gem
end

function Bash:DestroyGem(gem)
	if gem then
		GameObjectManager:Destroy(gem)
	end
end

function Bash:DestroyGemAt(x, y)
	assert(x)
	assert(y)
	
	self:DestroyGem(self.grid[x][y])
	self.grid[x][y] = nil
end

function Bash:Swap(x1, y1, x2, y2)
	assert(x1)
	assert(y1)
	assert(x2)
	assert(y2)
	
	LOGF("Bash:Swap(%d,%d,%d,%d)", x1, y1, x2, y2)
	
	local valid, reason = self.grid:IsValidMove(x1, y1, x2, y2)
	if not valid then
		LOGF("Rejected move %d,%d <-> %d,%d because: %s", x1, y1, x2, y2, reason)
		self:HandleSwapRejection(x1, y1, x2, y2)
		return
	end
	
	self:AddToTurnsLeft(-1)
	
	SoundSystem.Play("pq2audio/board/gemmove")
	self:SwapGems(x1, y1, x2, y2)
	
	-- wait for half a second (this is how long it takes the views to swap to gems) before cascading
	local checkForMatches = GameEventManager:Construct("CheckForMatches")
	checkForMatches:SetAttribute("count", 1)
	GameEventManager:SendDelayed(checkForMatches, self, GetGameTime() + Constants.PAUSE_AFTER_MOVE)
end

function Bash:SendSwapRequest(x1, y1, x2, y2)
	return self:Swap(x1, y1, x2, y2)
end

function Bash:HandleSwapRejection(x1, y1, x2, y2)
	self:SwapGems(x1, y1, x2, y2)
	
	SoundSystem.Play("pq2audio/ui/illegal")
	local px1, py1 = self:GemCoordToPixel(x1, y1)
	local px2, py2 = self:GemCoordToPixel(x2, y2)
	local x = (px1 + px2) / 2
	local y = (py1 + py2) / 2
	SCREENS.EffectsMenu:PlayFallingTextEffectAt(x, y, "[ILLEGAL_MOVE]")
	
	-- Then send another event to make sure they swap back
	local event = GameEventManager:Construct("SwapBack")
	event:SetAttribute("source_x", x2)
	event:SetAttribute("source_y", y2)
	event:SetAttribute("target_x", x1)
	event:SetAttribute("target_y", y1)
	GameEventManager:SendDelayed(event, self, GetGameTime() + Constants.GEM_SWAP_TIME)
end

function Bash:OnEventSwapBack(event)
	if not self.swappedBack then
		self.swappedBack = true
		local x1, y1 = event:GetAttribute("source_x"), event:GetAttribute("source_y")
		local x2, y2 = event:GetAttribute("target_x"), event:GetAttribute("target_y")
		self:SwapGems(x1, y1, x2, y2)
		GameEventManager:SendDelayed(event, self, GetGameTime() + 1.1 * Constants.GEM_SWAP_TIME)
	else
		self.swappedBack = false
		self.isMoving = false
		self:SetupHintAndSelection()
	end
end

function Bash:SwapGems(x1, y1, x2, y2)
	self.isMoving = true
	
	-- clear old overlays
	self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
	self.selectGemX, self.selectGemY = nil
	
	if self.targetGemX and self.targetGemY then
		self:ClearSelectedGemOverlay(self.targetGemX, self.targetGemY)
		self.targetGemX, self.targetGemY = nil
	end
	
	local gem1 = self.grid[x1][y1]
	local gem2 = self.grid[x2][y2]
	
	-- set new overlays
	if gem1 then
		self:SetSelectedGemOverlay(x1, y1, true, true)
	end
	if gem2 then
		self:SetSelectedGemOverlay(x2, y2, false, true)
	end
	
	if gem1 then
		self:MoveGemTo(gem1, x2, y2)
	end
	if gem2 then
		self:MoveGemTo(gem2, x1, y1)
	end
end

function Bash:MoveGemTo(gem, target_x, target_y)
	local target_x_pix, target_y_pix = self:GemCoordToPixel(target_x, target_y)
	local x_pix, y_pix = gem:GetX(), gem:GetY()
	
	gem.x = assert(target_x)
	gem.y = assert(target_y)
	
	self.grid[target_x][target_y] = gem
	
	local moveToTarget = MovementManager:Construct("Smooth")
	
	moveToTarget:SetAttribute("Duration", Constants.GEM_SWAP_TIME)
	moveToTarget:SetAttribute("StartX", x_pix)
	moveToTarget:SetAttribute("StartY", y_pix)
	moveToTarget:SetAttribute("EndX", target_x_pix)
	moveToTarget:SetAttribute("EndY", target_y_pix)
	
	gem:SetMovementController(moveToTarget)
end

function Bash:HandleOfAKind(match)
	SoundSystem.Play("pq2audio/board/fx_extraturn")
	local gx, gy = SCREENS.EffectsMenu:GetAverageGridPos(match.gems)
	local x, y = self:GemCoordToPixel(gx, gy)
	
	if match.count == 4 then	-- 4 of a kind - add 5 bonus seconds
		SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[OFKIND_FOUR]", "[OFKIND_TURN1]")
		self:AddToTurnsLeft(MATCH4_TURN_BONUS)
	else	-- 5 of a kind - add 10 bonus seconds
		SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[OFKIND_FIVE]", "[OFKIND_TURN2]")
		self:AddToTurnsLeft(MATCH5_TURN_BONUS)
	end
	
	local mid = match.gems[math.floor(#match.gems / 2)]
	self:ReplaceGem("Lock", mid.x, mid.y)
	if match.count >= 5 then
		local mid2 = match.gems[math.floor(#match.gems / 2) - 1]
		self:ReplaceGem("Lock", mid2.x, mid2.y)
		
		return mid, mid2
	end
	
	return mid
end

function Bash:DestroyMatches(matches)
	local function IsCoordsInTable(table, coords)
		for i, geminfo in ipairs(table) do
			if (coords.x == geminfo.x) and (coords.y == geminfo.y) then
				return true
			end
		end
		return false
	end
	
	-- explode matches
	self:PlayMatchAudio(matches)
	
	local newDoors = {}
	for i, match in ipairs(matches) do
		local mid = match.gems[math.floor(#match.gems / 2) + 1]
		self:ReplaceGem("Lock", mid.x, mid.y, true)
		table.insert(newDoors, mid)
		
		if match.count >= Constants.BONUS_TURN_COUNT then
			local mid1, mid2 = self:HandleOfAKind(match)
			if mid1 then
				table.insert(newDoors, mid1)
			end
			if mid2 then
				table.insert(newDoors, mid2)
			end
		end
	end
	
	for i, match in ipairs(matches) do
		for j, geminfo in ipairs(match.gems) do
			if not IsCoordsInTable(newDoors, geminfo) then
				self:ExplodeGem(geminfo.x, geminfo.y)
			end
		end
		
		if Gems[match.name].givesDamage then
			self.shake_id = 1 + (self.shake_id or 0)
			self.shake_id = math.min(self.shake_id, #SHAKE)
			screen:StartAnimation(SHAKE[self.shake_id])
			local intensity = self.shake_id / #SHAKE
			AddRumble(PlayerToUser(1), intensity, 800)
		end
		
		if Gems[match.name].givesDamage and self.state.health_left > 0 then
			local damage = match.score * match.bonus
			while damage > 0 and self.state.health_left > 0 do
				damage = damage - 1
				self.state.health_left = self.state.health_left - 1
				if (self.state.health_left > 0) and (math.mod(self.state.health_left, 10) == 0) then
					SCREENS.EffectsMenu:PlayDefaultSpellTextEffect(substitute(translate_text("[X_HP_TO_GO]"), self.state.health_left))
				end
			end
		end
	end
	screen:UpdateHealth(self.state.health_left, self.health_max)
	
--	self.grid:Dump()
	
	-- cascade
	local event = GameEventManager:Construct("Cascade")
	
	local delay = self:Cascade(event)
	
	GameEventManager:Send(event, self)
	
	return delay
end

function Bash:CheckEndOfTurn()
	if self.state.health_left < 1 then	-- Minigame won
		self:GameWon()
	elseif self.turns_left < 1 then		-- Minigame lost
		self:GameLost()
	else
		PuzzleHelpers.TurnsLeftFeedback(self.turns_left)
		PuzzleHelpers.StartNextTurn(self)
		self.shake_id = 0
	end
end

function Bash:GameWon()
	self.isMoving = true
	MusicManager:SetState("MINIGAME_VICTORY")
	SCREENS.EffectsMenu:PlayGameWonTextEffect("[GAME_WON]")
	local gameOverEvent = GameEventManager:Construct("GameOver")
	gameOverEvent:SetAttribute("won", 1)
	GameEventManager:SendDelayed(gameOverEvent, self, GetGameTime() + 2500)
	Achievements.AwardAchievement(1, "SMASHING", 1)
end

function Bash:GameLost()
	self.isMoving = true
	MusicManager:SetState("MINIGAME_DEFEAT")
	SCREENS.EffectsMenu:PlayGameOverTextEffect("[GAME_OVER]", "[NO_ENERGY_LEFT]")
	GameEventManager:SendDelayed(GameEventManager:Construct("GameOver"), self, GetGameTime() + 2500)
end

function Bash:OnEventGameOver(event)
	local won = event:GetAttribute("won")
	if won == 1 then
		if self.endMinigame then
			SCREENS.MinigameResultsMenu:SetupAndOpen(HeroManager:GetActiveHero(), "[MG_WON_HEAD]", "[MG_WON_BASH]", 50, true, function () EndMinigame(self.endMinigame, true) end)
		else
			EndMinigame(false, true)
		end
	else
		if self.endMinigame then
			SCREENS.MinigameResultsMenu:SetupAndOpen(HeroManager:GetActiveHero(), "[MG_LOST_HEAD]", "[MG_LOST_BASH]", 10, false, function () EndMinigame(self.endMinigame, false) end)
		else
			EndMinigame(false, false)
		end
	end
end

function Bash:AddToTurnsLeft(value)
	assert(value)
	
	if not self.turns_left then
		self.turns_left = self.max_turns
	end
	
	self.turns_left = self.turns_left + value
	
	if self.turns_left > self.max_turns then
		self.turns_left = self.max_turns
	end
	
	screen:UpdateUI(self.turns_left, self.max_turns)
end

function Bash:Cascade(event)
	assert(event)
	
	local calc = Constants.MakeDropTimeCalculator(self.gem_step)
	local delay = self.grid:Cascade(event, calc)
	
	local top = event:GetAttribute("top")
	local bottom = event:GetAttribute("bottom")
	local right = event:GetAttribute("right")
	local left = event:GetAttribute("left")
	
	local function ExpandRegion(x, y)
		if y > top then
			top = y
		end
		
		if y < bottom then
			bottom = y
		end
		
		if x < left then
			left = x
		end
		
		if x > right then
			right = x
		end
	end
	
	for x=1, self.width do
		for y=1, self.height do
			if not self.grid[x][y] then
				self.grid[x][y] = self:CreateGem(nil, x, y, false)
				
				event:PushAttribute("newx", x)
				event:PushAttribute("newy", y)
				
				ExpandRegion(x, y)
				
				delay = math.max(delay, calc(x, self.height+1, x, y, bottom))
			end
		end
	end
	
	event:SetAttribute("top", top)
	event:SetAttribute("bottom", bottom)
	event:SetAttribute("right", right)
	event:SetAttribute("left", left)
	
	self.grid:Dump()
	
	return delay
end

function Bash:ExplodeGem(x, y)
	assert(x)
	assert(y)
	
	if self.grid[x][y] then
		local particleEffects = Gems[self.grid[x][y].name].particleEffects
		if type(particleEffects) == "table" then
			for _, sys in ipairs(particleEffects) do
				AttachParticles(self, sys, self:GemCoordToPixel(x, y))
			end
		end
		
		self:DestroyGemAt(x, y)
	end
end

function Bash:ReplaceGem(name, x, y)
	assert(x)
	assert(y)
	
	if self.grid[x][y] then
		local particleEffects = Gems[self.grid[x][y].name].particleEffects
		if type(particleEffects) == "table" then
			for _, sys in ipairs(particleEffects) do
				AttachParticles(self, sys, self:GemCoordToPixel(x, y))
			end
		end
		
		self:DestroyGem(self.grid[x][y])	--At(x, y)
		self.grid[x][y] = self:CreateGem(name, x, y, true)
	end
end

local FIVE_OF_A_KIND_BONUS = 100
local FOUR_OF_A_KIND_BONUS = 50
local DOOR_MULTIPLIER = 3

function Bash:CalculateMatchScore(match)
	local score = match.score * match.bonus
	
	-- if this match gives damage then give it a bonus score
	local name = match.name
	
	if Gems[name].givesDamage then
		score = score * DOOR_MULTIPLIER
	end
	
	-- the length of the run determines if the player gets a bonus turn, or bonus turn and wildcard
	if match.count >= 5 then
		score = score + FIVE_OF_A_KIND_BONUS
	elseif match.count >= 4 then
		score = score + FOUR_OF_A_KIND_BONUS
	end
	
	return score
end

function Bash:IsValidMove()
	return true
end

function Bash:SetupHint()
	self:RemoveHint()
	
	if not self.isMoving and (screen.hintdelay > 0) then
		if self.hint_thread then
			scheduler:StopThread(self.hint_thread)
			self.hint_thread = nil
		end
		
		self.hint_thread = scheduler:AddChildThread(self, function ()
			scheduler:Sleep(screen.hintdelay)
			
			if self:CanShowHint() then
				local x1, y1, x2, y2 = self.grid:FindBestMove(function (...) return self:CalculateMatchScore(...) end)
				if self and x1 and y1 and x2 and y2 then
					local px_x1, px_y1 = self:GemCoordToPixel(x1, y1)
					px_y1 = px_y1 + (LARGE_GEM_SIZE / 2)
					local px_x2, px_y2 = self:GemCoordToPixel(x2, y2)
					px_y2 = px_y2 + (LARGE_GEM_SIZE / 2)
					self:AddOverlay("hint1", self.hint1, px_x1, px_y1)
					self:AddOverlay("hint2", self.hint2, px_x2, px_y2)
				end
			end
		end, "Bash:SetupHint")
	end
end

function Bash:SetupHintAndSelection()
	self:SetupHint()
	
	-- if gamepad connected then show the selection marker
	if IsGamepadActive() then
		self.selectGemX = self.x or math.floor(self.width / 2)
		self.selectGemY = self.y or math.floor(self.height / 2)
		self:SetSelectedGemOverlay(self.selectGemX, self.selectGemY, true)
	end
end

function Bash:RemoveHint()
	self:RemoveOverlay("hint1")
	self:RemoveOverlay("hint2")
end

function Bash:HideHint()
	self:RemoveOverlay("hint1")
	self:RemoveOverlay("hint2")
end

function Bash:CanShowHint()
	if self.isMoving then
		return false
	end
	
	if Settings:Read("hintarrowdelay", 5000) == 0 then
		return false	-- hint arrow has been disabled since request sent (#4108)
	end
	
	if IsGamepadActive() then
		if not screen.selected then
			return true
		end
	else
		if not self.selectGemX and not self.selectGemY then
			return true
		end
	end
	
	return false
end

function Bash:OnEventCascade(event)
	-- event attributes come from grid, not this class
	local top = event:GetAttribute("top")
	local left = event:GetAttribute("left")
	local right = event:GetAttribute("right")
	local bottom = event:GetAttribute("bottom")
	
	-- move all falling gems
	for i=1, event:NumAttributes("sx") do
		local sx = event:GetAttributeAt("sx", i)
		local sy = event:GetAttributeAt("sy", i)
		local tx = event:GetAttributeAt("tx", i)
		local ty = event:GetAttributeAt("ty", i)
		
		-- dont bother moving empty gems
		if self.grid[tx][ty] then
			local startGemFalling = GameEventManager:Construct("StartGemFalling")
			
			startGemFalling:SetAttribute("sx", sx)
			startGemFalling:SetAttribute("sy", sy)
			startGemFalling:SetAttribute("tx", tx)
			startGemFalling:SetAttribute("ty", ty)
			startGemFalling:SetAttribute("gem", self.grid[tx][ty])
			
			local delay = ty - bottom
			
			GameEventManager:SendDelayed(startGemFalling, self, GetGameTime() + delay * Constants.DROP_DELAY + tx * Constants.DROP_XDELAY)
		end
	end
	
	-- move all new gems
	for i=1,event:NumAttributes("newx") do
		local x = event:GetAttributeAt("newx", i)
		local y = event:GetAttributeAt("newy", i)
		
		local startGemFalling = GameEventManager:Construct("StartGemFalling")
		
		startGemFalling:SetAttribute("tx", x)
		startGemFalling:SetAttribute("ty", y)
		startGemFalling:SetAttribute("sx", x)
		startGemFalling:SetAttribute("sy", self.height+1)
		startGemFalling:SetAttribute("gem", self.grid[x][y])
		
		local delay = y - bottom
		
		GameEventManager:SendDelayed(startGemFalling, self, GetGameTime() + delay * Constants.DROP_DELAY + x * Constants.DROP_XDELAY)
	end
end

function Bash:OnEventStartGemFalling(event)
	assert(event)
	
	local gem = event:GetAttribute("gem")
	if not gem then
		-- gem was destroyed while falling, slightly unusual!
		return
	end
	
	local sx = event:GetAttribute("sx")
	local sy = event:GetAttribute("sy")
	local tx = event:GetAttribute("tx")
	local ty = event:GetAttribute("ty")
	local sx_pix, sy_pix = self:GemCoordToPixel(sx, sy)
	local tx_pix, ty_pix = self:GemCoordToPixel(tx, ty)
	
	gem:SetPos(sx_pix, sy_pix)
	gem.x = tx
	gem.y = ty
	
	if gem:GetView():GetAlpha() < 1.0 then
		gem:GetView():SetAlpha(1.0)
	end
	
	local moveToTarget = MovementManager:Construct("Gravity")
	
	moveToTarget:SetAttribute("Accel", Constants.FALL_ACCELERATION)
	moveToTarget:SetAttribute("StartX", sx_pix)
	moveToTarget:SetAttribute("StartY", sy_pix)
	moveToTarget:SetAttribute("EndX", tx_pix)
	moveToTarget:SetAttribute("EndY", ty_pix)
	
	gem:SetMovementController(moveToTarget)
end

function Bash:OnEventCheckForMatches(event)
	if event:GetAttribute("count") == Constants.HEROIC_EFFORT_COUNT then
		SoundSystem.Play("pq2audio/board/fx_heroic")
		SCREENS.EffectsMenu:PlayHeroicTextEffect("[HEROIC_ADDLOCK]")
		
--		self:AddToTurnsLeft(5)	-- To Murray - would this be needed?
		
		-- DO NOT FIX - on its own this does not causes crashes (after talking about it to Murray - is adding locks to the board a good idea?)
		local x, y, name
		local num_not_locks = (self.height * self.width) - (self.grid:CountByName("Lock"))
		for i=1, math.min(10, num_not_locks) do
			repeat
				x = math.random(self.width)
				y = math.random(self.height)
				name = "Lock"
				if self.grid[x][y] then
					name = self.grid[x][y].name
				end
			until name ~= "Lock"
			self:ReplaceGem("Lock", x, y, true)
		end
	end
	
	local matches = self.grid:FindMatches()
	
	if matches then
		local delay = self:DestroyMatches(matches)
		
		local checkForMatches = GameEventManager:Construct("CheckForMatches")
		checkForMatches:SetAttribute("count", event:GetAttribute("count") + 1)
		GameEventManager:SendDelayed(checkForMatches, self, GetGameTime() + delay + Constants.PAUSE_BETWEEN_CASCADES)
	else
		self.isMoving = false	-- no more matches, cascade has finished
		self:CheckEndOfTurn()
	end
end

function Bash:SetSpinSpeed(fast)
	if self.selectGemX and self.selectGemY then
		local gem = self.grid[self.selectGemX][self.selectGemY]
		if gem and self.spinHandle and self.spinHandle[gem] then
			SpinHelpers.EndSpin(self.spinHandle[gem])
			if fast then
				self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
				gem:AddOverlay("target", self.selectFast, 0, 0)
				if not self.spinHandle then
					self.spinHandle = {}
				end
				self.spinHandle[gem] = SpinHelpers.StartOverlaySpin(gem, "target", Constants.FAST_SPIN_RATE)
			else
				self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
				gem:AddOverlay("target", self.select1, 0, 0)
				if not self.spinHandle then
					self.spinHandle = {}
				end
				self.spinHandle[gem] = SpinHelpers.StartOverlaySpin(gem, "target", Constants.SPIN_RATE)
			end
		end
	end
end

function Bash:SetSelectedGemOverlay(gridX, gridY, isFirst, noSpin)
	local gem = self.grid[gridX][gridY]
	if gem then
		SoundSystem.Play("pq2audio/board/select1")
		local overlay = self.select1
		if not isFirst then
			overlay = self.select2
		end
		
		self:ClearSelectedGemOverlay(gridX, gridY)	--  For safety's sake, clear the old one
		gem:AddOverlay("target", overlay, 0, 0)
		if not noSpin then
			if not self.spinHandle then
				self.spinHandle = {}
			end
			self.spinHandle[gem] = SpinHelpers.StartOverlaySpin(gem, "target", Constants.SPIN_RATE)
		end
	end
end

function Bash:ClearSelectedGemOverlay(gridX, gridY)
	if gridX and gridY then
		local gem = self.grid[gridX][gridY]
		if gem then
			SoundSystem.Play("pq2audio/board/select0")
			if self.spinHandle and self.spinHandle[gem] then
				SpinHelpers.EndSpin(self.spinHandle[gem])
			end
			gem:RemoveOverlay("target")
		end
	end
	self:OnDragReset()
end

function Bash:IsAdjacent(x1, y1, x2, y2)
	if x1 == x2 then
		if math.abs(y1 - y2) == 1 then
			return true
		end
	elseif y1 == y2 then
		if math.abs(x1 - x2) == 1 then
			return true
		end
	end
	
	return false
end

function Bash:CanSelect(x, y)
	assert(x)
	assert(y)
	
	if x > 0 and x <= self.width and y > 0 and y <= self.height then
		return true
	end
	
	return false
end

function Bash:OnEventCursorEntered(event)
	local obj = event:GetAttribute("object")
	
	if obj and obj.IS_GEM then
		self:OnDragUpdate(obj)
	end
end

function Bash:OnEventCursorExited(event)
	local obj = event:GetAttribute("object")
	
	if obj and obj.IS_GEM then
		self:OnDragUpdate(nil)
	end
end

function Bash:OnEventCursorAction(event)
	local x = event:GetAttribute("x")
	local y = event:GetAttribute("y")
	local obj = event:GetAttribute("object")
	local up = event:GetAttribute("up")
	
	if self.isMoving then
		return
	end
	
	local gridX, gridY = self:GemPixelToCoord(x, y)
	LOGF("Bash:OnEventCursorAction: x=%d, y=%d, obj=%s, up=%s (%d, %d)", x, y, tostring(obj), tostring(up), gridX, gridY)
	if up then
		if self:OnDragEnd(obj, gridX, gridY) then
			return
		end
	else
		return self:OnDragStart(obj)
	end
	
	
	if self:CanSelect(gridX, gridY) and obj.IS_GEM and obj.def.canSwap then
		-- BORT: play click sound
		if self.selectGemX and self.selectGemY then
			if self.selectGemX == gridX and self.selectGemY == gridY then	-- clicked the selected gem, deselect it
				self:SetupHint()
				self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
				self.selectGemX, self.selectGemY = nil
			elseif not self:IsAdjacent(self.selectGemX, self.selectGemY, gridX, gridY) then	-- clicked a gem that is not adjacent, select it instead
				self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
				self.selectGemX, self.selectGemY = gridX, gridY
				self:SetSelectedGemOverlay(self.selectGemX, self.selectGemY, true)
			else	-- clicked a different gem, swap the selected and targeted gems
				self.targetGemX, self.targetGemY = gridX, gridY	-- Need to separate out the x and y because gems do not store their gridx and gridy positions
				self:SetSelectedGemOverlay(self.targetGemX, self.targetGemY, false)
				self:Swap(self.selectGemX, self.selectGemY, self.targetGemX, self.targetGemY)
			end
		else	-- clicked a new gem
			self:RemoveHint()
			self.selectGemX, self.selectGemY = gridX, gridY	-- Need to separate out the x and y because gems do not store their gridx and gridy positions
			self:SetSelectedGemOverlay(self.selectGemX, self.selectGemY, true)
		end
	end
end

function Bash:GetNextDisplacement(x, y)
	assert(x)
	assert(y)
	
	if x < 0 then
		return x-1, 0
	elseif x > 0 then
		return x+1, 0
	elseif y < 0 then
		return 0, y-1
	elseif y > 0 then
		return 0, y+1
	else
		return 0, 0
	end
end

function Bash:ShowGamepadSelection(xDisp, yDisp)
	assert(xDisp)
	assert(yDisp)
	
	if self.isMoving or not self.selectGemX or not self.selectGemY then
		return
	end
	
	LOGF("Bash:ShowGamepadSelection(x=%d, y=%d)", xDisp, yDisp)
	
	local x, y = self.selectGemX, self.selectGemY
	if self:CanSelect(x + xDisp, y + yDisp) then
		local obj = self.grid[x + xDisp][y + yDisp]
		if obj.IS_GEM and obj.def.canSwap then
			self:ClearSelectedGemOverlay(x, y)
			self.selectGemX = x + xDisp
			self.selectGemY = y + yDisp
			self:SetSelectedGemOverlay(x + xDisp, y + yDisp, true)
		else
			self:ShowGamepadSelection(self:GetNextDisplacement(xDisp, yDisp))
		end
	end
end

function Bash:DoGamepadSwap(xDisp, yDisp)
	assert(xDisp)
	assert(yDisp)
	
	if self.isMoving or not self.selectGemX or not self.selectGemY then
		return
	end
	
	LOGF("Bash:DoGamepadSwap(x=%d, y=%d)", xDisp, yDisp)
	
	local x, y = self.selectGemX, self.selectGemY
	if self:CanSelect(x + xDisp, y + yDisp) then
		local obj = self.grid[x + xDisp][y + yDisp]
		if obj.IS_GEM and obj.def.canSwap then
			if self:IsAdjacent(x, y, x + xDisp, y + yDisp) then
				self.x = x + xDisp
				self.y = y + yDisp
			else	-- exact diagonal, pick which direction to move
				if math.random(1,2) == 1 then
					self.x = x
					self.y = y + yDisp
				else
					self.x = x + xDisp
					self.y = y
				end
			end
			self:Swap(x, y, self.x, self.y)
		end
	else	-- cannot swap outside of board, reset selection
		screen.selected = false	-- HAX!
		self:SetupHint()
		self:SetSpinSpeed(false)
	end
end

function Bash:PlayMatchAudio(matches)
	local total = 0
	local num_doors = 0
	
	for _, match in ipairs(matches) do
		total = total + match.count
		if match.name == "Lock" then
			num_doors = num_doors + match.count
		end
	end
	
	if total < 7 then
		SoundSystem.Play("pq2audio/board/gemmatch_0")
	elseif total < 13 then
		SoundSystem.Play("pq2audio/board/gemmatch_1")
	else
		SoundSystem.Play("pq2audio/board/gemmatch_2")
	end
	
	if num_doors > 4 then
		SoundSystem.Play(string.format("pq2audio/board/gembashwood_1", self.state.door))
	elseif num_doors > 1 then
		SoundSystem.Play(string.format("pq2audio/board/gembashwood_0", self.state.door))
	end
end

return ExportClass("Bash")
