-- The HOARD mini game

use_safeglobals()

local Grid = require("Battlegrounds/Grid")
local ProbabilityTable = require("Utils/ProbabilityTable")
local Constants = require("Battlegrounds/BattleGroundConstants")
local SpinHelpers = require("Utils/SpinHelpers")

class "Hoard" (GameObject)

require("Utils/DraggingSupport"):AddToClass(Hoard)

Hoard.AttributeDescriptions = AttributeDescriptionList()

local LARGE_GEM_SIZE 	= Constants.PUZZLE_GEM_SIZE
local LARGE_GEM_SCALE	= Constants.PUZZLE_GEM_SCALE

local screen = SCREENS.HoardGameMenu

local function EndMinigame(endMinigame, result)
	if endMinigame then
		CallScreenSequencer("HoardGameMenu", function() endMinigame(result) end)
	else
		CallScreenSequencer("HoardGameMenu", "MinigameSelectionMenu")
	end
end

function Hoard:__init(clid)
	super("PHRD")
	
	self.gem_step = Constants.PUZZLE_GEM_STEP
	self.isMoving = false	-- this will be set to true the very instance a move is made, and unset when the board is in a steady state
	self.collection = {}
	
	self.hint1 = GameObjectViewManager:Construct("GraphicAsset", "anim_hint")
	self.hint1:SetScale(LARGE_GEM_SCALE)
	self.hint1:SetSortingValue(Constants.HINT_ARROW_SORTING_VALUE)
	self.hint1:SetAlpha(Constants.HINT_ARROW_ALPHA_VALUE)
	
	self.hint2 = GameObjectViewManager:Construct("GraphicAsset", "anim_hint")
	self.hint2:SetScale(LARGE_GEM_SCALE)
	self.hint2:SetSortingValue(Constants.HINT_ARROW_SORTING_VALUE)
	self.hint2:SetAlpha(Constants.HINT_ARROW_ALPHA_VALUE)
	
	self.select1 = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_g")
	self.select1:SetScale(LARGE_GEM_SCALE)
	self.select1:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self.select2 = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_g")
	self.select2:SetScale(LARGE_GEM_SCALE)
	self.select2:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self.selectFast = GameObjectViewManager:Construct("GraphicAsset", "img_target_select_b")
	self.selectFast:SetScale(LARGE_GEM_SCALE)
	self.selectFast:SetSortingValue(Constants.GEM_SELECTION_SORTING_VALUE)
	
	self:SetCursorInteract(false)
	self:OnDragInit(Constants.GEM_SELECTION_SORTING_VALUE, LARGE_GEM_SCALE)
end

function Hoard:PreDestroy()
	if leakdetect then
		leakdetect.PreDestroyHook(self)
	end
	scheduler:StopAllThreadsForObject(self)
end

function Hoard:GemCoordToPixel(x, y)
	return math.round((x - 1) * self.gem_step + (self.gem_step / 2)), math.round(y * self.gem_step - (self.gem_step / 2))
end

function Hoard:GemPixelToCoord(px, py)
	return math.round((px - (self.gem_step / 2)) / self.gem_step + 1), math.round((py + (self.gem_step / 2)) / self.gem_step)
end

function Hoard:Initialise(layout, gemProbabilityTable, difficultyLevel, endMinigame)
	assert(layout)
	assert(gemProbabilityTable)
	assert(difficultyLevel)
	
	self.width = layout.grid_width
	self.height = layout.grid_height
	self.gemProbabilityTable = gemProbabilityTable
	self.difficulty = difficultyLevel
	self.endMinigame = endMinigame
	
	self.num_turns = 0
	self.cycle = 2
	
	self.grid = Grid:Construct(self.width, self.height, self.grid)
	
	for row = self.height, 1, -1 do
		if layout[row] then
			local y = (1 + self.height) - row
			for x = 1, self.width do
				local gem = layout[row][x]
				if gem and gem ~= " " then
					self.grid[x][y] = self:CreateGem(Gems:GetNameFromSymbol(gem), x, y, true)
				end
			end
		end
	end
	
	repeat
		for x=1, self.width do
			for y=1, self.height do
				if not self.grid[x][y] then
					self.grid[x][y] = self:CreateGem(nil, x, y, true)
				end
			end
		end
		
		local matches = self.grid:FindMatches()
		
		if matches then
			-- Pop the middle gems in each match
			for i, match in ipairs(matches) do
				for j=2, #match.gems-1 do
					local geminfo = match.gems[j]
					
					self:DestroyGemAt(geminfo.x, geminfo.y)
				end
			end
		end
	until not matches
	
	self:SetupHintAndSelection()
	
	scheduler:AddChildThread(self, function ()
		scheduler:Sleep(1000)
		SCREENS.EffectsMenu:PlayDefaultSpellTextEffect("[HOARD_TEXT]")
	end, "Hoard:Initialise")
end

function Hoard:RefreshUI()
	local gold, lqItems, hqItems = self:GetCollectionTotals()
	screen:UpdateGold(gold)
	screen:UpdateItems(lqItems, hqItems)
	screen:UpdateTurns(self.num_turns)
end

function Hoard:CreateGem(name, x, y, visible)
	assert(x)
	assert(y)
	
	if not name then
		name = self.gemProbabilityTable:RandomItem()
	end
	
	if name == "Wildcard" then
		if not self.wildcardProbabilityTable then
			self.wildcardProbabilityTable = ProbabilityTable:Construct(Constants.WILDCARD_DIST)
		end
		name = "Wildcard" .. self.wildcardProbabilityTable:RandomItem()
	end
	
	local gem = GameObjectManager:ConstructLocal("BGEM")
	gem:SetInfo(Gems[name], LARGE_GEM_SCALE, LARGE_GEM_SIZE)
	gem:SetPos(self:GemCoordToPixel(x, y))
	gem.x = x
	gem.y = y

	if not visible then
		gem:GetView():SetAlpha(0.0)
	end
	
	self:AddChild(gem)
	
	return gem
end

function Hoard:DestroyGem(gem)
	if gem then
		GameObjectManager:Destroy(gem)
	end
end

function Hoard:DestroyGemAt(x, y)
	assert(x)
	assert(y)
	
	self:DestroyGem(self.grid[x][y])
	self.grid[x][y] = nil
end

function Hoard:Swap(x1, y1, x2, y2)
	assert(x1)
	assert(y1)
	assert(x2)
	assert(y2)
	
	LOGF("Hoard:Swap(%d,%d,%d,%d)", x1, y1, x2, y2)
	
	local valid, reason = self.grid:IsValidMove(x1, y1, x2, y2)
	if not valid then
		LOGF("Rejected move %d,%d <-> %d,%d because: %s", x1, y1, x2, y2, reason)
		self:HandleSwapRejection(x1, y1, x2, y2)
		return
	end
	
	self.num_turns = self.num_turns + 1
	screen:UpdateTurns(self.num_turns)
	
	SoundSystem.Play("pq2audio/board/gemmove")
	self:SwapGems(x1, y1, x2, y2)
	
	-- wait for half a second (this is how long it takes the views to swap to gems) before cascading
	local checkForMatches = GameEventManager:Construct("CheckForMatches")
	checkForMatches:SetAttribute("count", 1)
	GameEventManager:SendDelayed(checkForMatches, self, GetGameTime() + Constants.PAUSE_AFTER_MOVE)
end

function Hoard:SendSwapRequest(x1, y1, x2, y2)
	return self:Swap(x1, y1, x2, y2)
end

function Hoard:HandleSwapRejection(x1, y1, x2, y2)
	self:SwapGems(x1, y1, x2, y2)
	
	SoundSystem.Play("pq2audio/ui/illegal")
	local px1, py1 = self:GemCoordToPixel(x1, y1)
	local px2, py2 = self:GemCoordToPixel(x2, y2)
	local x = (px1 + px2) / 2
	local y = (py1 + py2) / 2
	SCREENS.EffectsMenu:PlayFallingTextEffectAt(x, y, "[ILLEGAL_MOVE]")
	
	-- Then send another event to make sure they swap back
	local event = GameEventManager:Construct("SwapBack")
	event:SetAttribute("source_x", x2)
	event:SetAttribute("source_y", y2)
	event:SetAttribute("target_x", x1)
	event:SetAttribute("target_y", y1)
	GameEventManager:SendDelayed(event, self, GetGameTime() + Constants.GEM_SWAP_TIME)
end

function Hoard:OnEventSwapBack(event)
	if not self.swappedBack then
		self.swappedBack = true
		local x1, y1 = event:GetAttribute("source_x"), event:GetAttribute("source_y")
		local x2, y2 = event:GetAttribute("target_x"), event:GetAttribute("target_y")
		self:SwapGems(x1, y1, x2, y2)
		GameEventManager:SendDelayed(event, self, GetGameTime() + 1.1 * Constants.GEM_SWAP_TIME)
	else
		self.swappedBack = false
		self.isMoving = false
		self:SetupHintAndSelection()
	end
end

function Hoard:SwapGems(x1, y1, x2, y2)
	self.isMoving = true
	
	-- clear old overlays
	self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
	self.selectGemX, self.selectGemY = nil
	
	if self.targetGemX and self.targetGemY then
		self:ClearSelectedGemOverlay(self.targetGemX, self.targetGemY)
		self.targetGemX, self.targetGemY = nil
	end
	
	local gem1 = self.grid[x1][y1]
	local gem2 = self.grid[x2][y2]
	
	-- set new overlays
	if gem1 then
		self:SetSelectedGemOverlay(x1, y1, true, true)
	end
	if gem2 then
		self:SetSelectedGemOverlay(x2, y2, false, true)
	end
	
	if gem1 then
		self:MoveGemTo(gem1, x2, y2)
	end
	if gem2 then
		self:MoveGemTo(gem2, x1, y1)
	end
end

function Hoard:MoveGemTo(gem, target_x, target_y)
	local target_x_pix, target_y_pix = self:GemCoordToPixel(target_x, target_y)
	local x_pix, y_pix = gem:GetX(), gem:GetY()
	
	gem.x = assert(target_x)
	gem.y = assert(target_y)
	
	self.grid[target_x][target_y] = gem
	
	local moveToTarget = MovementManager:Construct("Smooth")
	
	moveToTarget:SetAttribute("Duration", Constants.GEM_SWAP_TIME)
	moveToTarget:SetAttribute("StartX", x_pix)
	moveToTarget:SetAttribute("StartY", y_pix)
	moveToTarget:SetAttribute("EndX", target_x_pix)
	moveToTarget:SetAttribute("EndY", target_y_pix)
	
	gem:SetMovementController(moveToTarget)
end

function Hoard:HandleOfAKind(match)
	SoundSystem.Play("pq2audio/board/fx_extraturn")
	local gx, gy = SCREENS.EffectsMenu:GetAverageGridPos(match.gems)
	local x, y = self:GemCoordToPixel(gx, gy)
	
	if match.count == 4 then	-- 4 of a kind - add 1 bonus chest
		SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[OFKIND_FOUR]", "[OFKIND_ADDCHEST]")
	else	-- 5 of a kind - add 1 bonus rare chest
		SCREENS.EffectsMenu:PlayRisingTextEffectAt(x, y, "[OFKIND_FIVE]", "[OFKIND_ADDRARECHEST]")
	end
	
	local mid = match.gems[math.floor(#match.gems / 2) + 1]	-- why +1?
	
	if match.count >= 5 then
		self:ReplaceGem("Chest5", mid.x, mid.y)
	else
		self:ReplaceGem("Chest1", mid.x, mid.y)
	end
	
	return mid
end

function Hoard:DestroyMatches(matches)
	local function IsCoordsInTable(table, coords)
		for i, geminfo in ipairs(table) do
			if (coords.x == geminfo.x) and (coords.y == geminfo.y) then
				return true
			end
		end
		return false
	end
	
	-- explode matches
	self:PlayMatchAudio(matches)
	
	local newChests = {}
	for i, match in ipairs(matches) do
		if match.count >= Constants.BONUS_TURN_COUNT then
			table.insert(newChests, self:HandleOfAKind(match))
		end
	end
	
	for i, match in ipairs(matches) do
		for j, geminfo in ipairs(match.gems) do
			if not IsCoordsInTable(newChests, geminfo) then
				self:ExplodeGem(geminfo.x, geminfo.y)
			end
		end
		
		if not self.collection[match.name] then
			self.collection[match.name] = 0
		end
		self.collection[match.name] = self.collection[match.name] + (match.count - 2)
	end
	
	local gold, lqItems, hqItems = self:GetCollectionTotals()
	screen:UpdateGold(gold)
	screen:UpdateItems(lqItems, hqItems)
	
--	self.grid:Dump()
	
	-- cascade
	local event = GameEventManager:Construct("Cascade")
	
	local delay = self:Cascade(event)
	
	GameEventManager:Send(event, self)
	
	return delay
end

function Hoard:GetCollectionTotals()
	local gold5 = self.collection["Gold5"] or 0
	local gold50 = self.collection["Gold50"] or 0
	local gold500 = self.collection["Gold500"] or 0
	local lqItems = self.collection["Chest1"] or 0
	local hqItems = self.collection["Chest5"] or 0
	
	local gold_total = ((gold5) + (gold50 * 3) + (gold500 * 5)) * 3
	local lqItems_total = lqItems
	local hqItems_total = hqItems
	
	return gold_total, lqItems_total, hqItems_total
end

function Hoard:CheckEndOfTurn()
	if (self.num_turns % self.cycle) == 0 then
		SoundSystem.Play("pq2audio/board/blocks")
		local y = self.num_turns / self.cycle
		for x=1, self.width do
			self:DestroyGem(self.grid[x][y])	--At(x, y)
			self.grid[x][y] = self:CreateGem("Blocking", x, y, true)
		end
		if y == self.y then
			self.y = self.y + 1
		end
	end
	
	if (self.num_turns == self.cycle * self.height) or (self.grid:IsLocked()) then
		self:GameWon()
	else
		if (self.num_turns % self.cycle) == 0 then
			local rows_left = self.height - self.num_turns / self.cycle
			if rows_left > 0 then
				if rows_left == 1 then
					SCREENS.EffectsMenu:PlayDefaultSpellTextEffect("[LAST_ROW]")
				else
					SCREENS.EffectsMenu:PlayDefaultSpellTextEffect(substitute(translate_text("[X_ROWS_LEFT]"), rows_left))
				end
			end
		end
		
		self:SetupHintAndSelection()
	end
end

function Hoard:GameWon()
	self.isMoving = true
	MusicManager:SetState("MINIGAME_VICTORY")
	SCREENS.EffectsMenu:PlayGameWonTextEffect("[GAME_WON]")
	local gameOverEvent = GameEventManager:Construct("GameOver")
	gameOverEvent:SetAttribute("won", 1)
	GameEventManager:SendDelayed(gameOverEvent, self, GetGameTime() + 2500)
	Stats.AdjustSPStat("OVERALL", "HOARD_WINS", 1)
	Stats.PushSinglePlayerStats()
end

function Hoard:OnEventGameOver(event)
	local won = event:GetAttribute("won")
	if won == 1 then
		if self.endMinigame then
			local gold, items = SCREENS.HoardGameMenu:GetGoldAndItems()
			SCREENS.HoardResultsMenu:SetupAndOpen("[MG_WON_HEAD]", gold, items, function () EndMinigame(self.endMinigame, true) end)
		else
			EndMinigame(false, true)
		end
	end
end

function Hoard:Cascade(event)
	assert(event)
	
	local calc = Constants.MakeDropTimeCalculator(self.gem_step)
	local delay = self.grid:Cascade(event, calc)
	
	local top = event:GetAttribute("top")
	local bottom = event:GetAttribute("bottom")
	local right = event:GetAttribute("right")
	local left = event:GetAttribute("left")
	
	local function ExpandRegion(x, y)
		if y > top then
			top = y
		end
		
		if y < bottom then
			bottom = y
		end
		
		if x < left then
			left = x
		end
		
		if x > right then
			right = x
		end
	end
	
	for x=1, self.width do
		for y=1, self.height do
			if not self.grid[x][y] then
				self.grid[x][y] = self:CreateGem(nil, x, y, false)
				
				event:PushAttribute("newx", x)
				event:PushAttribute("newy", y)
				
				ExpandRegion(x, y)
				
				delay = math.max(delay, calc(x, self.height+1, x, y, bottom))
			end
		end
	end
	
	event:SetAttribute("top", top)
	event:SetAttribute("bottom", bottom)
	event:SetAttribute("right", right)
	event:SetAttribute("left", left)
	
	self.grid:Dump()
	
	return delay
end

function Hoard:ExplodeGem(x, y)
	assert(x)
	assert(y)
	
	if self.grid[x][y] then
		local particleEffects = Gems[self.grid[x][y].name].particleEffects
		if type(particleEffects) == "table" then
			for _, sys in ipairs(particleEffects) do
				AttachParticles(self, sys, self:GemCoordToPixel(x, y))
			end
		end
		
		self:DestroyGemAt(x, y)
	end
end

function Hoard:ReplaceGem(name, x, y)
	assert(x)
	assert(y)
	
	if self.grid[x][y] then
		local particleEffects = Gems[self.grid[x][y].name].particleEffects
		if type(particleEffects) == "table" then
			for _, sys in ipairs(particleEffects) do
				AttachParticles(self, sys, self:GemCoordToPixel(x, y))
			end
		end
		
		self:DestroyGem(self.grid[x][y])	--At(x, y)
		self.grid[x][y] = self:CreateGem(name, x, y, true)
	end
end

local FIVE_OF_A_KIND_BONUS = 100
local FOUR_OF_A_KIND_BONUS = 50
local CHEST_MULTIPLIER = 150

function Hoard:CalculateMatchScore(match)
	local score = match.score * match.bonus
	
	-- if this match is a chest then give it a bonus score, otherwise weight the score in the order of gold
	local name = match.name
	local manaWeights = {
		Gold5 = 1,
		Gold50 = 2,
		Gold500 = 3,
	}
	
	if string.sub(name, 1, 5) == "Chest" then
		score = score * CHEST_MULTIPLIER * tonumber(string.sub(name, 6, 6))
	elseif manaWeights[name] then
		score = score * manaWeights[name]
	end
	
	-- the length of the run determines if the player gets a bonus turn, or bonus turn and wildcard
	if match.count >= 5 then
		score = score + FIVE_OF_A_KIND_BONUS
	elseif match.count >= 4 then
		score = score + FOUR_OF_A_KIND_BONUS
	end
	
	return score
end

function Hoard:IsValidMove()
	return true
end

function Hoard:SetupHint()
	self:RemoveHint()
	
	if not self.isMoving and (screen.hintdelay > 0) then
		if self.hint_thread then
			scheduler:StopThread(self.hint_thread)
			self.hint_thread = nil
		end
		
		self.hint_thread = scheduler:AddChildThread(self, function ()
			scheduler:Sleep(screen.hintdelay)
			
			if self:CanShowHint() then
				local x1, y1, x2, y2 = self.grid:FindBestMove(function (...) return self:CalculateMatchScore(...) end)
				if self and x1 and y1 and x2 and y2 then
					local px_x1, px_y1 = self:GemCoordToPixel(x1, y1)
					px_y1 = px_y1 + (LARGE_GEM_SIZE / 2)
					local px_x2, px_y2 = self:GemCoordToPixel(x2, y2)
					px_y2 = px_y2 + (LARGE_GEM_SIZE / 2)
					self:AddOverlay("hint1", self.hint1, px_x1, px_y1)
					self:AddOverlay("hint2", self.hint2, px_x2, px_y2)
				end
			end
		end, "Hoard:SetupHint")
	end
end

function Hoard:SetupHintAndSelection()
	self:SetupHint()
	
	-- if gamepad connected then show the selection marker
	if IsGamepadActive() then
		self.selectGemX = self.x or math.floor(self.width / 2)
		self.selectGemY = self.y or math.floor(self.height / 2)
		self:SetSelectedGemOverlay(self.selectGemX, self.selectGemY, true)
	end
end

function Hoard:RemoveHint()
	self:RemoveOverlay("hint1")
	self:RemoveOverlay("hint2")
end

function Hoard:HideHint()
	self:RemoveOverlay("hint1")
	self:RemoveOverlay("hint2")
end

function Hoard:CanShowHint()
	if self.isMoving then
		return false
	end
	
	if Settings:Read("hintarrowdelay", 5000) == 0 then
		return false	-- hint arrow has been disabled since request sent (#4108)
	end
	
	if IsGamepadActive() then
		if not screen.selected then
			return true
		end
	else
		if not self.selectGemX and not self.selectGemY then
			return true
		end
	end
	
	return false
end

function Hoard:OnEventCascade(event)
	-- event attributes come from grid, not this class
	local top = event:GetAttribute("top")
	local left = event:GetAttribute("left")
	local right = event:GetAttribute("right")
	local bottom = event:GetAttribute("bottom")
	
	-- move all falling gems
	for i=1, event:NumAttributes("sx") do
		local sx = event:GetAttributeAt("sx", i)
		local sy = event:GetAttributeAt("sy", i)
		local tx = event:GetAttributeAt("tx", i)
		local ty = event:GetAttributeAt("ty", i)
		
		-- dont bother moving empty gems
		if self.grid[tx][ty] then
			local startGemFalling = GameEventManager:Construct("StartGemFalling")
			
			startGemFalling:SetAttribute("sx", sx)
			startGemFalling:SetAttribute("sy", sy)
			startGemFalling:SetAttribute("tx", tx)
			startGemFalling:SetAttribute("ty", ty)
			startGemFalling:SetAttribute("gem", self.grid[tx][ty])
			
			local delay = ty - bottom
			
			GameEventManager:SendDelayed(startGemFalling, self, GetGameTime() + delay * Constants.DROP_DELAY + tx * Constants.DROP_XDELAY)
		end
	end
	
	-- move all new gems
	for i=1,event:NumAttributes("newx") do
		local x = event:GetAttributeAt("newx", i)
		local y = event:GetAttributeAt("newy", i)
		
		local startGemFalling = GameEventManager:Construct("StartGemFalling")
		
		startGemFalling:SetAttribute("tx", x)
		startGemFalling:SetAttribute("ty", y)
		startGemFalling:SetAttribute("sx", x)
		startGemFalling:SetAttribute("sy", self.height+1)
		startGemFalling:SetAttribute("gem", self.grid[x][y])
		
		local delay = y - bottom
		
		GameEventManager:SendDelayed(startGemFalling, self, GetGameTime() + delay * Constants.DROP_DELAY + x * Constants.DROP_XDELAY)
	end
end

function Hoard:OnEventStartGemFalling(event)
	assert(event)
	
	local gem = event:GetAttribute("gem")
	if not gem then
		-- gem was destroyed while falling, slightly unusual!
		return
	end
	
	local sx = event:GetAttribute("sx")
	local sy = event:GetAttribute("sy")
	local tx = event:GetAttribute("tx")
	local ty = event:GetAttribute("ty")
	local sx_pix, sy_pix = self:GemCoordToPixel(sx, sy)
	local tx_pix, ty_pix = self:GemCoordToPixel(tx, ty)
	
	gem:SetPos(sx_pix, sy_pix)
	gem.x = tx
	gem.y = ty

	if gem:GetView():GetAlpha() < 1.0 then
		gem:GetView():SetAlpha(1.0)
	end
	
	local moveToTarget = MovementManager:Construct("Gravity")
	
	moveToTarget:SetAttribute("Accel", Constants.FALL_ACCELERATION)
	moveToTarget:SetAttribute("StartX", sx_pix)
	moveToTarget:SetAttribute("StartY", sy_pix)
	moveToTarget:SetAttribute("EndX", tx_pix)
	moveToTarget:SetAttribute("EndY", ty_pix)
	
	gem:SetMovementController(moveToTarget)
end

function Hoard:OnEventCheckForMatches(event)
	if event:GetAttribute("count") == Constants.HEROIC_EFFORT_COUNT then
		SoundSystem.Play("pq2audio/board/fx_heroic")
		SCREENS.EffectsMenu:PlayHeroicTextEffect("[HEROIC_ADDCHESTS]")
		
		local x, y, name
		for i=1, 2 do
			repeat
				x = math.random(self.width)
				y = math.random(math.min(1+math.floor(self.num_turns/self.cycle), self.height), self.height)	-- make sure chests won't be swallowed up by an impending rise
				name = self.grid[x][y].name
			until name ~= "Blocking" and string.sub(name, 1, 5) ~= "Chest"
			
			self:ReplaceGem("Chest5", x, y, true)
		end
	end
	
	local matches = self.grid:FindMatches()
	
	if matches then
		local delay = self:DestroyMatches(matches)
		
		local checkForMatches = GameEventManager:Construct("CheckForMatches")
		checkForMatches:SetAttribute("count", event:GetAttribute("count") + 1)
		GameEventManager:SendDelayed(checkForMatches, self, GetGameTime() + delay + Constants.PAUSE_BETWEEN_CASCADES)
	else
		self.isMoving = false  -- no more matches, cascade has finished
		self:CheckEndOfTurn()
	end
end

function Hoard:SetSpinSpeed(fast)
	if self.selectGemX and self.selectGemY then
		local gem = self.grid[self.selectGemX][self.selectGemY]
		if gem and self.spinHandle and self.spinHandle[gem] then
			SpinHelpers.EndSpin(self.spinHandle[gem])
			if fast then
				self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
				gem:AddOverlay("target", self.selectFast, 0, 0)
				if not self.spinHandle then
					self.spinHandle = {}
				end
				self.spinHandle[gem] = SpinHelpers.StartOverlaySpin(gem, "target", Constants.FAST_SPIN_RATE)
			else
				self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
				gem:AddOverlay("target", self.select1, 0, 0)
				if not self.spinHandle then
					self.spinHandle = {}
				end
				self.spinHandle[gem] = SpinHelpers.StartOverlaySpin(gem, "target", Constants.SPIN_RATE)
			end
		end
	end
end

function Hoard:SetSelectedGemOverlay(gridX, gridY, isFirst, noSpin)
	local gem = self.grid[gridX][gridY]
	if gem then
		SoundSystem.Play("pq2audio/board/select1")
		local overlay = self.select1
		if not isFirst then
			overlay = self.select2
		end
		
		self:ClearSelectedGemOverlay(gridX, gridY)	--  For safety's sake, clear the old one
		gem:AddOverlay("target", overlay, 0, 0)
		if not noSpin then
			if not self.spinHandle then
				self.spinHandle = {}
			end
			self.spinHandle[gem] = SpinHelpers.StartOverlaySpin(gem, "target", Constants.SPIN_RATE)
		end
	end
end

function Hoard:ClearSelectedGemOverlay(gridX, gridY)
	if gridX and gridY then
		local gem = self.grid[gridX][gridY]
		if gem then
			SoundSystem.Play("pq2audio/board/select0")
			if self.spinHandle and self.spinHandle[gem] then
				SpinHelpers.EndSpin(self.spinHandle[gem])
			end
			gem:RemoveOverlay("target")
		end
	end
	self:OnDragReset()
end

function Hoard:IsAdjacent(x1, y1, x2, y2)
	if x1 == x2 then
		if math.abs(y1 - y2) == 1 then
			return true
		end
	elseif y1 == y2 then
		if math.abs(x1 - x2) == 1 then
			return true
		end
	end
	
	return false
end

function Hoard:CanSelect(x, y)
	assert(x)
	assert(y)
	
	if x > 0 and x <= self.width and y > 0 and y <= self.height then
		return true
	end
	
	return false
end

function Hoard:OnEventCursorEntered(event)
	local obj = event:GetAttribute("object")
	
	if obj and obj.IS_GEM then
		self:OnDragUpdate(obj)
	end
end

function Hoard:OnEventCursorExited(event)
	local obj = event:GetAttribute("object")
	
	if obj and obj.IS_GEM then
		self:OnDragUpdate(nil)
	end
end

function Hoard:OnEventCursorAction(event)
	local x = event:GetAttribute("x")
	local y = event:GetAttribute("y")
	local obj = event:GetAttribute("object")
	local up = event:GetAttribute("up")
	
	if self.isMoving then
		return
	end
	
	local gridX, gridY = self:GemPixelToCoord(x, y)
	LOGF("Hoard:OnEventCursorAction: x=%d, y=%d, obj=%s, up=%s (%d, %d)", x, y, tostring(obj), tostring(up), gridX, gridY)
	
	if up then
		if self:OnDragEnd(obj, gridX, gridY) then
			return
		end
	else
		return self:OnDragStart(obj)
	end

	if self:CanSelect(gridX, gridY) and obj.IS_GEM and obj.def.canSwap then
		-- BORT: play click sound
		if self.selectGemX and self.selectGemY then
			if self.selectGemX == gridX and self.selectGemY == gridY then	-- clicked the selected gem, deselect it
				self:SetupHint()
				self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
				self.selectGemX, self.selectGemY = nil
			elseif not self:IsAdjacent(self.selectGemX, self.selectGemY, gridX, gridY) then	-- clicked a gem that is not adjacent, select it instead
				self:ClearSelectedGemOverlay(self.selectGemX, self.selectGemY)
				self.selectGemX, self.selectGemY = gridX, gridY
				self:SetSelectedGemOverlay(self.selectGemX, self.selectGemY, true)
			else	-- clicked a different gem, swap the selected and targeted gems
				self.targetGemX, self.targetGemY = gridX, gridY	-- Need to separate out the x and y because gems do not store their gridx and gridy positions
				self:SetSelectedGemOverlay(self.targetGemX, self.targetGemY, false)
				self:Swap(self.selectGemX, self.selectGemY, self.targetGemX, self.targetGemY)
			end
		else	-- clicked a new gem
			self:RemoveHint()
			self.selectGemX, self.selectGemY = gridX, gridY	-- Need to separate out the x and y because gems do not store their gridx and gridy positions
			self:SetSelectedGemOverlay(self.selectGemX, self.selectGemY, true)
		end
	end
end

function Hoard:GetNextDisplacement(x, y)
	assert(x)
	assert(y)
	
	if x < 0 then
		return x-1, 0
	elseif x > 0 then
		return x+1, 0
	elseif y < 0 then
		return 0, y-1
	elseif y > 0 then
		return 0, y+1
	else
		return 0, 0
	end
end

function Hoard:ShowGamepadSelection(xDisp, yDisp)
	assert(xDisp)
	assert(yDisp)
	
	if self.isMoving or not self.selectGemX or not self.selectGemY then
		return
	end
	
	LOGF("Hoard:ShowGamepadSelection(x=%d, y=%d)", xDisp, yDisp)
	
	local x, y = self.selectGemX, self.selectGemY
	if self:CanSelect(x + xDisp, y + yDisp) then
		local obj = self.grid[x + xDisp][y + yDisp]
		if obj.IS_GEM and obj.def.canSwap then
			self:ClearSelectedGemOverlay(x, y)
			self.selectGemX = x + xDisp
			self.selectGemY = y + yDisp
			self:SetSelectedGemOverlay(x + xDisp, y + yDisp, true)
		else
			self:ShowGamepadSelection(self:GetNextDisplacement(xDisp, yDisp))
		end
	end
end

function Hoard:DoGamepadSwap(xDisp, yDisp)
	assert(xDisp)
	assert(yDisp)
	
	if self.isMoving or not self.selectGemX or not self.selectGemY then
		return
	end
	
	LOGF("Hoard:DoGamepadSwap(x=%d, y=%d)", xDisp, yDisp)
	
	local x, y = self.selectGemX, self.selectGemY
	if self:CanSelect(x + xDisp, y + yDisp) then
		local obj = self.grid[x + xDisp][y + yDisp]
		if obj.IS_GEM and obj.def.canSwap then
			if self:IsAdjacent(x, y, x + xDisp, y + yDisp) then
				self.x = x + xDisp
				self.y = y + yDisp
			else	-- exact diagonal, pick which direction to move
				if math.random(1,2) == 1 then
					self.x = x
					self.y = y + yDisp
				else
					self.x = x + xDisp
					self.y = y
				end
			end
			self:Swap(x, y, self.x, self.y)
		end
	else	-- cannot swap outside of board, reset selection
		screen.selected = false	-- HAX!
		self:SetupHint()
		self:SetSpinSpeed(false)
	end
end

function Hoard:PlayMatchAudio(matches)	
	local total = 0
	local gem_num = 
	{
		Gold5   = 0,
		Gold50  = 0,
		Gold500 = 0,
		Chest1  = 0,
		Chest5  = 0,
	}
	for _, match in ipairs(matches) do
		gem_num[match.name] = gem_num[match.name] + match.count
		total = total + match.count
	end
	
	if total < 7 then
		SoundSystem.Play("pq2audio/board/gemmatch_0")
	elseif total < 13 then
		SoundSystem.Play("pq2audio/board/gemmatch_1")
	else
		SoundSystem.Play("pq2audio/board/gemmatch_2")
	end
	
	if gem_num.Gold5 > 4 then
		SoundSystem.Play("pq2audio/board/gemcoin_1")
	elseif gem_num.Gold5 > 0 then
		SoundSystem.Play("pq2audio/board/gemcoin_0")
	end
	
	if gem_num.Gold50 > 4 then
		SoundSystem.Play("pq2audio/board/gemcup_1")
	elseif gem_num.Gold50 > 0 then
		SoundSystem.Play("pq2audio/board/gemcup_0")
	end
	
	if gem_num.Gold500 > 4 then
		SoundSystem.Play("pq2audio/board/gemcrown_1")
	elseif gem_num.Gold500 > 0 then
		SoundSystem.Play("pq2audio/board/gemcrown_0")
	end
	
	if gem_num.Chest5 > 0 then
		SoundSystem.Play("pq2audio/board/gemchest_1")
	elseif gem_num.Chest1 > 0 then
		SoundSystem.Play("pq2audio/board/gemchest_0")
	end
end

return ExportClass("Hoard")
