use_safeglobals()

local ProbabilityTable = require("Utils/ProbabilityTable")
local MenuHelpers = require("Utils/MenuHelpers")
local Configuration = require("Puzzles/PuzzleConfiguration").HOARD
local ItemGenerator = require("Items.Generator")

class "HoardGameMenu" (cGameMenu);

local MAX_ITEMS_SHOWN = 7
local MAX_LQITEMS_SHOWN = 5
local MAX_HQITEMS_SHOWN = 2
local MAX_TURNS = 16
local MAX_GOLD = 800

function HoardGameMenu:__init()
	super()
	
	SoundSystem.Load("board_bank")
	
	_G.Joystick.InitJoystick()
	
	self:Initialize("Assets\\Screens\\HoardGameMenu.xml")
end

function HoardGameMenu:SetBoard(board)
	self.board = board
end

function HoardGameMenu:SetDifficulty(difficulty)
	self.difficulty = difficulty
end

function HoardGameMenu:SetEndMinigame(endMinigame)
	self.endMinigame = endMinigame
end

function HoardGameMenu:SetGameOptions(gameOptions)
	self.gameOptions = gameOptions
end

function HoardGameMenu:OnOpen()
	SCREENS.EffectsMenu:Open()
	SCREENS.EffectsMenu:MoveToFront()
	
	MenuHelpers.Init()

	self.hintdelay = Settings:Read("hintarrowdelay", 5000)
	
	self.gold = 0
	self.numLqItems = 0
	self.numHqItems = 0
	self.lqitems = {}
	self.hqitems = {}
	
	self:set_progress_max("prog_turns", MAX_TURNS)
	self:set_progress("prog_turns", MAX_TURNS)
	self:UpdateTurns(0)
	self:set_progress_max("prog_gold", MAX_GOLD)
	self:UpdateGold(0)
	
	for i=1, MAX_ITEMS_SHOWN do
		self:set_image(string.format("icon_chest%d", i), "")
	end
	
	if not self.board then
		self.board = "Default8x8"
	end
	
	self:Initialise(require(string.format("Grids.%s", self.board)))
	
	-- TUTORIAL TODO: Hoard minigame begins
	
	self.hero = HeroManager:GetActiveHero()
	
	-- Play special minigame music
	MusicManager:SetState("MINIGAME")
	
	--Set Tutorial on first game - but don't overide quest conversations
	local hero = HeroManager:GetActiveHero()
	if hero then
		if (not self.gameOptions or self.gameOptions and not self.gameOptions.conversation) and not hero:GetAttribute("hoardTutorialShown") then
			self.gameOptions = self.gameOptions or {}
			self.gameOptions.conversation = "HOARD_TUTORIAL"
			hero:SetAttribute("hoardTutorialShown",true)
		end
		
		-- Play conversation if set
		if self.gameOptions and self.gameOptions.conversation then
			open_conversation_menu(
				"Assets/Conversations/" .. self.gameOptions.conversation .. ".xml",
				hero:GetName(), 
				hero:GetConversationPortrait(), 
				hero:GetGenderForConversation(), 
				"", 
				"",
				"", 
				function () 
					-- conversation finished, start game
				end, 
				x or GetScreenWidth()*0.5,
				y or GetScreenHeight()*0.5
			)
		end
	end
	
	-- Update presence
	self.previousPresence = Presence:SaveRichPresenceMode()
	
	Presence:SetPresenceMode("MINIGAME")
	Presence:SetPresenceContext("MINIGAME", "HOARD")

	return cGameMenu.OnOpen(self)
end

function HoardGameMenu:OnRefresh()
	MenuHelpers.Init()
	
	self:set_progress_max("prog_turns", MAX_TURNS)
	self:set_progress_max("prog_gold", MAX_GOLD)
	
	self:hide_widget("icon_close_highlight")
	self:hide_widget("icon_info_highlight")
	
	if self.lqitems then
		self:UpdateLQDisplay()
	end
	if self.hqitems then
		self:UpdateHQDisplay()
	end
	
	self.view:RefreshUI()
end

function HoardGameMenu:Initialise(layout)
	self.view = GameObjectManager:ConstructLocal("PHRD")
	
	local gemProbabilityTable = ProbabilityTable:Construct(Configuration.GEM_DISTRIBUTION)
	
	self.view:Initialise(layout, gemProbabilityTable, self.difficulty, self.endMinigame)
	
	self:SetWorld(self.view)
	
	self.tradeProbabilityTable = ProbabilityTable:Construct(Configuration.TRADE_DISTRIBUTION)
end

function HoardGameMenu:OnButton(buttonId, clickX, clickY)
	if buttonId == self:get_widget_id("butt_info") then
		open_message_menu("[HOARD_HELP_HEAD]", "[HOARD_HELP_TEXT]")
	elseif buttonId == self:get_widget_id("butt_close") then
		self:OpenPauseMenu()
	else
		return cGameMenu.OnButton(self, buttonId, clickX, clickY)
	end
	
	return Menu.MESSAGE_HANDLED
end

function HoardGameMenu:OnKey(key)
	if key == Keys.SK_ESCAPE then
		self:OpenPauseMenu()
	elseif key == Keys.SK_F12 then
		if not IsComputerBuildslaveAndBuildConfigRetail() then
			if self.view then
				self:UpdateGold(self.gold + math.random(15, 1500))
				self:GiveLQItem()
				self:GiveHQItem()
				self.view:GameWon()
			end
		end
	elseif key == Keys.SK_F11 then
		if not IsComputerBuildslaveAndBuildConfigRetail() then
			self:GiveLQItem()
		end
	elseif key == Keys.SK_F10 then
		if not IsComputerBuildslaveAndBuildConfigRetail() then
			self:GiveHQItem()
		end
	elseif key == Keys.SK_F10 then
		if not IsComputerBuildslaveAndBuildConfigRetail() then
			self:UpdateGold(self.gold + 555)
		end
	end
	
	return cGameMenu.OnKey(self, key)
end

function HoardGameMenu:UpdateTurns(turns)
	local turns_left = MAX_TURNS - turns
	MenuHelpers.UpdateStringTag(self, "turns", turns_left, "str_turns", "[TURNS_X_OF_Y]", MAX_TURNS)
	MenuHelpers.UpdateProgressBar(self, turns_left, "prog_turns")
end

function HoardGameMenu:UpdateGold(gold)
	self.gold = gold
	MenuHelpers.UpdateStringTag(self, "gold", gold, "str_gold", "[GOLD_X]")
	MenuHelpers.UpdateProgressBar(self, gold, "prog_gold")
end

function HoardGameMenu:UpdateItems(numLqItems, numHqItems)
	if numLqItems > self.numLqItems then
		for i=1, (numLqItems - self.numLqItems) do
			self:GiveLQItem()
		end
		self.numLqItems = numLqItems
	end
	
	if numHqItems > self.numHqItems then
		for i=1, (numHqItems - self.numHqItems) do
			self:GiveHQItem()
		end
		self.numHqItems = numHqItems
	end
end

function HoardGameMenu:GiveLQItem()
	local trade = self.tradeProbabilityTable:RandomItem()
	local item = CreateTradeItem(trade, math.random(self.difficulty, 2+self.difficulty))
	table.insert(self.lqitems, item)
	self:UpdateLQDisplay()
end

function HoardGameMenu:UpdateLQDisplay()
	local item_id
	for i=1, MAX_LQITEMS_SHOWN do
		if #self.lqitems > MAX_LQITEMS_SHOWN then
			item_id = #self.lqitems - (MAX_LQITEMS_SHOWN - i)
		else
			item_id = i
		end
		if self.lqitems[item_id] then
			self:set_image(string.format("icon_chest%d", i + MAX_HQITEMS_SHOWN), self.lqitems[item_id]:GetIconAsset())
		end
	end
end

function HoardGameMenu:GiveHQItem()
	local level = math.random(1, 50)
	if self.hero then
		level = self.hero:GetAttribute("level")
	end
	local item = ItemGenerator.RandomEquippableItemWithLevelVariance({level = level, baseType = nil, race = nil, rarity = "RareChestRandom"})
	assert(item, string.format("HoardGameMenu:GiveHQItem() failed to create a random item of params {level = %s, baseType = nil, race = nil, rarity = \"RareChestRandom\"}", tostring(level)))
	table.insert(self.hqitems, item)
	self:UpdateHQDisplay()
	
	local rarity = item:GetRarity()
	SCREENS.EffectsMenu:PlaySpellTextEffect(string.format("[%s_ITEM]", string.upper(rarity)))
end

function HoardGameMenu:UpdateHQDisplay()
	local item_id
	for i=1, MAX_HQITEMS_SHOWN do
		if #self.hqitems > MAX_HQITEMS_SHOWN then
			item_id = #self.hqitems - (MAX_HQITEMS_SHOWN - i)
		else
			item_id = i
		end
		if self.hqitems[item_id] then
			self:set_image(string.format("icon_chest%d", i), self.hqitems[item_id]:GetIconAsset())
		end
	end
end

function HoardGameMenu:GetGoldAndItems()
	local items = {}
	
	if self.hqitems then
		for _, item in ipairs(self.hqitems) do
			table.insert(items, item)
		end
	end
	
	if self.lqitems then
		for _, item in ipairs(self.lqitems) do
			table.insert(items, item)
		end
	end
	
	self.hqitems = nil
	self.lqitems = nil
	
	return self.gold, items
end

function HoardGameMenu:OnClose()
	require("FXContainer").StopAll()
	self.board = nil
	self.difficulty = nil
	self.endMinigame = nil
	self.gameOptions = nil
	self.selected = nil
	
	if self.lqitems then
		for _, item in ipairs(self.lqitems) do
			GameObjectManager:Destroy(item)
		end
	end
	
	if self.hqitems then
		for _, item in ipairs(self.hqitems) do
			GameObjectManager:Destroy(item)
		end
	end
	
	self.lqitems = nil
	self.hqitems = nil
	
	GameObjectManager:Destroy(self.view)
	self.view = nil
	self:SetWorld(nil)
	
	
	
	SoundSystem.Unload("board_bank")

	-- Return to previous music
	if rawget(SCREENS, "WorldMapMenu") and SCREENS.WorldMapMenu:IsOpen() then
		MusicManager:SetState("WORLD_BEGIN")
	else
		MusicManager:SetState("MENU")
	end
	
	-- Restore previous presence information
	Presence:RestoreRichPresence(self.previousPresence)
	
	if is_message_menu_open() then
		close_message_menu()
	end
	
	return cGameMenu.OnClose(self)
end

function HoardGameMenu:OpenPauseMenu()
	SCREENS.PauseMenu:Open("HoardGameMenu", function(quit)
		if quit then
			if self.endMinigame then
				CallScreenSequencer("HoardGameMenu", self.endMinigame)
			else
				CallScreenSequencer("HoardGameMenu", "MinigameSelectionMenu")
			end
		else
			self.hintdelay = Settings:Read("hintarrowdelay", 5000)
			self.view:RemoveHint()
			self.view:SetupHint()
		end
	end)
end

function HoardGameMenu:OnGamepadButton(user_id, button_id, value)
	if value == 0 then
		return Menu.MESSAGE_HANDLED
	end
	
	LOGF("HoardGameMenu:OnGamepadButton(user_id=%d, button_id=%d, value=%d)", user_id, button_id, value)
	
	if button_id == Gamepad.MapGenericInput(user_id, GamepadInput.GAMEPAD_BUTTON_OKAY) then
		if not self.selected then
			self.selected = true
			self.view:RemoveHint()
			self.view:SetSpinSpeed(true)
		else
			self.selected = false
			self.view:SetupHint()
			self.view:SetSpinSpeed(false)
		end
	elseif button_id == Gamepad.MapGenericInput(user_id, GamepadInput.GAMEPAD_BUTTON_CANCEL) then
		if self.selected then
			self.selected = false
			self.view:SetupHint()
			self.view:SetSpinSpeed(false)
		end
	elseif button_id == Gamepad.MapGenericInput(user_id, GamepadInput.GAMEPAD_BUTTON_START) then
		self:OpenPauseMenu()
	
	elseif button_id == Gamepad.MapGenericInput(user_id, GamepadInput.GAMEPAD_BUTTON_SHOULDER_RIGHT) then
		if not IsComputerBuildslaveAndBuildConfigRetail() then
			if self.view then
				self:UpdateGold(self.gold + math.random(15, 1500))
				self:GiveLQItem()
				self:GiveHQItem()
				self.view:GameWon()
			end
		end
	elseif button_id == Gamepad.MapGenericInput(user_id, GamepadInput.GAMEPAD_BUTTON_SHOULDER_LEFT) then
		if not IsComputerBuildslaveAndBuildConfigRetail() then
			self:UpdateGold(self.gold + 555)
		end
	elseif button_id == Gamepad.MapGenericInput(user_id, GamepadInput.GAMEPAD_BUTTON_CROSS_LEFT) then
		if not IsComputerBuildslaveAndBuildConfigRetail() then
			self:GiveLQItem()
			self:GiveHQItem()
		end
	end
	
	return Menu.MESSAGE_HANDLED
end

function HoardGameMenu:OnGamepadDPad(user_id, dpad_id, x_value, y_value)
	if x_value == 0 and y_value == 0 then
		return Menu.MESSAGE_HANDLED
	end
	
	if user_id then		--This line needs to be removed after the OnGamepadJoystick bug is fixed
		LOGF("HoardGameMenu:OnGamepadDPad(user_id=%d, dpad_id=%d, x_value=%d, y_value=%d", user_id, dpad_id, x_value, y_value)
	end
	
	if self.selected or (dpad_id == Gamepad.MapGenericInput(user_id, GamepadInput.GAMEPAD_JOYSTICK_2)) then
		self.selected = false
		self.view:RemoveHint()
		self.view:DoGamepadSwap(x_value, y_value)
	else
		self.view:ShowGamepadSelection(x_value, y_value)
	end
	
	return Menu.MESSAGE_HANDLED
end

--These functions need to be removed after the OnGamepadJoystick bug is fixed
function HoardGameMenu:OnGamepadJoystick(user, joystick, x, y)
	if _G.Joystick then
		_G.Joystick.OnGamepadJoystick(user, joystick, x, y)
	end
	return Menu.MESSAGE_HANDLED
end

function HoardGameMenu:OnTimer(time)
	if _G.Joystick then
		_G.Joystick.OnTimer(self, time)
	end
	return Menu.MESSAGE_HANDLED
end

return ExportSingleInstance("HoardGameMenu")
