use_safeglobals()

local MenuHelpers = require("Utils/MenuHelpers")

class "SelectHeroMenu" (Menu);

function SelectHeroMenu:__init()
	super()
	
	self:Initialize("Assets\\Screens\\SelectHeroMenu.xml")
end

function SelectHeroMenu:OnOpen()
	assert(self.callback, "SelectHeroMenu was not opened correctly.  Please open it with SetupAndOpen")

	LoadAssetGroup("Rooms/Thumbnails")
	
	if HeroManager:AreSavesInvalid() or HeroManager:HasDeviceChangedSinceLoad() then
		self:ReloadHeroSaves()
	else
		self:SetupHeroList()
	end
	
	return Menu.OnOpen(self)
end

function SelectHeroMenu:OnRefresh()
	if self.limitToContinue then
		self:hide_widget("butt_new")
		self:hide_widget("butt_delete")
	else
		self:activate_widget("butt_new")
		self:activate_widget("butt_delete")
	end
	
	if self.saveList and #self.saveList > 1 then
		self:activate_widget("butt_next")
		self:activate_widget("butt_prev")
	else 
		self:hide_widget("butt_next")
		self:hide_widget("butt_prev")
	end
	
	if self.currentHeroIndex then
		self:SetCurrentHero(self.currentHeroIndex)
	end
end

function SelectHeroMenu:ReloadHeroSaves()
	self.saveList = nil
	
	if is_yesno_menu_open() then
		close_yesno_menu(false, false)
	end
	
	if not _G.IsDemo() then
		if not IsOpen("CustomLoadingMenu") then
			SCREENS.CustomLoadingMenu:SetupAndOpen(nil, function()
				HeroManager:SetPlayer(1)
				local function Continue()
					if HeroManager:AreSavesInvalid() then
						self:ReloadHeroSaves()
					else
						self:SetupHeroList()
					end
				end
				local function Shutdown(heading, message) 
					self:activate_widget("str_text")
					_G.ResetToSplashWithError(heading, message)
				end
				HeroManager:LoadHeroes(Continue, Shutdown)
				SCREENS.CustomLoadingMenu:Close()
			end, "[LOADING]")
		end
	else
		CallScreenSequencer("SelectHeroMenu", "CreateHeroMenu")
	end
end

function SelectHeroMenu:HeroesLost()
	HeroManager:ClearHeroes()
	self:StartAnimation("fadeout")
end

function SelectHeroMenu:SetupHeroList()
	self.saveList = HeroManager:GetHeroList()
	if (#self.saveList == 0) then
		self:HeroesLost()
	else
		local last_save
		if Settings:ValueExists("last_save") then
			last_save = Settings:Read("last_save", "")
		end
		self:SetCurrentHero(1)

		if last_save and last_save ~= "" then
			for index, heroData in ipairs(self.saveList) do
				if heroData.saveHandle and (heroData.saveHandle:GetName() == last_save) then
					self:SetCurrentHero(index)
					break -- early out
				end
			end
		end
		
		if #self.saveList <= 1 then
			self:hide_widget("butt_next")
			self:hide_widget("butt_prev")
		else 
			self:activate_widget("butt_next")
			self:activate_widget("butt_prev")
		end
	end
end

function SelectHeroMenu:OnClose()
	self.callback = nil
	self.limitToContinue = nil
	
	UnloadAssetGroup("Rooms/Thumbnails")
	
	return Menu.OnClose(self)
end

function SelectHeroMenu:SetupAndOpen(callback, limitToContinue)
	self.callback = assert(callback)
	self.limitToContinue = limitToContinue
	self:Open()
end

function SelectHeroMenu:SetCurrentHero(index)
	assert(type(index) == "number")
	
	self.currentHeroIndex = index
	
	local heroData = self.saveList[index]
	if not heroData.hero then
		HeroManager:RetrieveHero(index)
	end
	
	local activeHero = heroData.hero
	if activeHero then
		activeHero:InitializeStats()
		activeHero:CalculateStats()
		--activeHero:PrintDebug(true)

		HeroLoadTutorialSystem(activeHero)	
		
		self:set_image("icon_portrait", string.format("%s_M", activeHero:GetAttribute("portrait")))
		self:set_text_raw("str_name_value", activeHero:GetAttribute("name"))
		self:set_text_raw("str_level_value", tostring(activeHero:GetAttribute("level")))
		self:set_text_raw("str_health_value", tostring(activeHero:GetHealth()))
		
		local gold_val = activeHero:GetAttribute("gold")
		local text_val, real_gold_val = MenuHelpers.ShortenValue(gold_val)
		
		if real_gold_val and text_val then
			self:set_text_raw("str_gold_value", substitute(translate_text(text_val), tostring(real_gold_val)))
		else
			self:set_text_raw("str_gold_value", tostring(gold_val))
		end

		-- set the save_image to be a room thumbnail coz we don't have the save snapshots working yet and they may not get done.
		local room = activeHero:GetCurrentRoom()
		local thumb = string.format("image_room_thumb_%s", room)
		self:set_image("icon_save_image", thumb)
		self:activate_widget("butt_continue")
		
		if IsGamepadActive() then
			MenuSystem.SetGamepadWidget(self, "butt_back")
			MenuSystem.SetGamepadWidget(self, "butt_continue")
		end
	elseif heroData.corrupt then
		self:set_image("icon_portrait", "")
		self:set_image("icon_save_image", "")
		self:set_text("str_name_value", "[CORRUPT_SAVE]")
		self:set_text("str_level_value", "[BLANK]")
		self:set_text("str_health_value", "[BLANK]")
		self:set_text("str_gold_value", "[BLANK]")
		self:deactivate_widget("butt_continue")
		open_yesno_menu("[ERROR]","[LOAD_FAILURE]", function(continue)
			if not continue then
				SCREENS.SelectHeroMenu:DeleteCurrentHero()
			end
		end, "[DELETE]", "[CONTINUE]")
	end
end

function SelectHeroMenu:OnKey(key)
	if key == Keys.SK_ESCAPE then
		Settings:Save()
		HeroManager:ClearHeroes()
		self:StartAnimation("fadeout")
	end
	
	return Menu.OnKey(self, key)
end

function SelectHeroMenu:DeleteCurrentHero()
	HeroManager:DeleteHero(self.currentHeroIndex, function (success) 
		if success then
			open_message_menu("[DELETE_HEADING]","[DELETE_CONFIRM]", function()
				self.saveList = HeroManager:GetHeroList()
				if #self.saveList == 0 then
					CallScreenSequencer("SelectHeroMenu", "CreateHeroMenu")
				else
					if #self.saveList <= 1 then
						self:hide_widget("butt_next")
						self:hide_widget("butt_prev")
					end
					
					if self.currentHeroIndex - 1 < 1 then
						self:SetCurrentHero(#self.saveList)
					else
						self:SetCurrentHero(self.currentHeroIndex - 1)
					end
				end
			end)
		else
			open_yesno_menu("[WARNING]", "[RETRY_DELETE_MSG]", function(continue)
				if not continue then
					self:DeleteCurrentHero()
				end
			end, "[RETRY]", "[CONTINUE]")
		end
	end)
end

function SelectHeroMenu:OnButton(buttonId, clickX, clickY)
	if buttonId == self:get_widget_id("butt_back") then
		Settings:Save()
		HeroManager:ClearHeroes()
		self:StartAnimation("fadeout")
	elseif buttonId == self:get_widget_id("butt_continue") then
		HeroManager:SetActiveHero(self.currentHeroIndex)
		HeroManager:ClearNonActiveHeroes()

		local saveHandle = HeroManager:GetActiveHeroSaveHandle()
		if saveHandle then
			Settings:Write("last_save", saveHandle:GetName())
			Settings:Save()
		end

		self.callback(true)
	elseif buttonId == self:get_widget_id("butt_new") then
		if HeroManager:GetNumHeroes() < HeroManager:GetMaxNumberOfSaves() then
			HeroManager:ClearHeroes()
			CallScreenSequencer("SelectHeroMenu", "CreateHeroMenu")
		else
			open_message_menu("[WARNING]","[TOO_MANY_SAVES]")
		end
	elseif buttonId == self:get_widget_id("butt_delete") then
		open_yesno_menu("[DELETE_HEADING]", "[DELETE_MSG]", function (res)
			if not res then
				SCREENS.SelectHeroMenu:DeleteCurrentHero()
			end
		end, "[YES]", "[NO]")
		
	elseif buttonId == self:get_widget_id("butt_next") then
		if self.saveList then
			if self.currentHeroIndex + 1 > #self.saveList then
				self:SetCurrentHero(1)
			else
				self:SetCurrentHero(self.currentHeroIndex + 1)
			end
		end
	elseif buttonId == self:get_widget_id("butt_prev") then
		if self.saveList then
			if self.currentHeroIndex - 1 < 1 then
				self:SetCurrentHero(#self.saveList)
			else
				self:SetCurrentHero(self.currentHeroIndex - 1)
			end
		end
	else
		return Menu.OnButton(self, buttonId, clickX, clickY)
	end
	
	return Menu.MESSAGE_HANDLED   
end

function SelectHeroMenu:OnAnimFadeout(data) 
	if (data == 1) then
		self.callback(false)
	end
end

dofile("Assets/Scripts/Screens/SelectHeroMenuPlatform.lua")

-- return an instance of SelectHeroMenu
return ExportSingleInstance("SelectHeroMenu")
